//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						(c) 1997 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalViewables.cc	--- implementation of CalViewable classes
//
//	DESCRIPTION:
//		This is the implementation part of the CalViewable classes
//
//	CHANGE HISTORY:
//		UL, 1999-03-23: 1.0.0: - first version
//		UL, 1999-09-16: 1.1.0: - added AbsCoef viewable implementation
//		UL, 1999-09-17: 1.2.0: - added CrossPsf viewable implementation
//		UL, 1999-10-21: 1.3.0: - added HKParmInt viewable implementation
//		UL, 2000-02-02: 1.4.0: - added FilterTransmission viewable
//								 implementation
//		UL, 2000-02-07: 1.5.0: - bug fix in AbsCoefSubViewable()
//		UL, 2000-02-17: 1.6.0: - added QuantumEfficiency viewable/subviewable
//								 implementation
//		UL, 2000-03-08: 1.7.0: - fixed minor compilation problem under
//								 gcc-2.95.2
//		UL, 2000-03-24: 1.8.0: - added Viewable effectiveArea + Vignetting
//		UL, 2000-04-25: 1.9.0: - added Viewable GratingProperties +
//								 subviewables
//		UL, 2000-06-14: 1.10.0:- in CcdPlot viewable: set
//									<randomize>=false
//									<ocb>=1
//		UL, 2000-08-30: 1.11.0:- new viewables: obscuration/intercept eff.
//								 vs. theta/phi
//		UL, 2000-09-05: 1.12.0:- added Psf Viewable
//							   - bug fix: forgot DEG->RAD conversion of phi
//								 in several viewables
//		UL, 2000-11-15: 1.12.0:- added RgaMisalignmentDistributionViewable
//		UL, 2000-11-20: 1.13.0:- added RgaScatteringDistributionViewable
//		UL, 2000-11-23: 1.14.0:- added DispersedPsfPlotViewable
//							   - CrossPsfPlot now gives 5 curves for different
//								 betas
//		UL, 2001-01-10: 1.15.0:- CrossPsf Viewable does now take theta/phi
//								 into account to compute source position
//								 in X-dispersion direction
//							   - added SelfVignetting-vs-beta plot as
//								 SubViewable of GratingDataServer
//							   - give incidence angle in all Grating Viewables
//		UL, 2001-01-18: 1.16.0:- fixed minor problem with plot title introduced
//								 in 1.15.0
//		UL, 2001-02-15: 1.17.0:- RGA scattering distribtion plot shows now
//								 also small + large angle components
//		UL, 2001-02-21: 1.18.0:- added RGA bowingDistribution viewable
//							   - added SubViewable RowlandDefocusPlot to
//								 GratingDataServer
//		UL, 2001-05-01: 1.19.0:- added SubViewable RowlandDefocusDist to
//								 GratingDataServer
//
//	AUTHOR:
//		U. Lammers, Wed Mar 24 16:24:14 MET 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/calview/CalViewables.cc,v 1.50 2001/05/01 12:41:20 ulammers Exp $
//

#include <cstdio>
#include <fstream>
#include <strstream>
#include <algorithm>
#include "CalServer.h"
#include "Xmm.h"
#include "common.h"
#include "CalViewables.h"
#include "errstr.h"

#include "Redistribution.h"
#include "HKvalRangesServer.h"
#include "AbsorptionDataServer.h"
#include "EncircledEnergyDataServer.h"
#include "FilterDataServer.h"
#include "GratingDataServer.h"
#include "CcdQuantumDataServer.h"
#include "EffectiveAreaDataServer.h"
#include "PsfDataServer.h"
#include "OmPsfDataServer.h"

//
//	---( AbsCoefPlot + SubViewables )-----------------------------------------
//
class AbsCoefSubViewable : public CalViewable {
public:
	AbsCoefSubViewable(const string &elementName,
					   const CalViewServer *server = 0)
		: CalViewable(server), _elementName(elementName) {
		PROC

		_name = _elementName;
		strstream s;
		s<<"Absorption coefficients for element "<<_name<<ends;
		char *p = s.str();
		_whatIs = p;
		delete p;

		ENDPROC
	} // AbsCoefSubViewable()

	CalViewProductDescriptor view(const string &fileName = "") const {
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Absorption Coefficients for "<<_elementName
		   <<" (Henke data)\"\n";
		set<<"@xaxis label \"energy [eV]\"\n";
		set<<"@yaxis label \"[arb. units]\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";

		//
		//	write data
		//
		set<<"@target s0"<<endl;
		set<<"@s0 color 1"<<endl;
		set<<"@type xy"<<endl;

		//	get the atom
		AbsorptionDataServer *abds = calServer.getAtom(abds);
		vector<const AbsCoefCurve *> f;
		f.push_back(&abds->henke(_elementName).f1());
		f.push_back(&abds->henke(_elementName).f2());

		const double fromE = 40.,
					 toE = 15000.,
					 inStepsOf = 2.;

		for (unsigned n=0; n<f.size(); ++n) {
			for (double E = fromE; E<=toE; E+=inStepsOf)
				set<<E<<" "<<(*f[n])(E)<<endl;
			set<<"&\n";
		}
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // AbsCoefSubViewable::view()

	~AbsCoefSubViewable() { PROC ENDPROC }

private:
	string	_elementName;
};


AbsCoefPlot::AbsCoefPlot(const CalViewServer *server)
		: CalViewable(server)
{
	PROC

	_name = "AbsCoef";
	_whatIs = "Absorption coefficients";

	//	get the atom and register SubViewables
	AbsorptionDataServer *abds = calServer.getAtom(abds);
	vector<string> materials = abds->materials();
	for (unsigned n=0; n<materials.size(); ++n)
		_subViewable.push_back(new AbsCoefSubViewable(materials[n], server));

	ENDPROC
} // AbsCoefPlot::AbsCoefPlot()


CalViewProductDescriptor AbsCoefPlot::view(const string &setName) const
{
	PROC
	RETURN _subViewable[0]->view(setName);
} // AbsCoefPlot::view()


//
//	---( GratingProperties + SubViewables )------------------------------------
//
class ReflectionEfficiencyViewable : public CalViewable {
public:
	ReflectionEfficiencyViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "Reflection efficiencies";
		_whatIs = "Reflection efficiency of grating";

		ENDPROC
	} // ReflectionEfficiencyViewable()

	CalViewProductDescriptor view(const string &fileName = "") const {
		PROC

		//	get the atom
		GratingDataServer *gds = calServer.getAtom(gds);

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double theta = p->state()->theta.get(),
			   phi = p->state()->phi.get(),
			   alphaIn = gds->alpha(theta, phi * DEG_TO_RAD) * RAD_TO_DEG;

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Reflection efficiencies\"\n";
		set<<"@subtitle \""
		   <<Xmm::toString(ins)<<" - "
		   <<"theta="<<theta<<"'' "
		   <<"phi="<<phi<<"deg "
		   <<"alpha="<<setprecision(5)<<alphaIn<<"deg\"\n";
		set<<"@xaxis label \"energy [eV]\"\n";
		set<<"@yaxis label \"efficiency\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";
		set<<"@legend on\n";

		for (int order=gds->maxOrder; order>=gds->minOrder; --order) {
			//
			//	write data
			//
			int setnr = gds->maxOrder-order;
			set<<"@target s"<<setnr<<endl;
			set<<"@type xy"<<endl;
			set<<"@s"<<setnr<<" legend \"order: "<<order<<"\"\n";

			InterpolatedCurve f = gds->reflectionEfficiency(theta,
															phi * DEG_TO_RAD,
															order);
			const double fromE = 200.,
						 toE = 2500.,
						 inStepsOf = 1.;

			for (double E = fromE; E<=toE; E+=inStepsOf)
				set<<E<<" "<<f(E)<<endl;
			set<<"&\n";
		}
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // ReflectionEfficiencyViewable::view()

	~ReflectionEfficiencyViewable() { PROC ENDPROC }
};


class InterceptObscurationEfficiencyVsEViewable : public CalViewable {
public:
	InterceptObscurationEfficiencyVsEViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "vs E";
		_whatIs = "Intercept/obscuration efficiency of grating";

		ENDPROC
	} // InterceptObscurationEfficiencyViewable()

	CalViewProductDescriptor view(const string &fileName = "") const {
		PROC

		//	get the atom
		GratingDataServer *gds = calServer.getAtom(gds);

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double theta = p->state()->theta.get(),
			   phi = p->state()->phi.get(),
			   alphaIn = gds->alpha(theta, phi * DEG_TO_RAD) * RAD_TO_DEG;

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Intercept/obscuration efficiencies\"\n";
		set<<"@subtitle \""
		   <<Xmm::toString(ins)<<" - "
		   <<"theta="<<theta<<"'' "
		   <<"phi="<<phi<<"deg "
		   <<"alpha="<<setprecision(5)<<alphaIn<<"deg\"\n";
		set<<"@xaxis label \"energy [eV]\"\n";
		set<<"@yaxis label \"efficiency\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";
		set<<"@legend on\n";

		//
		//	write data
		//
		set<<"@target s0"<<endl;
		set<<"@type xy"<<endl;
		set<<"@s0 legend \"intercept\"\n";

		InterpolatedCurve f = gds->interceptEfficiency(theta, phi *DEG_TO_RAD);
		const double fromE = 100.,
					 toE = 15000.,
					 inStepsOf = 2.;

		for (double E = fromE; E<=toE; E+=inStepsOf)
			set<<E<<" "<<f(E)<<endl;
		set<<"&\n";

		set<<"@target s1"<<endl;
		set<<"@type xy"<<endl;
		set<<"@s1 legend \"obscuration\"\n";
		f = gds->obscurationEfficiency(theta, phi);
		for (double E = fromE; E<=toE; E+=inStepsOf)
			set<<E<<" "<<f(E)<<endl;
		set<<"&\n";
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // InterceptObscurationEfficiencyVsEViewable::view()

	~InterceptObscurationEfficiencyVsEViewable() { PROC ENDPROC }
};


class InterceptObscurationEfficiencyVsThetaViewable : public CalViewable {
public:
	InterceptObscurationEfficiencyVsThetaViewable(const CalViewServer *server
		= 0) : CalViewable(server) {
		PROC

		_name = "vs theta";
		_whatIs = "Intercept/obscuration efficiency of grating";

		ENDPROC
	} // InterceptObscurationEfficiencyViewable()

	CalViewProductDescriptor view(const string &fileName = "") const {
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double E = p->state()->E.get(),
			   phi = p->state()->phi.get();

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Intercept/obscuration efficiencies\"\n";
		set<<"@subtitle \""
		   <<Xmm::toString(ins)<<" - "
		   <<"E="<<E<<"eV "
		   <<phi<<"deg\"\n";
		set<<"@xaxis label \"energy [eV]\"\n";
		set<<"@yaxis label \"efficiency\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";
		set<<"@legend on\n";

		//	get the atom
		GratingDataServer *gds = calServer.getAtom(gds);

		//
		//	write data
		//
		set<<"@target s0"<<endl;
		set<<"@type xy"<<endl;
		set<<"@s0 legend \"intercept\"\n";

		const double fromTheta = 0.,
					 toTheta = 900.,
					 inStepsOf = 5.;

		InterpolatedCurve f;
		for (double theta = fromTheta; theta <=toTheta; theta += inStepsOf) {
			f = gds->interceptEfficiency(theta, phi * DEG_TO_RAD);
			set<<theta<<" "<<f(E)<<endl;
		}
		set<<"&\n";

		set<<"@target s1"<<endl;
		set<<"@type xy"<<endl;
		set<<"@s1 legend \"obscuration\"\n";
		for (double theta = fromTheta; theta <=toTheta; theta += inStepsOf) {
			f = gds->obscurationEfficiency(theta, phi * DEG_TO_RAD);
			set<<theta<<" "<<f(E)<<endl;
		}	
		set<<"&\n";
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // InterceptObscurationEfficiencyVsThetaViewable::view()

	~InterceptObscurationEfficiencyVsThetaViewable() { PROC ENDPROC }
};


class InterceptObscurationEfficiencyVsPhiViewable : public CalViewable {
public:
	InterceptObscurationEfficiencyVsPhiViewable(const CalViewServer *server
		= 0) : CalViewable(server) {
		PROC

		_name = "vs phi";
		_whatIs = "Intercept/obscuration efficiency of grating";

		ENDPROC
	} // InterceptObscurationEfficiencyViewable()

	CalViewProductDescriptor view(const string &fileName = "") const {
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double E = p->state()->E.get(),
			   theta = p->state()->theta.get();

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Intercept/obscuration efficiencies";
		set<<"@subtitle \""
		   <<Xmm::toString(ins)<<" - "
		   <<"E="<<E<<"eV "
		   <<"theta="<<theta<<"''\"\n";
		set<<"@xaxis label \"energy [eV]\"\n";
		set<<"@yaxis label \"efficiency\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";
		set<<"@legend on\n";

		//	get the atom
		GratingDataServer *gds = calServer.getAtom(gds);

		//
		//	write data
		//
		set<<"@target s0"<<endl;
		set<<"@type xy"<<endl;
		set<<"@s0 legend \"intercept\"\n";

		const double fromPhi = 0.,
					 toPhi = 360.,
					 inStepsOf = 5.;

		InterpolatedCurve f;
		for (double phi = fromPhi; phi<=toPhi; phi += inStepsOf) {
			f = gds->interceptEfficiency(theta, phi * DEG_TO_RAD);
			set<<phi<<" "<<f(E)<<endl;
		}
		set<<"&\n";

		set<<"@target s1"<<endl;
		set<<"@type xy"<<endl;
		set<<"@s1 legend \"obscuration\"\n";
		for (double phi = fromPhi; phi <=toPhi; phi += inStepsOf) {
			f = gds->obscurationEfficiency(theta, phi * DEG_TO_RAD);
			set<<phi<<" "<<f(E)<<endl;
		}	
		set<<"&\n";
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // InterceptObscurationEfficiencyVsPhiViewable::view()

	~InterceptObscurationEfficiencyVsPhiViewable() { PROC ENDPROC }
};


class InterceptObscurationEfficiencyViewable : public CalViewable {
public:
	InterceptObscurationEfficiencyViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "Intercept/obscuration efficienty";
		_whatIs = "Intercept/obscuration efficiency of grating";

		//	register SubViewables
		_subViewable.push_back(
					new	InterceptObscurationEfficiencyVsEViewable(server));
		_subViewable.push_back(
					new InterceptObscurationEfficiencyVsThetaViewable(server));
		_subViewable.push_back(
					new InterceptObscurationEfficiencyVsPhiViewable(server));

		ENDPROC
	} // InterceptObscurationEfficiencyViewable()


	CalViewProductDescriptor view(const string &setName) const
	{
		PROC
		RETURN _subViewable[0]->view(setName);
	} // InterceptObscurationEfficiencyViewable::view()

	~InterceptObscurationEfficiencyViewable() { PROC ENDPROC }
};


class SelfVignettingEfficiencyVsEViewable : public CalViewable {
public:
	SelfVignettingEfficiencyVsEViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "vs E";
		_whatIs = "Self-vignetting efficiency of grating";

		ENDPROC
	} // SelfVignettingEfficiencyVsEViewable()

	CalViewProductDescriptor view(const string &fileName = "") const {
		PROC

		GratingDataServer *gds = calServer.getAtom(gds);

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double theta = p->state()->theta.get(),
			   phi = p->state()->phi.get(),
			   alphaIn = gds->alpha(theta, phi * DEG_TO_RAD) * RAD_TO_DEG;

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Self-vignetting efficiencies";
		set<<"@subtitle \""
		   <<Xmm::toString(ins)<<" - "
		   <<"theta="<<theta<<"'' "
		   <<"phi="<<phi<<"deg "
		   <<"alpha="<<setprecision(5)<<alphaIn<<"deg\"\n";
		set<<"@xaxis label \"energy [eV]\"\n";
		set<<"@yaxis label \"efficiency\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";
		set<<"@legend on\n";

		//	get the atom
		const double fromE = 200.,
					 toE = 2500.,
					 inStepsOf = 1.;
		unsigned N = UNT((toE - fromE)/inStepsOf+1);
		vector<double> E(N);
		for (unsigned n=0; n<N; ++n)
			E[n] = fromE + n*inStepsOf;

		for (int order=gds->maxOrder; order>=gds->minOrder; --order) {
			//
			//	write data
			//
			int setnr = gds->maxOrder-order;
			set<<"@target s"<<setnr<<endl;
			set<<"@type xy"<<endl;
			set<<"@s"<<setnr<<" legend \"order: "<<order<<"\"\n";

			vector<double> v = gds->selfVignettingEfficiency(E, theta,
															 phi * DEG_TO_RAD,
															 order);
			for (unsigned n=0; n<N; ++n)
				set<<E[n]<<" "<<v[n]<<endl;
			set<<"&\n";
		}
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // SelfVignettingEfficiencyVsEViewable::view()

	~SelfVignettingEfficiencyVsEViewable() { PROC ENDPROC }
};


class SelfVignettingEfficiencyVsBetaViewable : public CalViewable {
public:
	SelfVignettingEfficiencyVsBetaViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "vs beta";
		_whatIs = "Self-vignetting efficiency of grating";

		ENDPROC
	} // SelfVignettingEfficiencyVsBetaViewable()

	CalViewProductDescriptor view(const string &fileName = "") const {
		PROC

		GratingDataServer *gds = calServer.getAtom(gds);

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double theta = p->state()->theta.get(),
			   phi = p->state()->phi.get(),
			   alphaIn = gds->alpha(theta, phi * DEG_TO_RAD) * RAD_TO_DEG;

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Self-vignetting efficiencies\"\n";
		set<<"@subtitle \""
		   <<Xmm::toString(ins)<<" - "
		   <<"theta="<<theta<<"'' "
		   <<"phi="<<phi<<"deg "
		   <<"alpha="<<setprecision(5)<<alphaIn<<"deg\"\n";
		set<<"@xaxis label \"beta [rad]\"\n";
		set<<"@yaxis label \"efficiency\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";

		const double fromBeta = 2.1*DEG_TO_RAD,
					 toBeta = 5.0*DEG_TO_RAD,
					 inStepsOf = .01*DEG_TO_RAD;
		unsigned N = UNT((toBeta - fromBeta)/inStepsOf+1);
		vector<double> beta(N);
		for (unsigned n=0; n<N; ++n)
			beta[n] = fromBeta + n*inStepsOf;

		//
		//	write data
		//
		set<<"@target s0\n";
		set<<"@type xy"<<endl;

		vector<double> v = gds->selfVignettingEfficiency(beta);
		for (unsigned n=0; n<N; ++n)
			set<<setprecision(8)<<beta[n]<<" "<<v[n]<<endl;
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // SelfVignettingEfficiencyVsBetaViewable::view()

	~SelfVignettingEfficiencyVsBetaViewable() { PROC ENDPROC }
};


class SelfVignettingEfficiencyViewable : public CalViewable {
public:
	SelfVignettingEfficiencyViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "Self-vignetting efficiency";
		_whatIs = "Self-vignetting efficiency of grating";

		//	register SubViewables
		_subViewable.push_back(
						new SelfVignettingEfficiencyVsBetaViewable(server));
		_subViewable.push_back(
						new SelfVignettingEfficiencyVsEViewable(server));

		ENDPROC
	} // SelfVignettingEfficiencyViewable()


	CalViewProductDescriptor view(const string &setName) const
	{
		PROC
		RETURN _subViewable[0]->view(setName);
	} // SelfVignettingEfficiencyViewable::view()
	
	~SelfVignettingEfficiencyViewable() { PROC ENDPROC }
};


class RgaMisalignmentDistributionViewable : public CalViewable {
public:
	RgaMisalignmentDistributionViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "RGA misalignment";
		_whatIs = "RGA Misalignment figure error distribution of grating";

		ENDPROC
	} // RgaMisalignmentDistributionViewable()

	CalViewProductDescriptor view(const string &fileName = "") const {
		PROC

		//	get the atom
		GratingDataServer *gds = calServer.getAtom(gds);

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double theta = p->state()->theta.get(),
			   phi = p->state()->phi.get(),
			   alphaIn = gds->alpha(theta, phi * DEG_TO_RAD) * RAD_TO_DEG,
			   E = p->state()->E.get();

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"RGA Misalignment Figure Distribution\"\n";
		set<<"@subtitle \""
		   <<Xmm::toString(ins)<<" - "
		   <<"theta="<<theta<<"'' "
		   <<"phi="<<phi<<"deg "
		   <<"alpha="<<setprecision(5)<<alphaIn<<"deg "
		   <<"E="<<E<<"eV\"\n";
		set<<"@xaxis label \"beta [rad]\"\n";
		set<<"@yaxis label \"probability\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";
		set<<"@legend on\n";

		for (int order=gds->maxOrder; order>=gds->minOrder; --order) {
			int setnr = gds->maxOrder-order;
			set<<"@target s"<<setnr<<endl;
			set<<"@type xy"<<endl;
			set<<"@s"<<setnr<<" legend \"order: "<<order<<"\"\n";

			double alphaIn = gds->alpha(theta, phi * DEG_TO_RAD),
				   betaOut = gds->beta(alphaIn, order, E),
				   chromaticMagnification =	sin(alphaIn)/sin(betaOut);
			const double NBETA = 10000.,
						 DELTA_BETA = .5,	//	beta span
						 binWidth = DELTA_BETA/NBETA;
			Spectrum spec(gds->misalignmentFigureDistribution(7.,
												   binWidth*DEG_TO_RAD,
												   chromaticMagnification));
			unsigned N = spec.size();
			for (unsigned n=0; n<N; ++n)
				set<<setprecision(6)
				   <<(DBL(n)-.5*DBL(N))*binWidth*DEG_TO_RAD+betaOut<<' '
				   <<setprecision(10)<<spec[n]<<endl;
			set<<"&\n";
		}
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // RgaMisalignmentDistributionViewable::view()

	~RgaMisalignmentDistributionViewable() { PROC ENDPROC }
};


class RgaBowingDistributionViewable : public CalViewable {
public:
	RgaBowingDistributionViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "RGA bowing";
		_whatIs = "Bowing figure error distribution of grating";

		ENDPROC
	} // RgaBowingDistributionViewable()


	CalViewProductDescriptor view(const string &fileName = "") const {
		PROC

		//	get the atom
		GratingDataServer *gds = calServer.getAtom(gds);

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double theta = p->state()->theta.get(),
			   phi = p->state()->phi.get(),
			   alphaIn = gds->alpha(theta, phi * DEG_TO_RAD) * RAD_TO_DEG,
			   E = p->state()->E.get();

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"RGA Bowing Figure Distribution\"\n";
		set<<"@subtitle \""
		   <<Xmm::toString(ins)<<" - "
		   <<"theta="<<theta<<"'' "
		   <<"phi="<<phi<<"deg "
		   <<"alpha="<<setprecision(5)<<alphaIn<<"deg "
		   <<"E="<<E<<"eV\"\n";
		set<<"@xaxis label \"beta [rad]\"\n";
		set<<"@yaxis label \"probability\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";
		set<<"@legend on\n";

		for (int order=gds->maxOrder; order>=gds->minOrder; --order) {
			int setnr = gds->maxOrder-order;
			set<<"@target s"<<setnr<<endl;
			set<<"@type xy"<<endl;
			set<<"@s"<<setnr<<" legend \"order: "<<order<<"\"\n";

			double alphaIn = gds->alpha(theta, phi * DEG_TO_RAD),
				   betaOut = gds->beta(alphaIn, order, E),
				   chromaticMagnification =	sin(alphaIn)/sin(betaOut);
			const double NBETA = 10000.,
						 DELTA_BETA = .5,	//	beta span
						 binWidth = DELTA_BETA/NBETA;
			Spectrum spec(gds->bowingFigureDistribution(7.,
												   binWidth*DEG_TO_RAD,
												   chromaticMagnification));
			unsigned N = spec.size();
			for (unsigned n=0; n<N; ++n)
				set<<setprecision(6)
				   <<(DBL(n)-.5*DBL(N))*binWidth*DEG_TO_RAD+betaOut<<' '
				   <<setprecision(10)<<spec[n]<<endl;
			set<<"&\n";
		}
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // RgaBowingDistributionViewable::view()

	~RgaBowingDistributionViewable() { PROC ENDPROC }
};

class RgaFigureErrorViewables : public CalViewable {
public:
	RgaFigureErrorViewables(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "Figure error distributions";
		_whatIs = "RGA figure error distributions";

		//	register subviewables
		_subViewable.push_back(
							new RgaMisalignmentDistributionViewable(server));
		_subViewable.push_back(new RgaBowingDistributionViewable(server));

		ENDPROC
	} // RgaFigureErrorViewables()


	CalViewProductDescriptor view(const string &setName) const
	{
		PROC
		RETURN _subViewable[0]->view(setName);
	} // view()


	~RgaFigureErrorViewables() { PROC ENDPROC }
};


class RgaScatteringDistributionViewable : public CalViewable {
public:
	RgaScatteringDistributionViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "RGA Scattering distribution";
		_whatIs = "Scattering distribution of grating";

		ENDPROC
	} // RgaScatteringDistributionViewable()

	CalViewProductDescriptor view(const string &fileName = "") const {
		PROC

		//	get the atom
		GratingDataServer *gds = calServer.getAtom(gds);

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double theta = p->state()->theta.get(),
			   phi = p->state()->phi.get(),
			   alphaIn = gds->alpha(theta, phi * DEG_TO_RAD),
			   E = p->state()->E.get();
		int	   order = p->state()->spectralOrder.get();
		double betaM = gds->beta(alphaIn, order, E);

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"RGA Scattering Distribution\"\n";
		set<<"@subtitle \""
		   <<Xmm::toString(ins)<<" - "
		   <<"theta="<<theta<<"'' "
		   <<"phi="<<phi<<"deg "
		   <<"alpha="<<setprecision(5)<<alphaIn*RAD_TO_DEG<<"deg "
		   <<"order="<<order<<' '
		   <<"E="<<E<<"eV\"\n";
		set<<"@xaxis label \"beta [rad]\"\n";
		set<<"@yaxis label \"probability\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";
		set<<"@legend on\n";

		bool contribs[][2] = { { true, true },	//	LA+SA
							   { true, false }, //	SA only
							   { false, true }	//	LA only
							 };
		char *legendstr[]   = { "total", "small angle", "large angle" };

		const double NBETA = 10000.,
					 DELTA_BETA = .5,	//	beta span
					 binWidth = DELTA_BETA/NBETA;
		for (unsigned n=0; n<NUM_OF(legendstr); ++n) {
			set<<"@target s"<<n<<endl;
			set<<"@type xy"<<endl;
			set<<"@s"<<n<<" legend \""<<legendstr[n]<<"\"\n";

			Spectrum spec(gds->scatteringDistribution(7.,
												   binWidth*DEG_TO_RAD,
												   alphaIn, order, E,
												   contribs[n][0],
												   contribs[n][1]));
			unsigned N = spec.size();
			for (unsigned n=0; n<N; ++n)
				set<<setprecision(6)
				   <<(DBL(n)-.5*DBL(N))*binWidth*DEG_TO_RAD+betaM<<' '
				   <<setprecision(10)<<spec[n]<<endl;
			set<<"&\n";
		}
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // RgaScatteringDistributionViewable::view()

	~RgaScatteringDistributionViewable() { PROC ENDPROC }
};


GratingPropertiesPlot::GratingPropertiesPlot(const CalViewServer *server)
		: CalViewable(server)
{
	PROC

	_name = "GratingProperties";
	_whatIs = "Grating properties";

	//	register SubViewables
	_subViewable.push_back(new ReflectionEfficiencyViewable(server));
	_subViewable.push_back(new InterceptObscurationEfficiencyViewable(server));
	_subViewable.push_back(new SelfVignettingEfficiencyViewable(server));
	_subViewable.push_back(new RgaFigureErrorViewables(server));
	_subViewable.push_back(new RgaScatteringDistributionViewable(server));

	ENDPROC
} // GratingPropertiesPlot::GratingPropertiesPlot()


CalViewProductDescriptor GratingPropertiesPlot::view(const string &setName)
const
{
	PROC
	RETURN _subViewable[0]->view(setName);
} // GratingPropertiesPlot::view()


bool GratingPropertiesPlot::isApplicable(Xmm::Instrument instrument) const
{
	PROC
	RETURN (instrument == Xmm::RGS1 || instrument == Xmm::RGS2);
} // GratingPropertiesPlot::isApplicable()	


//
//	---( DefocusLengths + SubViewables )------------------------------------
//
class RowlandDefocusDirectViewable : public CalViewable {
public:
	RowlandDefocusDirectViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "Defocus lengths (direct)";
		_whatIs = "Rowland defocus lengths";

		ENDPROC
	} // RowlandDefocusDirectViewable()

	CalViewProductDescriptor view(const string &fileName = "") const {
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		GratingDataServer *gd = calServer.getAtom(gd);
		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument inst = calServer.instrument()->id();
		double theta = p->state()->theta.get(),
			   phi = p->state()->phi.get(),
			   alphaIn = gd->alpha(theta, phi * DEG_TO_RAD) * RAD_TO_DEG;

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Rowland defocus lengths\"\n";
		set<<"@subtitle \""
		   <<Xmm::toString(inst)<<" - "
		   <<"theta="<<theta<<"'' "
		   <<"phi="<<phi<<"deg "
		   <<"alpha="<<setprecision(5)<<alphaIn<<"deg\"\n";
		set<<"@xaxis label \"X (CC1) [mm]\"\n";
		set<<"@yaxis label \"defocus length [mm]\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";
		set<<"@legend on\n";
		set<<"@target s0"<<endl;
		set<<"@type xy"<<endl;

		calServer.pushState();
		const XmmInstrument *ins = calServer.instrument();
		calServer.state()->randomize()->set(false);
		calServer.state()->ocb()->set(1);
	
		unsigned minX = ins->minX(), maxX = ins->maxX(),
				 minY = ins->minY(), maxY = ins->maxY(),
				 xPixels = maxX - minX + 1,
				 halfY = (minY+maxY)/2;
		short chipX[xPixels], chipY[xPixels];
		for (unsigned n=minX; n<=maxX; ++n) {
			chipX[n-1] = n;
			chipY[n-1] = halfY;
		}

		for (unsigned ccd = 1; ccd<=ins->nCcds(); ++ccd) {
			 calServer.state()->ccd()->set(ccd);

			CalInt16Vector chipXVec(chipX, xPixels),
						   chipYVec(chipY, xPixels);
			GeometryDataServer *cdf = calServer.getAtom(cdf);
			XmmPixelList pl = cdf->chipCoordList(chipXVec, chipYVec);
			CalReal32Vector x(0, xPixels), y(0, xPixels), z(0, xPixels);
			XmmCartesianPointList cpl = pl.toCamCoord1(x, y, z);
			CalReal32Vector xc(x.clone());
			CalReal32Vector	defl = cpl.rowlandDefocusLength(theta,
															phi*DEG_TO_RAD);

			//
			//	write data
			//
			for (unsigned n=0; n<xPixels; ++n)
				set<<xc[n]<<' '<<defl[n]<<endl;
		}
		set<<"&\n";
		set.close();

		calServer.popState();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // RowlandDefocusDirectViewable::view()

	~RowlandDefocusDirectViewable() { PROC ENDPROC }
};


class RowlandDefocusDistributionViewable : public CalViewable {
public:
	RowlandDefocusDistributionViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "Defocus lengths (distribution to LSF)";
		_whatIs = "Rowland defocus lengths distribution";

		ENDPROC
	} // RowlandDefocusDistributionViewable()

	CalViewProductDescriptor view(const string &fileName = "") const {
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		GratingDataServer *gds = calServer.getAtom(gds);
		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument inst = calServer.instrument()->id();
		double theta = p->state()->theta.get(),
			   phi = p->state()->phi.get(),
			   E = p->state()->E.get(),
			   alphaIn = gds->alpha(theta, phi * DEG_TO_RAD) * RAD_TO_DEG;

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Rowland defocus lengths distribution\"\n";
		set<<"@subtitle \""
		   <<Xmm::toString(inst)<<" - "
		   <<"theta="<<theta<<"'' "
		   <<"phi="<<phi<<"deg "
		   <<"alpha="<<setprecision(5)<<alphaIn<<"deg "
		   <<"E="<<E<<"eV\"\n";
		set<<"@xaxis label \"beta [rad]\"\n";
		set<<"@yaxis label \"probability\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";
		set<<"@legend on\n";
		set<<"@target s0"<<endl;
		set<<"@type xy"<<endl;

		GeometryDataServer *geomDS = calServer.getAtom(geomDS);
		for (int order=gds->maxOrder; order>=gds->minOrder; --order) {
			int setnr = gds->maxOrder-order;
			set<<"@target s"<<setnr<<endl;
			set<<"@type xy"<<endl;
			set<<"@s"<<setnr<<" legend \"order: "<<order<<"\"\n";

			double alphaIn = gds->alpha(theta, phi * DEG_TO_RAD),
				   betaM = gds->beta(alphaIn, order, E);
			const double NBETA = 10000.,
						 DELTA_BETA = .5,	//	beta span
						 binWidth = DELTA_BETA/NBETA;
			Spectrum spec(geomDS->lsfDefocusDistribution(7.,
												   binWidth*DEG_TO_RAD,
												   theta, phi, order, E));
			unsigned N = spec.size();
			for (unsigned n=0; n<N; ++n)
				set<<setprecision(6)
				   <<(DBL(n)-.5*DBL(N))*binWidth*DEG_TO_RAD+betaM<<' '
				   <<setprecision(10)<<spec[n]<<endl;
			set<<"&\n";
		}
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // RowlandDefocusDistributionViewable::view()

	~RowlandDefocusDistributionViewable() { PROC ENDPROC }
};


DefocusLengthPlot::DefocusLengthPlot(const CalViewServer *server)
	: CalViewable(server)
{
	PROC

	_name = "Defocus lengths";
	_whatIs = "Defocus lengths";

	//	register SubViewables
	_subViewable.push_back(new RowlandDefocusDirectViewable(server));
	_subViewable.push_back(new RowlandDefocusDistributionViewable(server));

	ENDPROC
} // DefocusLengthPlot::DefocusLengthPlot()


CalViewProductDescriptor DefocusLengthPlot::view(const string &setName)
const
{
	PROC
	RETURN _subViewable[0]->view(setName);
} // DefocusLengthPlot::view()


bool DefocusLengthPlot::isApplicable(Xmm::Instrument instrument) const
{
	PROC
	RETURN (instrument == Xmm::RGS1 || instrument == Xmm::RGS2);
} // DefocusLengthPlot::isApplicable()	


//
//	---( EffectiveAreaPlot + SubViewables )------------------------------------
//
class AeVsEViewable : public CalViewable {
public:
	AeVsEViewable(const CalViewServer *server = 0) : CalViewable(server) {
		PROC

		_name = "vs E";
		_whatIs = "effective area as function of energy";

		ENDPROC
	} // AeVsEViewable()

	CalViewProductDescriptor AeVsEViewable::view(const string &fileName)
	const
	{
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double theta = p->state()->theta.get(),
			   phi = p->state()->phi.get();
		int	spectralOrder = p->state()->spectralOrder.get();
		bool isRgs = ins==Xmm::RGS1 || ins==Xmm::RGS2;

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Effective area for "
		   <<Xmm::toString(ins)<<" at theta/phi = "
		   <<theta<<"'' / "<<phi<<"deg";
		if (isRgs)
		set<<" order="<<spectralOrder;
		set<<"\"\n";
		set<<"@xaxis label \"energy [eV]\"\n";
		set<<"@yaxis label \"effective area [cm^2]\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";

		//
		//	write data
		//
		set<<"@target s0"<<endl;
		set<<"@s0 color 1"<<endl;
		set<<"@type xy"<<endl;

		//	get the atom
		EffectiveAreaDataServer *aeds = calServer.getAtom(aeds);

		const double fromE = 0.,
					 toE = isRgs ? 2500. : 15000.,
					 inStepsOf = 1.;

		EffectiveAreaCurve f = !isRgs
			? aeds->effectiveAreaCurve(theta, phi * DEG_TO_RAD)
			: aeds->effectiveAreaCurve(spectralOrder, theta, phi * DEG_TO_RAD);

		for (double E = fromE; E<=toE; E+=inStepsOf)
			set<<E<<" "<<f(E)<<endl;
		set<<"&\n";
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // AeVsEViewable::view()


	~AeVsEViewable() { PROC ENDPROC }
};


class AeVsThetaViewable : public CalViewable {
public:
	AeVsThetaViewable(const CalViewServer *server = 0) : CalViewable(server) {
		PROC

		_name = "vs theta";
		_whatIs = "effective area as function of off-axis angle";

		ENDPROC
	} // AeVsThetaViewable()

	CalViewProductDescriptor AeVsThetaViewable::view(const string &fileName)
	const
	{
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double E = p->state()->E.get(),
			   phi = p->state()->phi.get();
		int	spectralOrder = p->state()->spectralOrder.get();
		bool isRgs = ins==Xmm::RGS1 || ins==Xmm::RGS2;

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Effective area for "
		   <<Xmm::toString(ins)<<" at E/phi = "
		   <<E<<"eV / "<<phi<<"deg";
		if (isRgs)
		set<<" order="<<spectralOrder;
		set<<"\"\n";
		set<<"@xaxis label \"off-axis angle ['']\"\n";
		set<<"@yaxis label \"effective area [cm^2]\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";

		//
		//	write data
		//
		set<<"@target s0"<<endl;
		set<<"@s0 color 1"<<endl;
		set<<"@type xy"<<endl;

		//	get the atom
		EffectiveAreaDataServer *aeds = calServer.getAtom(aeds);

		const double fromTheta = 0.,
					 toTheta = 900.,
					 inStepsOf = 5.;

		for (double theta = fromTheta; theta <=toTheta; theta += inStepsOf) {
			EffectiveAreaCurve f = !isRgs
				? aeds->effectiveAreaCurve(theta, phi * DEG_TO_RAD)
				: aeds->effectiveAreaCurve(spectralOrder, theta,
										  phi * DEG_TO_RAD);

			set<<theta<<" "<<f(E)<<endl;
		}
		set<<"&\n";
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // AeVsThetaViewable::view()


	~AeVsThetaViewable() { PROC ENDPROC }
};


class AeVsPhiViewable : public CalViewable {
public:
	AeVsPhiViewable(const CalViewServer *server = 0) : CalViewable(server) {
		PROC

		_name = "vs phi";
		_whatIs = "effective area as function of azimuth angle";

		ENDPROC
	} // AeVsPhiViewable()

	CalViewProductDescriptor AeVsPhiViewable::view(const string &fileName)
	const
	{
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double E = p->state()->E.get(),
			   theta = p->state()->theta.get();
		int	spectralOrder = p->state()->spectralOrder.get();
		bool isRgs = ins==Xmm::RGS1 || ins==Xmm::RGS2;

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Effective area for "
		   <<Xmm::toString(ins)<<" at E/theta = "
		   <<E<<"eV / "<<theta<<"''";
		if (isRgs)
		set<<" order="<<spectralOrder;
		set<<"\"\n";
		set<<"@xaxis label \"azimuth angle [deg]\"\n";
		set<<"@yaxis label \"effective area [cm^2]\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";

		//
		//	write data
		//
		set<<"@target s0"<<endl;
		set<<"@s0 color 1"<<endl;
		set<<"@type xy"<<endl;

		//	get the atom
		EffectiveAreaDataServer *aeds = calServer.getAtom(aeds);

		const double fromPhi = 0.,
					 toPhi = 360.,
					 inStepsOf = 5.;

		for (double phi = fromPhi; phi<=toPhi; phi += inStepsOf) {
			EffectiveAreaCurve f = !isRgs
				? aeds->effectiveAreaCurve(theta, phi * DEG_TO_RAD)
				: aeds->effectiveAreaCurve(spectralOrder, theta,
										  phi * DEG_TO_RAD);

			set<<phi<<" "<<f(E)<<endl;
		}
		set<<"&\n";
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // AeVsPhiViewable::view()


	~AeVsPhiViewable() { PROC ENDPROC }
};


EffectiveAreaPlot::EffectiveAreaPlot(const CalViewServer *server)
		: CalViewable(server)
{
	PROC

	_name = "Effective area";
	_whatIs = "Effective area";

	//	register SubViewables
	_subViewable.push_back(new AeVsEViewable(server));
	_subViewable.push_back(new AeVsThetaViewable(server));
	_subViewable.push_back(new AeVsPhiViewable(server));

	ENDPROC
} // EffectiveAreaPlot::EffectiveAreaPlot()


CalViewProductDescriptor EffectiveAreaPlot::view(const string &setName) const
{
	PROC
	RETURN _subViewable[0]->view(setName);
} // EffectiveAreaPlot::view()


bool EffectiveAreaPlot::isApplicable(Xmm::Instrument instrument) const
{
	PROC
	RETURN instrument != Xmm::OM && instrument != Xmm::ERM;
} // EffectiveAreaPlot::isApplicable()	


//
//	---( VignettingPlot + SubViewables )---------------------------------------
//
class VignettingVsEViewable : public CalViewable {
public:
	VignettingVsEViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "vs E";
		_whatIs = "vignetting factor as function of energy";

		ENDPROC
	} // VignettingVsEViewable()

	CalViewProductDescriptor VignettingVsEViewable::view(const string
														 &fileName)
	const
	{
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double theta = p->state()->theta.get(),
			   phi = p->state()->phi.get();
		int	spectralOrder = p->state()->spectralOrder.get();
		bool isRgs = ins==Xmm::RGS1 || ins==Xmm::RGS2;

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Vignetting factor for "
		   <<Xmm::toString(ins)<<" at theta/phi = "
		   <<theta<<"'' / "<<phi<<"deg";
		if (isRgs)
		set<<" order="<<spectralOrder;
		set<<"\"\n";
		set<<"@xaxis label \"energy [eV]\"\n";
		set<<"@yaxis label \"vignetting factor\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";

		//
		//	write data
		//
		set<<"@target s0"<<endl;
		set<<"@s0 color 1"<<endl;
		set<<"@type xy"<<endl;

		//	get the atom
		EffectiveAreaDataServer *aeds = calServer.getAtom(aeds);

		const double fromE = 100.,
					 toE = isRgs ? 2500. : 15000.,
					 inStepsOf = 1.;

		EffectiveAreaCurve f = !isRgs
			? aeds->vignettingCurve(theta, phi * DEG_TO_RAD)
			: aeds->vignettingCurve(spectralOrder, theta, phi * DEG_TO_RAD);

		for (double E = fromE; E<=toE; E+=inStepsOf)
			set<<E<<" "<<f(E)<<endl;
		set<<"&\n";
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // VignettingVsEViewable::view()


	~VignettingVsEViewable() { PROC ENDPROC }
};


class VignettingVsThetaViewable : public CalViewable {
public:
	VignettingVsThetaViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "vs theta";
		_whatIs = "vignetting factor as function of off-axis angle";

		ENDPROC
	} // VignettingVsThetaViewable()

	CalViewProductDescriptor VignettingVsThetaViewable::view(const string
															 &fileName)
	const
	{
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double E = p->state()->E.get(),
			   phi = p->state()->phi.get();
		int	spectralOrder = p->state()->spectralOrder.get();
		bool isRgs = ins==Xmm::RGS1 || ins==Xmm::RGS2;

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Vignetting factor for "
		   <<Xmm::toString(ins)<<" at E/phi = "
		   <<E<<"eV / "<<phi<<"deg";
		if (isRgs)
		set<<" order="<<spectralOrder;
		set<<"\"\n";
		set<<"@xaxis label \"off-axis angle ['']\"\n";
		set<<"@yaxis label \"vignetting factor\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";

		//
		//	write data
		//
		set<<"@target s0"<<endl;
		set<<"@s0 color 1"<<endl;
		set<<"@type xy"<<endl;

		//	get the atom
		EffectiveAreaDataServer *aeds = calServer.getAtom(aeds);

		const double fromTheta = 0.,
					 toTheta = 900.,
					 inStepsOf = 5.;

		for (double theta = fromTheta; theta <=toTheta; theta += inStepsOf) {
			EffectiveAreaCurve f = !isRgs
				? aeds->vignettingCurve(theta, phi * DEG_TO_RAD)
				: aeds->vignettingCurve(spectralOrder, theta, phi *DEG_TO_RAD);

			set<<theta<<" "<<f(E)<<endl;
		}
		set<<"&\n";
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // VignettingVsThetaViewable::view()


	~VignettingVsThetaViewable() { PROC ENDPROC }
};


class VignettingVsPhiViewable : public CalViewable {
public:
	VignettingVsPhiViewable(const CalViewServer *server = 0) : CalViewable(server) {
		PROC

		_name = "vs phi";
		_whatIs = "vignetting factor as function of azimuth angle";

		ENDPROC
	} // VignettingVsPhiViewable()

	CalViewProductDescriptor VignettingVsPhiViewable::view(const string
														   &fileName)
	const
	{
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double E = p->state()->E.get(),
			   theta = p->state()->theta.get();
		int	spectralOrder = p->state()->spectralOrder.get();
		bool isRgs = ins==Xmm::RGS1 || ins==Xmm::RGS2;

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Vignetting factor for "
		   <<Xmm::toString(ins)<<" at E/theta = "
		   <<E<<"eV / "<<theta<<"''";
		if (isRgs)
		set<<" order="<<spectralOrder;
		set<<"\"\n";
		set<<"@xaxis label \"azimuth angle [deg]\"\n";
		set<<"@yaxis label \"vignetting factor\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";

		//
		//	write data
		//
		set<<"@target s0"<<endl;
		set<<"@s0 color 1"<<endl;
		set<<"@type xy"<<endl;

		//	get the atom
		EffectiveAreaDataServer *aeds = calServer.getAtom(aeds);

		const double fromPhi = 0.,
					 toPhi = 360.,
					 inStepsOf = 5.;

		for (double phi = fromPhi; phi<=toPhi; phi += inStepsOf) {
			EffectiveAreaCurve f = !isRgs
				? aeds->vignettingCurve(theta, phi * DEG_TO_RAD)
				: aeds->vignettingCurve(spectralOrder, theta, phi *DEG_TO_RAD);

			set<<phi<<" "<<f(E)<<endl;
		}
		set<<"&\n";
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // VignettingVsPhiViewable::view()


	~VignettingVsPhiViewable() { PROC ENDPROC }
};


VignettingPlot::VignettingPlot(const CalViewServer *server)
		: CalViewable(server)
{
	PROC

	_name = "Vignetting";
	_whatIs = "Vignetting curve";

	//	register SubViewables
	_subViewable.push_back(new VignettingVsEViewable(server));
	_subViewable.push_back(new VignettingVsThetaViewable(server));
	_subViewable.push_back(new VignettingVsPhiViewable(server));

	ENDPROC
} // VignettingPlot::VignettingPlot()


CalViewProductDescriptor VignettingPlot::view(const string &setName) const
{
	PROC
	RETURN _subViewable[0]->view(setName);
} // VignettingPlot::view()


bool VignettingPlot::isApplicable(Xmm::Instrument instrument) const
{
	PROC
	RETURN instrument != Xmm::OM && instrument != Xmm::ERM;
} // VignettingPlot::isApplicable()	


//
//	---( FilterTransPlot )-----------------------------------------------------
//
FilterTransPlot::FilterTransPlot(const CalViewServer *server)
		: CalViewable(server)
{
	PROC

	_name = "FilterTrans";
	_whatIs = "Filter Transmission";

	ENDPROC
} // FilterTransPlot::FilterTransPlot()


CalViewProductDescriptor FilterTransPlot::view(const string &fileName) const
{
	PROC

	static string dftName(tmpnam(NULL));
	const string &setName = fileName != "" ? fileName : dftName;

	//	open Grace parameter file
	ofstream set(setName.c_str());
	Filter filter = Xmm::THIN1;
	if (calServer.state()->filter()->isActive())
		filter = calServer.state()->filter()->get();
	set<<"@title \"Transmission of EPIC filter "<<Xmm::toString(filter)
	   <<"\"\n";
	set<<"@xaxis label \"energy [eV]\"\n";
	set<<"@yaxis label \"transmission [unitless]\"\n";
	set<<"@string loctype world\n";
	set<<"@string just 2\n";

	//
	//	write data
	//
	set<<"@target s0"<<endl;
	set<<"@s0 color 1"<<endl;
	set<<"@type xy"<<endl;

	//	get the atom
	FilterDataServer *fds = calServer.getAtom(fds);

	const double fromE = 0.,
				 toE = 15000.,
				 inStepsOf = 1.;

	const FilterTransmissionCurve &f = fds->at(0., 0.);
	for (double E = fromE; E<=toE; E+=inStepsOf)
		set<<E<<" "<<f(E)<<endl;
	set<<"&\n";
	set.close();

	errstr<<"Grace parameter file written to "<<setName<<endl
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	RETURN CalViewProductDescriptor(setName);
} // FilterTransPlot::view()


bool FilterTransPlot::isApplicable(Xmm::Instrument instrument) const
{
	PROC
	RETURN (instrument == Xmm::EMOS1 || instrument == Xmm::EMOS2 ||
			instrument == Xmm::EPN);
} // FilterTransPlot::isApplicable()	


//
//	---( PsfPlot )-----------------------------------------------------
//
PsfPlot::PsfPlot(const CalViewServer *server) : CalViewable(server)
{
	PROC

	_name = "PSF";
	_whatIs = "Point-Spread-Function";

	ENDPROC
} // PsfPlot::PsfPlot()


CalViewProductDescriptor PsfPlot::view(const string &fileName) const
{
	PROC

	static string dftName(tmpnam(NULL));
	const string &setName = fileName != "" ? fileName : dftName;
	DataSet *set = dataSetServer->open(setName, Dal::Create);

	CalViewServer *p = const_cast<CalViewServer *>(parent());

	//	get PSF server atom - for OM we need to do something special
	if (calServer.instrument()->id() != Xmm::OM) {
		PsfDataServer *pds = calServer.getAtom(pds);
		const Psf &psf = pds->psf(p->state()->E.get(),
								  p->state()->theta.get(),
								  p->state()->phi.get() * DEG_TO_RAD);

		//	if (psf.binSizeChangeable())
		//		psf.binSize(Psf::BinSize(.02, .02));

		set<<psf.image(SpatialWindow<double>(4., 4.));
	} else {
		OmPsf *psf = calServer.getAtom(psf);
		psf->forCountFrameRateRatio(.5).withCenterAt(1024., 1024.);
		OmPsf::OmPsfWindow win = { 0, 0, 0, 0 };
		set<<psf->image(win, 64, 64);
	}

	dataSetServer->close(set);

	errstr<<"Image file written to "<<setName<<endl
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	RETURN CalViewProductDescriptor(setName,
									CalViewProductDescriptor::FITS_IMG);
} // PsfPlot::view()


bool PsfPlot::isApplicable(Xmm::Instrument instrument) const
{
	PROC
	RETURN (instrument == Xmm::EMOS1 || instrument == Xmm::EMOS2 ||
			instrument == Xmm::EPN || instrument == Xmm::OM);
} // PsfPlot::isApplicable()	


//
//	---( EncircledEnergyPlot + SubViewables )----------------------------------
//
class EncircledEnergyPsfViewable : public CalViewable {
public:
	EncircledEnergyPsfViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "Encircled energy (PSF)";
		_whatIs = "Encircled energy from PSF integration";

		ENDPROC
	} // EncircledEnergyPsfViewable()

	CalViewProductDescriptor view(const string &fileName) const {
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		//	get view state variables
		CalViewServer *p = const_cast<CalViewServer *>(parent());
		double E = p->state()->E.get(),
			   theta = p->state()->theta.get(),
			   phi = p->state()->phi.get();

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Encircled energy from PSF integration\"\n";
		set<<"@subtitle \""
		   <<Xmm::toString(calServer.instrument()->id())<<" - "
		   <<"theta="<<theta<<"'' "
		   <<"phi="<<phi<<"deg "
		   <<"E="<<E<<"eV\"\n";

		set<<"@xaxis label \"radius [mm]\"\n";
		set<<"@yaxis label \"PSF fraction [unitless]\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";

		//	get PSF server atom
		PsfDataServer *pds = calServer.getAtom(pds);
		const Psf &psf = pds->psf(E, theta, phi*DEG_TO_RAD);
		double fracs[] = { .05, .1, .2, .3, .4, .5, .55,
						   .6, .65, .7, .75, .8, .85, .9, .91, .92, .93, .94,
						   .95, .96, .97, .98 };
		unsigned N = NUM_OF(fracs);
		vector<double> rVec(N);
		for (unsigned n=0; n<N; ++n)
			rVec[n] = psf.encircledEnergy(fracs[n]);

		//	remove duplicates
		vector<double> rv, fv;
		rv.push_back(rVec[0]);
		fv.push_back(fracs[0]);
		double oldr = rVec[0];
		for (unsigned n=1; n<N; ++n) {
			SAS_DEBUG(cout<<n<<' '<<rVec[n]<<' '<<fracs[n]<<endl;)
			if (rVec[n] > oldr) {
				rv.push_back(rVec[n]);
				fv.push_back(fracs[n]);
				oldr = rVec[n];
			}
		}

		SAS_DEBUG(cout<<endl;
		for (unsigned n=0; n<rv.size(); ++n)
			cout<<n<<' '<<rv[n]<<' '<<fv[n]<<endl;)
		InterpolatedCurve rc = InterpolatedCurve(rv, fv);

		//
		//	write data
		//
		set<<"@target s0"<<endl;
		set<<"@s0 color 1"<<endl;
		set<<"@type xy"<<endl;

		const double fromR = 0.,
					 toR = 5.,
					 inStepsOf = .01;

		double oldv = 0.;
		for (double r = fromR+.05; r<=toR; r+=inStepsOf) {
			double v=rc(r);
			if (v<=oldv || v>=1.)
				break;
			set<<r<<" "<<v<<endl;
			oldv = v;
		}
		set<<"&\n";
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // EncircledEnergyPsfViewable::view()

	~EncircledEnergyPsfViewable() { PROC ENDPROC }
};


class EncircledEnergyCcfViewable : public CalViewable {
public:
	EncircledEnergyCcfViewable(const CalViewServer *server = 0)
		: CalViewable(server) {
		PROC

		_name = "Encircled energy (CCF)";
		_whatIs = "Encircled energy from CCF";

		ENDPROC
	} // EncircledEnergyCcfViewable()

	CalViewProductDescriptor view(const string &fileName) const {
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		double E = p->state()->E.get(),
			   theta = p->state()->theta.get(),
			   phi = p->state()->phi.get();

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Encircled energy from CCF\"\n";
		set<<"@subtitle \""
		   <<Xmm::toString(calServer.instrument()->id())<<" - "
		   <<"theta="<<theta<<"'' "
		   <<"phi="<<phi<<"deg "
		   <<"E="<<E<<"eV\"\n";

		set<<"@xaxis label \"radius [arcsec]\"\n";
		set<<"@yaxis label \"PSF fraction [unitless]\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";

		//	get PSF server atom
		EncircledEnergyDataServer *eeds = calServer.getAtom(eeds);
		const EncircledEnergyFunction &ee =
							eeds->encircledEnergy(E, theta, phi * DEG_TO_RAD);

		//
		//	write data
		//
		set<<"@target s0"<<endl;
		set<<"@s0 color 1"<<endl;
		set<<"@type xy"<<endl;

		const double fromR = 0.,
					 toR = 300.,
					 inStepsOf = .1;

		for (double r = fromR+.05; r<=toR; r+=inStepsOf)
			set<<r<<' '<<ee(r)<<endl;
		set<<"&\n";
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // EncircledEnergyCcfViewable::view()

	~EncircledEnergyCcfViewable() { PROC ENDPROC }
};


EncircledEnergyPlot::EncircledEnergyPlot(const CalViewServer *server)
		: CalViewable(server)
{
	PROC

	_name = "Encircled energy";
	_whatIs = "Encircled energy function";

	//	register SubViewables
	_subViewable.push_back(new EncircledEnergyPsfViewable(server));
	_subViewable.push_back(new EncircledEnergyCcfViewable(server));

	ENDPROC
} // EncircledEnergyPlot::EncircledEnergyPlot()


CalViewProductDescriptor EncircledEnergyPlot::view(const string &setName)
const
{
	PROC
	RETURN _subViewable[0]->view(setName);
} // EncircledEnergyPlot::view()


bool EncircledEnergyPlot::isApplicable(Xmm::Instrument instrument) const
{
	PROC
	RETURN (instrument == Xmm::EMOS1 || instrument == Xmm::EMOS2 ||
			instrument == Xmm::EPN);
} // EncircledEnergyPlot::isApplicable()	


//
//	---( QuantumEfficiencyPlot + SubViewables )--------------------------------
//
class QuantumEfficiencySubViewable : public CalViewable {
public:
	QuantumEfficiencySubViewable(unsigned patternId,
								 const CalViewServer *server = 0)
		: CalViewable(server), _patternId(patternId) {
		PROC

		_name = toString(patternId);
		strstream s;
		s<<"Quantum Efficiency for pattern "<<_name<<ends;
		char *p = s.str();
		_whatIs = p;
		delete p;

		ENDPROC
	} // QuantumEfficiencySubViewable()


	CalViewProductDescriptor view(const string &fileName = "") const {
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		//	open Grace parameter file
		ofstream set(setName.c_str());
		Ccd ccd = 1;
		if (calServer.state()->ccd()->isActive())
			ccd = calServer.state()->ccd()->get();

		const XmmInstrument *ins = calServer.instrument();		
		unsigned rawX = ins->maxX()/2,
				 rawY = ins->maxY()/2;
		set<<"@title \"Quantum Efficiency - CCD #"<<ccd<<" - pixel "
		   <<rawX<<'/'<<rawY<<" - pattern #"<<_patternId<<"\"\n";
		set<<"@xaxis label \"energy [eV]\"\n";
		set<<"@yaxis label \"QE [unitless]\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";

		//
		//	write data
		//
		set<<"@target s0"<<endl;
		set<<"@s0 color 1"<<endl;
		set<<"@type xy"<<endl;

		//	get the atom
		CcdQuantumDataServer *qds = calServer.getAtom(qds);

		const double fromE = 0.,
					 toE = 15000.,
					 inStepsOf = 1.;

		const CcdQuantumEfficiencyCurve &f =
									qds->at(rawX, rawY).pattern(_patternId);
		for (double E = fromE; E<=toE; E+=inStepsOf)
			set<<E<<" "<<f(E)<<endl;
		set<<"&\n";
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // QuantumEfficiencySubViewable::view()

	~QuantumEfficiencySubViewable() { PROC ENDPROC }

private:
	unsigned _patternId;
};


QuantumEfficiencyPlot::QuantumEfficiencyPlot(const CalViewServer *server)
		: CalViewable(server)
{
	PROC

	_name = "QuantumEfficiency";
	_whatIs = "Quantum Efficiency";

	//	get the atom and register SubViewables
	CcdQuantumDataServer *qds = calServer.getAtom(qds);
	unsigned patterns = qds->patterns();
	for (unsigned n=0; n<patterns; ++n)
		_subViewable.push_back(new QuantumEfficiencySubViewable(n, server));

	ENDPROC
} // QuantumEfficiencyPlot::QuantumEfficiencyPlot()


CalViewProductDescriptor QuantumEfficiencyPlot::view(const string &setName)
const
{
	PROC
	RETURN _subViewable[0]->view(setName);
} // QuantumEfficiencyPlot::view()


bool QuantumEfficiencyPlot::isApplicable(Xmm::Instrument instrument) const
{
	PROC
	RETURN (instrument == Xmm::EMOS1 || instrument == Xmm::EMOS2 ||
			instrument == Xmm::EPN ||
			instrument == Xmm::RGS1 || instrument == Xmm::RGS2);
} // QuantumEfficiencyPlot::isApplicable()	


//
//	---( CrossPsfPlot + SubViewables )-----------------------------------------
//
class CrossPsfVsChiSubViewable : public CalViewable {
public:
	CrossPsfVsChiSubViewable(const CalViewServer *server = 0)
		: CalViewable(server)
	{
		PROC

		_name = "CrossPsf vs chi";
		_whatIs = "Point-Spread-Function in cross dispersion direction";

		ENDPROC
	} // CrossPsfVsChiSubViewable::CrossPsfSubViewable()


	CalViewProductDescriptor view(const string &fileName)const
	{
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double theta = p->state()->theta.get(),
			   phi = p->state()->phi.get();

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Cross dispersion Point-Spread-Function\"\n";
		set<<"@subtitle \""
		   <<Xmm::toString(ins)<<" - "
		   <<"theta="<<theta<<"'' "
		   <<"phi="<<phi<<"deg\"\n";
		set<<"@xaxis label \"cross dispersion angle [deg]\"\n";
		set<<"@yaxis label \"probability\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";
		set<<"@legend on\n";

	    CanonicalCrossPsf *cpsf = calServer.getAtom(cpsf);
		GratingDataServer *gds = calServer.getAtom(gds);

		//	construct the spectrum
		const double betaM = .056335894,
					 deltaBeta = .01;
		double betaVec[] = { betaM, betaM+deltaBeta, betaM+2.*deltaBeta,
									betaM-deltaBeta, betaM-2.*deltaBeta };
		double xAnglesMin = -.05 * DEG_TO_RAD;
		unsigned nBins = 400;
		double binWidth = (-xAnglesMin - xAnglesMin)/DBL(nBins);
		CalReal64Vector xAngles(0, nBins);
		for (unsigned n=0; n<nBins; ++n)
			xAngles[n] = xAnglesMin + (DBL(n)-.5) * binWidth;

		double sourceXdisp = gds->chi(theta, phi * DEG_TO_RAD);
		for (unsigned n=0; n<NUM_OF(betaVec); ++n) {
			set<<"@target s"<<n<<endl;
			set<<"@type xy"<<endl;
			set<<"@s"<<n<<" legend \"beta: "<<setprecision(5)<<betaVec[n];
			if (!n)
				set<<" (M-point)";
			set<<"\"\n";

			Spectrum spec(cpsf->probability(0., betaVec[n], sourceXdisp,
						  xAngles, binWidth));

			unsigned N = spec.size();
			for (unsigned n=0; n<N; ++n)
				set<<setprecision(6)
				   <<xAngles[n]*RAD_TO_DEG<<' '<<setprecision(10)<<spec[n]
				   <<endl;
			set<<"&\n";
		}
		set.close();
		xAngles.free();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // CrossPsfVsChiSubViewable::view()
};


class CrossPsfVsBetaSubViewable : public CalViewable {
public:
	CrossPsfVsBetaSubViewable(const CalViewServer *server = 0)
		: CalViewable(server)
	{
		PROC

		_name = "CrossPsf vs beta";
		_whatIs = "Point-Spread-Function in cross dispersion direction";

		ENDPROC
	} // CrossPsfVsBetaSubViewable::CrossPsfSubViewable()


	CalViewProductDescriptor view(const string &fileName)const
	{
		PROC

		static string dftName(tmpnam(NULL));
		const string &setName = fileName != "" ? fileName : dftName;

		CalViewServer *p = const_cast<CalViewServer *>(parent());
		Xmm::Instrument ins = calServer.instrument()->id();
		double theta = p->state()->theta.get(),
			   phi = p->state()->phi.get();

		//	open Grace parameter file
		ofstream set(setName.c_str());
		set<<"@title \"Cross dispersion Point-Spread-Function\"\n";
		set<<"@subtitle \""
		   <<Xmm::toString(ins)<<" - "
		   <<"theta="<<theta<<"'' "
		   <<"phi="<<phi<<"deg - 95% level\"\n";
		set<<"@xaxis label \"beta [rad]\"\n";
		set<<"@yaxis label \"chi [rad]\"\n";
		set<<"@string loctype world\n";
		set<<"@string just 2\n";
		set<<"@legend on\n";

		//	get the atom
	    CanonicalCrossPsf *cpsf = calServer.getAtom(cpsf);
		GratingDataServer *gds = calServer.getAtom(gds);

		const double fromBeta = 2.1*DEG_TO_RAD,
					 toBeta = 5.0*DEG_TO_RAD,
					 inStepsOf = .002*DEG_TO_RAD;
		unsigned N = UNT((toBeta - fromBeta)/inStepsOf+1);
		vector<double> beta(N);
		for (unsigned n=0; n<N; ++n)
			beta[n] = fromBeta + n*inStepsOf;

		double xAnglesMin = -.035 * DEG_TO_RAD;
		unsigned nBins = 3000;
		double binWidth = (-xAnglesMin - xAnglesMin)/DBL(nBins);
		CalReal64Vector xAngles(0, nBins);
		for (unsigned n=0; n<nBins; ++n)
			xAngles[n] = xAnglesMin + (DBL(n)-.5) * binWidth;

		double sourceXdisp = gds->chi(theta, phi * DEG_TO_RAD);
		vector<double> xdsp95[2];
		xdsp95[0].resize(N);
		xdsp95[1].resize(N);
		for (unsigned n=0; n<N; ++n) {
			Spectrum spec(cpsf->probability(0., beta[n], sourceXdisp,
						  xAngles, binWidth));
			unsigned peakChannel = spec.peakChannel();
			int i = peakChannel;
			double sum = 0.;
			while(i>=0 && sum <.5*.95)
				sum += spec[i--];
			xdsp95[0][n] = i>=0 ? xAngles[i] : 0.;
			i = peakChannel;
			int maxN = spec.size();
			sum = 0.;
			while(i<maxN && sum <.5*.95)
				sum += spec[i++];
			xdsp95[1][n] = i<maxN ? xAngles[i] : 0.;
		}

		for (unsigned n=0; n<2; ++n) {
			set<<"@target s"<<n<<endl;
			set<<"@type xy"<<endl;

			for (unsigned nn=0; nn<N; ++nn)
				set<<setprecision(6)<<beta[nn]<<' '<<xdsp95[n][nn]<<endl;
			set<<"&\n";
		}
		set.close();
		xAngles.free();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);

		RETURN CalViewProductDescriptor(setName);
	} // CrossPsfVsBetaSubViewable::view()
};


CrossPsfPlot::CrossPsfPlot(const CalViewServer *server)
	: CalViewable(server)
{
	PROC

	_name = "CrossPsf";
	_whatIs = "Point-Spread-Function in cross-dispersion direction";

	_subViewable.push_back(new CrossPsfVsChiSubViewable(server));
	_subViewable.push_back(new CrossPsfVsBetaSubViewable(server));

	ENDPROC
} // CrossPsfPlt::CrossPsfPlot()


CalViewProductDescriptor CrossPsfPlot::view(const string &setName) const
{
	PROC
	RETURN _subViewable[0]->view(setName);
} // CrossPsfPlot::view()


bool CrossPsfPlot::isApplicable(Xmm::Instrument instrument) const
{
	PROC
	RETURN (instrument == Xmm::RGS1 || instrument == Xmm::RGS2);
} // CrossPsfPlot::isApplicable()	


//
//	---( DispersedPsfPlot )----------------------------------------------------
//
DispersedPsfPlot::DispersedPsfPlot(const CalViewServer *server)
		: CalViewable(server)
{
	PROC

	_name = "DispersedPsf";
	_whatIs = "Point-Spread-Function in dispersion direction";

	ENDPROC
} // DispersedPsfPlot::DispersedPsfPlot()


CalViewProductDescriptor DispersedPsfPlot::view(const string &fileName) const
{
	PROC

	//	get the atom
	PsfDataServer *pds = calServer.getAtom(pds);
	GratingDataServer *gds = calServer.getAtom(gds);

	static string dftName(tmpnam(NULL));
	const string &setName = fileName != "" ? fileName : dftName;

	CalViewServer *p = const_cast<CalViewServer *>(parent());
	Xmm::Instrument ins = calServer.instrument()->id();
	double theta = p->state()->theta.get(),
		   phi = p->state()->phi.get(),
		   alphaIn = gds->alpha(theta, phi*DEG_TO_RAD) * RAD_TO_DEG,
		   E = p->state()->E.get();

	//	open Grace parameter file
	ofstream set(setName.c_str());
	set<<"@title \"PSF in dispersion direction\"\n";
	set<<"@subtitle \""
	   <<Xmm::toString(ins)<<" - "
	   <<"theta="<<theta<<"'' "
	   <<"phi="<<phi<<"deg "
	   <<"alpha="<<setprecision(5)<<alphaIn<<"deg "
	   <<"E="<<E<<"eV\"\n";

	set<<"@xaxis label \"beta [rad]\"\n";
	set<<"@yaxis label \"probability\"\n";
	set<<"@string loctype world\n";
	set<<"@string just 2\n";
	set<<"@legend on\n";

	for (int order=gds->maxOrder; order>=gds->minOrder; --order) {
		int setnr = gds->maxOrder-order;
		set<<"@target s"<<setnr<<endl;
		set<<"@type xy"<<endl;
		set<<"@s"<<setnr<<" legend \"order: "<<order<<"\"\n";

		double alphaIn = gds->alpha(theta, phi * DEG_TO_RAD),
			   betaOut = gds->beta(alphaIn, order, E),
			   chromaticMagnification =	sin(alphaIn)/sin(betaOut);
		const double NBETA = 10000.,
					 DELTA_BETA = .5,	//	beta span
					 binWidth = DELTA_BETA/NBETA;
		Spectrum spec(pds->dispersionFigureDistribution(7.,
												    binWidth*DEG_TO_RAD,
												    chromaticMagnification,
													.65));
		unsigned N = spec.size();
		for (unsigned n=0; n<N; ++n)
			set<<setprecision(6)
			   <<(DBL(n)-.5*DBL(N))*binWidth*DEG_TO_RAD+betaOut<<' '
			   <<setprecision(10)<<spec[n]<<endl;
		set<<"&\n";
	}
	set.close();

	errstr<<"Grace parameter file written to "<<setName<<endl
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	RETURN CalViewProductDescriptor(setName);
} // DispersedPsfPlot::view()


bool DispersedPsfPlot::isApplicable(Xmm::Instrument instrument) const
{
	PROC
	RETURN (instrument == Xmm::RGS1 || instrument == Xmm::RGS2);
} // DispersedPsfPlot::isApplicable()	


//
//	---( CcdPlot )--------------------------------------------------------
//
CcdPlot::CcdPlot(const CalViewServer *server)
		: CalViewable(server)
{
	PROC

	_name = "LinCoord";
	_whatIs = "A plot showing the positions of the CCDs in the camera";

	ENDPROC
} // CcdPlot::CcdPlot()



CalViewProductDescriptor CcdPlot::view(const string &fileName) const
{
	PROC

	static string dftName(tmpnam(NULL));
	const string &setName = fileName != "" ? fileName : dftName;

	//	open Grace parameter file
	ofstream set(setName.c_str());
	set<<"@title \"Positions of CCDs in "
	   <<Xmm::toString(calServer.instrument()->id())<<"\"\n";
	set<<"@xaxis label \"x [mm]\"\n";
	set<<"@yaxis label \"y [mm]\"\n";
	set<<"@string loctype world\n";
	set<<"@string just 2\n";

	calServer.pushState();
	const XmmInstrument *ins = calServer.instrument();
	calServer.state()->randomize()->set(false);
	calServer.state()->ocb()->set(1);
	unsigned minX = ins->minX(), maxX = ins->maxX(),
			 minY = ins->minY(), maxY = ins->maxY(),
			 xPixels = maxX - minX + 1,
			 yPixels = maxY - minY + 1,
			 N = xPixels * yPixels;

	unsigned everyNthPixel = N > 15000 ? 25 : 10;
	int16 chipXv[N/everyNthPixel+everyNthPixel],
		  chipYv[N/everyNthPixel+everyNthPixel];
	unsigned n = 0;
	for (unsigned x=minX; x<=maxX; x+=everyNthPixel)
		for (unsigned y=minY; y<maxY; y+=everyNthPixel) {
			chipXv[n] = x;
			chipYv[n++] = y;
		}
	N = n;
	CalInt16Vector chipX(chipXv, N),
				   chipY(chipYv, N);

	for (unsigned ccd = 1; ccd<=ins->nCcds(); ++ccd) {
		 calServer.state()->ccd()->set(ccd);
		short rawX[] = { minX, maxX, maxX, minX, minX },
			  rawY[] = { minY, minY, maxY, maxY, minY };

		CalInt16Vector rawXVec(rawX, NUM_OF(rawX)),
					   rawYVec(rawY, NUM_OF(rawY));
		float xmm[NUM_OF(rawX)], ymm[NUM_OF(rawX)], zmm[NUM_OF(rawX)];
		GeometryDataServer *cdf = calServer.getAtom(cdf);
		XmmPixelList pl = cdf->chipCoordList(rawXVec, rawYVec);
		XmmCartesianPointList cpl = pl.toCamCoord1(xmm, ymm, zmm);
		cpl.toCamCoord2();

		float cenx = .25 * (xmm[0]+xmm[1]+xmm[2]+xmm[3]),
			  ceny = .25 * (ymm[0]+ymm[1]+ymm[2]+ymm[3]);
		//
		//	write CCD number
		//
		set<<"@with string\n";
		set<<"@string on\n";
		set<<"@string char size 2\n";
		set<<"@string "<<cenx<<", "<<ceny<<endl;
		set<<"@string def \""<<ccd<<"\"\n";

		//
		//	coordinate systems: x
		//
		set<<"@with line\n";
		set<<"@line on\n";
		set<<"@line loctype world\n";
		set<<"@line "<<xmm[0]<<", "<<ymm[0]<<", "
					 <<xmm[1]<<", "<<ymm[1]<<endl;
		set<<"@line arrow 2\n";
		set<<"@line arrow size 1.5\n";
		set<<"@line def\n";
		set<<"@with string\n";
		set<<"@string on\n";
		set<<"@string char size 1\n";
		set<<"@string "<<xmm[1]<<", "<<ymm[1]<<endl;
		set<<"@string def \"X\"\n";

		//
		//	coordinate systems: y
		//
		set<<"@with line\n";
		set<<"@line on\n";
		set<<"@line loctype world\n";
		set<<"@line "<<xmm[0]<<", "<<ymm[0]<<", "
					 <<xmm[3]<<", "<<ymm[3]<<endl;
		set<<"@line arrow 2\n";
		set<<"@line arrow size 1.5\n";
		set<<"@line def\n";
		set<<"@with string\n";
		set<<"@string on\n";
		set<<"@string char size 1\n";
		set<<"@string "<<xmm[3]<<", "<<ymm[3]<<endl;
		set<<"@string def \"Y\"\n";

		//
		//	write data
		//
		set<<"@target s"<<2*(ccd-1)<<endl;
		set<<"@s"<<2*(ccd-1)<<" color 1\n";
		set<<"@type xy\n";

		for (unsigned n=0; n<NUM_OF(rawX); ++n)
			set<<xmm[n]<<" "<<ymm[n]<<endl;

		set<<"&\n";

		//
		//	pixels in the FOV
		//
		XmmPixelList allpixels = cdf->chipCoordList(chipX, chipY);
		vector<bool> inFOV = allpixels.insideFOV();
		XmmCartesianPointList allmm = allpixels.toCamCoord1().toCamCoord2();

		unsigned i = 2*(ccd-1)+1;
		set<<"@target s"<<i<<endl;
		set<<"@s"<<i<<" color 1\n";
		set<<"@type xy\n";
		set<<"@s"<<i<<" symbol "<<8<<endl;
		set<<"@s"<<i<<" symbol size .2\n";
		set<<"@s"<<i<<" symbol color "<<1<<endl;
		set<<"@s"<<i<<" symbol pattern 1\n";
		set<<"@s"<<i<<" symbol fill pattern 0\n";
		set<<"@s"<<i<<" line type 0\n";

		unsigned N = inFOV.size();
		CalReal32Vector &xv = allmm.x(),
						&yv = allmm.y(),
						&zv = allmm.z();
		for (unsigned n=0; n<N; n++)
			if (inFOV[n])
				set<<xv[n]<<" "<<yv[n]<<endl;

		set<<"&\n";

		xv.free();
		yv.free();
		zv.free();
	}
	calServer.popState();
	set.close();

	errstr<<"Grace parameter file written to "<<setName<<endl
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	RETURN CalViewProductDescriptor(setName);
} // CcdPlot::view()


//
//	---( HKParmIntPlot )------------------------------------------------------
//
HKParmIntPlot::HKParmIntPlot(const CalViewServer *server)
		: CalViewable(server)
{
	PROC

	_name = "HKParmInt";
	_whatIs = "HK parameter validity ranges";

	ENDPROC
} // HKParmIntPlot::HKParmIntPlot()



CalViewProductDescriptor HKParmIntPlot::view(const string &fileName) const
{
	PROC

	static string dftName(tmpnam(NULL));
	const string &setName = fileName != "" ? fileName : dftName;

	//	open file
	ofstream set(setName.c_str());

	//	get the atom
	HKvalRangesServer *p = calServer.getAtom(p);

	set<<"HK parameter validity intervals:\n\n";
	for (HKvalRanges::MCIter i = p->begin(); i != p->end(); ++i)
		set<<(*i).first<<" : "<<(*i).second<<endl;

	set<<"\nlong->short HK name translation table:\n\n";

	const HKvalRanges::MapType &m = p->translationMap();
	for (HKvalRanges::MCIter i = m.begin(); i != m.end(); ++i)
		set<<(*i).first<<" => "<<(*i).second<<endl;

	set.close();

	errstr<<"ASCII file written to "<<setName<<endl
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	RETURN CalViewProductDescriptor(setName, CalViewProductDescriptor::ASCII);
} // HKParmIntPlot::view()


//
//	---( RedistributionPlot )--------------------------------------------------
//
RedistributionPlot::RedistributionPlot(const CalViewServer *server)
		: CalViewable(server)
{
	PROC

	_name = "Redist";
	_whatIs = "CCD response";

	ENDPROC
} // RedistributionPlot::RedistributionPlot()


bool RedistributionPlot::isApplicable(Xmm::Instrument instrument) const
{
	PROC

	RETURN (instrument == Xmm::EMOS1 ||	instrument == Xmm::EMOS2 ||
			instrument == Xmm::EPN ||
			instrument == Xmm::RGS1 || instrument == Xmm::RGS2);
} // RedistributionPlot::isApplicable()	


CalViewProductDescriptor RedistributionPlot::view(const string &fileName) const
{
	PROC

	static string dftName(tmpnam(NULL));
	const string &setName = fileName != "" ? fileName : dftName;

	//	open Grace parameter file
	ofstream set(setName.c_str());
	set<<"@title \"CCD response for "
	   <<Xmm::toString(calServer.instrument()->id())<<"\"\n";
	set<<"@xaxis label \"energy [eV]\"\n";
	set<<"@yaxis label \"W\"\n";
	set<<"@string loctype world\n";
	set<<"@string just 2\n";

	//
	//	write data
	//
	set<<"@target s0"<<endl;
	set<<"@s0 color 1"<<endl;
	set<<"@type xy"<<endl;

	//	get the rmf
    CanonicalRedist *rmf = calServer.getAtom(rmf);
	CalViewServer *p = const_cast<CalViewServer *>(parent());
	double E = p->state()->E.get();
	const BinnedAxis &eBins = rmf->binnedPIeBounds();
	Spectrum spec(rmf->failSafe(true).spectrum(E, eBins));

	if (spec.size()) {
		unsigned N = eBins.bins();
		for (unsigned n=0; n<N; ++n)
			set<<eBins.bin(n).mid()<<" "<<spec[n]<<endl;

		set<<"&\n";
		set.close();

		errstr<<"Grace parameter file written to "<<setName<<endl
			  <<Sas::message(Msg::Lib, Msg::Verbose);
	} else
		errstr<<"Spectrum is empty"<<Sas::error("NullSpectrum");

	RETURN CalViewProductDescriptor(setName);
} // RedistributionPlot::view()


//
//	---( BadPixPlot )--------------------------------------------------------
//
BadPixPlot::BadPixPlot(const CalViewServer *server)
		: CalViewable(server)
{
	PROC

	_name = "BadPix";
	_whatIs = "A plot the location of bad pixels on the CCD";

	ENDPROC
} // BadPixPlot::BadPixPlot()


CalViewProductDescriptor BadPixPlot::view(const string &fileName) const
{
	PROC

	static string dftName(tmpnam(NULL));
	const string &setName = fileName != "" ? fileName : dftName;

	//	open Grace parameter file
	ofstream set(setName.c_str());
	set<<"@title \"Bad pixels on "
	   <<Xmm::toString(calServer.instrument()->id())<<" - CCD #"
	   <<calServer.state()->ccd()->get();
	if (calServer.state()->node()->isActive())
		set<<" - node #"<<Xmm::toString(calServer.state()->node()->get());
	set<<"\"\n";
	set<<"@xaxis label \"x [pixel]\"\n";
	set<<"@yaxis label \"y [pixel]\"\n";
	set<<"@string loctype world\n";
	set<<"@string just 2\n";

	const XmmInstrument *ins = calServer.instrument();
	set<<"@with g0"<<endl;
	set<<"@g0 on"<<endl;
	set<<"@world xmin "<<ins->minX()<<endl;
	set<<"@world xmax "<<ins->maxX()<<endl;
	set<<"@world ymin "<<ins->minY()<<endl;
	set<<"@world ymax "<<ins->maxY()<<endl;

	set<<"@s0 symbol 10"<<endl;
	set<<"@s0 symbol size 0.4"<<endl;
	set<<"@s0 symbol fill 0"<<endl;
	set<<"@s0 linestyle 0"<<endl;
	set<<"@xaxis tick on"<<endl;
	set<<"@xaxis tick major 100"<<endl;
	set<<"@xaxis tick minor 50"<<endl;
	set<<"@yaxis tick on"<<endl;
	set<<"@yaxis tick major 100"<<endl;
	set<<"@yaxis tick minor 50"<<endl;

	//
	//	write data
	//
	set<<"@target s0"<<endl;
	set<<"@type xy"<<endl;

    BadPixelDataServer *bpds = calServer.getAtom(bpds);
	BadPixelMap map = bpds->map();
	for (BadPixelMap::const_iterator i=map.begin();	i!=map.end(); ++i) {
		unsigned x = (*i).first.first,
		y = (*i).first.second;
		for (unsigned n=0; n<(*i).second.yextent(); ++n)
			set<<x<<" "<<y+n<<endl;
	}
	set<<"&\n";

	set.close();

	errstr<<"Grace parameter file written to "<<setName<<endl
		  <<Sas::message(Msg::Lib, Msg::Verbose);

	RETURN CalViewProductDescriptor(setName);
} // BadPixPlot::view()
