//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//						(c) 1997 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalviewBack.cc --- backend for calview
//
//	DESCRIPTION:
//		This contains the backend for the calview task
//
//	CHANGE HISTORY:
//		UL, 1998-05-08: 1.0.0: - first version
//		UL, 1999-01-19: 1.1.0: - make it work with changes in ../src
//		UL, 1999-03-10: 1.1.1: - minor change in openCcf()
//
//	AUTHOR:
//		U. Lammers, Fri May  8 12:45:29 MET DST 1998
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: TBW note about calview
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/calview/CalviewBackend.cc,v 1.9 2001/02/16 17:22:37 ulammers Exp $
//

#include "common.h"
#include "../src/CalServer.h"
#include "../src/Ccf.h"
#include "STime.h"
#include "errstr.h"
#include "Tracer.h"

#include "CalviewBackend.h"

//
//	----------------------------( openCcf )-----------------------------------
//
void openCcf(const char *cifName)
{
	PROC

	if(calServer.ccf())
		errstr<<"Closing previously opened CIF and opening new one `"
			  <<cifName<<"'\n"
			  <<Sas::warning("calview::openCCF");

	calServer.newCcf(cifName);

	SAS_DEBUG(clog<<"Read new CIF:\n"<<calServer.ccf()<<endl;)

	ENDPROC
} // openCcf()


//
//	----------------------( setCalibrationFile )-----------------------------
//
void setCalibrationFile(const char *calibrationFileName)
{
	PROC

	const CifEntry e = CifEntry(calibrationFileName);
    calServer.ccf()->registerNewCifEntry(e);
    SAS_DEBUG(clog<<"CIF after replacing constituent:\n"<<calServer.ccf()
              <<endl;)
	
	ENDPROC
} // setCalibrationFile()


//
//	---------------------( setCalState )------------------------------------
//
void setCalState(const CalState *calState)
{
	PROC

	calServer.setInstrument(
		S_CAST(UNT(calState->instrument), Xmm::Instruments));

	calServer.setCcd(calState->ccdChip);
	calServer.setNode(calState->ccdNode);
	calServer.setMode(
		S_CAST(UNT(calState->dataMode), Xmm::DataModes));
	calServer.setFilter(
		S_CAST(UNT(calState->filter), Xmm::Filters));
	calServer.setTemperature(calState->temperature);
	calServer.setObservationStartTime(calState->observationStartTime);
	calServer.setAccuracyLevel(
		S_CAST(UNT(calState->accuracyLevel), AccuracyLevel));

	calServer.setRandomizeSwitch(calState->randomizeSwitch);

	ENDPROC
} // setCalState()


//
//	---------------------( setCalStateFromScienceFile )-----------------------
//
void setCalStateFromScienceFile(const char *fileName)
{
	PROC

	calServer.setState(fileName);

	ENDPROC
} // setCalStateFromScienceFile()


//
//	---------------------( getCalState )--------------------------------------
//
const CalState *getCalState(void)
{
	PROC

	static CalState calState;

	calState.instrument = S_CAST(UNT(calServer.instrument()), XmmInstruments);
	calState.ccdChip = calServer.ccd();
	calState.ccdNode = calServer.node();
	calState.dataMode =	S_CAST(UNT(calServer.mode()), XmmDataModes);
	calState.filter = S_CAST(UNT(calServer.filter()), XmmFilters);
	calState.temperature = calServer.temperature();
	STime t = calServer.observationStartTime();
	calState.observationStartTime = t.modifiedJulianDay();
	calState.accuracyLevel = S_CAST(UNT(calServer.accuracyLevel()),
									AccuracyLevels);
	calState.randomizeSwitch = calServer.randomizeSwitch();

	RETURN &calState;
} // getCalState()


static ViewState currentViewState = { 0., 0., 0., "" };

//
//	------------------------( setViewState )--------------------------------
//
void setViewState(const ViewState *viewState)
{
	PROC

	currentViewState = *viewState;

	ENDPROC
} // setViewState()


//
//	----------------------( getViewState )---------------------------------
//
const ViewState	*getViewState(void)
{
	PROC
	RETURN &currentViewState;
} // getViewState()


static void setViewParameters(CalAtom *atom, const ViewState *viewState)
{
	PROC

	atom->setE(viewState->E);
	atom->setThetaPhi(viewState->theta, viewState->phi);
	//	atom->setFileName(string(viewState->dumpFileName));

	ENDPROC
} // setViewParameters()


//
//	---------------------( view )--------------------------------------------
//
void view(CalAtoms calAtom)
{
	PROC

	CalAtom *atom =	calServer.getAtom(S_CAST(UNT(calAtom), Ccf::CalDataTypes));
	setViewParameters(atom, &currentViewState);
	atom->view();

	ENDPROC
} // view()


//
//	-----------------------( dump )------------------------------------------
//
void dump(CalAtoms calAtom)
{
	PROC

	CalAtom *atom =	calServer.getAtom(S_CAST(UNT(calAtom), Ccf::CalDataTypes));
	setViewParameters(atom, &currentViewState);
	atom->dump();

	ENDPROC
} // dump()





