//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	AbsorptionDataServer.cc --- implementation of AbsorptionDataServer class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-09-10: 1.0.0: - first incarnation (from CalAtomsXmm.cc)
//		UL, 2000-01-31: 1.1.0: - adapted to new CCF::AbsCoef structure
//
//	AUTHOR:
//		U. Lammers, Fri Sep 10 13:00:46 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.
//		[2]: Calibration Access and Data Handbook, C. Erd et al.
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/AbsorptionDataServer.cc,v 1.6 2000/12/20 16:04:49 ulammers Exp $
//

#include "CalOalUtility.h"
#include "slatec++.h"
#include "AbsorptionDataServer.h"

using namespace CalOalUtility;

//
//	---( AbsCoefCurves )------------------------------------------------------
//
const AbsCoefCurve &AbsCoefCurves::coefficient(unsigned n = 0) const
{
	PROC

	if (n >= this->size())
		errstr<<"Cannot access absorption coefficient curve #"<<n
			  <<" for material "<<_name<<" - there are only "<<this->size()
			  <<Sas::error("NoSuchData");

	RETURN *(*this)[n];
} // AbsCoefCurves::coefficient()


//
//	---( AbsorptionDataServer )------------------------------------------------
//
CalAtomCollection<AbsorptionDataServer>::CalAtomCollection(
					const XmmInstrument *instrument,
					const Ccf *ccf,
					Xmm::Filter filter,
					AccuracyLevel accuracy)
{
	PROC

	_filter		= filter;
	_accuracy	= accuracy;
	_collection	= new vector<AbsorptionDataServer *>;
	_collection->push_back(new AbsorptionDataServer);
	AbsorptionDataServer &acds = *(*_collection)[0];

	CifEntry *e = ccf->entry(instrument, Ccf::AbsCoefs);
	DataSet *dataSet = e->open();

	//
	//	access table with atomic properties
	//
	CcfConstituentTable ccftab(dataSet, "ATOMIC_PROPERTIES");
	const Table *apTable = ccftab.block();
	unsigned nMat = apTable->rows();
	ColumnData *cdmat = apTable->column("Material")->data(),
			   *cdz = apTable->column("Z")->data(),
			   *cda = apTable->column("A")->data(),
			   *cdRho = apTable->column("Rho")->data();
	real32	   *Z = cdz->real32Data(),
			   *A = cda->real32Data(),
			   *rho = cdRho->real32Data();

	//
	//	loop over all HENKE data tables
	//
	for (unsigned n=0; n<dataSet->blocks(); ++n) {
		CcfConstituentTable ccftab(dataSet, n);
		const Table *table = ccftab.block();
		string tabName = table->name();
		if (tabName.substr(0, 6) != "HENKE-")
			continue;

		string element(tabName.substr(6, tabName.length()-6));
		ColumnData *cdE = table->column("ENERGY")->data();
		const Column *cdcf = table->column("SCATTERING_FACTORS");
		ColumnData *cdf = cdcf->data();
		real32	*E = cdE->real32Data(),
				*f = cdf->real32Data();

		if (cdf->elements() != 2)
			errstr<<"Expected column `"<<cdcf->qualifiedName()<<"' to be of "
					"length 2"<<Sas::error("UnexpectedColumnLength");

		//	read energy column
		unsigned nRows = table->rows();
		vector<double> energy(nRows),
					   f1(nRows), f2(nRows);

		for (unsigned i=0; i<nRows; ++i) {
			energy[i] = DBL(E[i]);
			f1[i] = DBL(f[cdf->position(i, 0)]);
			f2[i] = DBL(f[cdf->position(i, 1)]);
		}

		delete cdf;
		delete cdE;

		AtomicProperties ap(element, 1, 0., 0.);
		unsigned i = 0;
		for (; i<nMat; ++i)
			if (stringColumnRow(cdmat, i) == element) {
				ap = AtomicProperties(element, UNT(Z[i]), A[i], rho[i]);
				break;
			}

		if (i>=nMat)
			errstr<<"Did not find atomic properties for `"<<element<<"' in "
				  <<apTable->qualifiedName()
				  <<Sas::warning("DataNotFound");

		AbsCoefCurves curves(element);
		curves.push_back(
				new Slatec::CubicHermiteInterpolant<double>(energy, f1));
		curves.push_back(
				new Slatec::CubicHermiteInterpolant<double>(energy, f2));

		acds[element] = AbsorptionData(curves, ap);
	}

	delete cdmat;
	delete cdz;
	delete cda;
	delete cdRho;

	dataSetServer->close(dataSet);

	ENDPROC
} // CalAtomCollection<AbsorptionData>::CalAtomCollection()


void AbsorptionDataServer::dump(ostream &os)
{
	PROC

	os<<"AbsorptionDataServer{\n";
	for (const_iterator i = begin(); i != end(); ++i)
		os<<(*i).first<<endl;
	os<<"}";

	ENDPROC
} // AbsorptionDataServer::dump()


AbsorptionDataServer *CalAtomCollection<AbsorptionDataServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	PROC
	RETURN (*_collection)[0];
} // AbsorptionDataCollection<AbsorptionDataServer>::select()



vector<string> AbsorptionDataServer::materials() const
{
	PROC

	vector<string>	v;
	for (const_iterator i = begin(); i != end(); v.push_back((*i++).first));

	RETURN v;
} // AbsorptionDataServer::materials()


bool AbsorptionDataServer::hasMaterial(const string &materialSymbol) const
{
	PROC

	AbsorptionDataServer::MapType::const_iterator i = find(materialSymbol);
	RETURN i != end();
} // AbsorptionDataServer::hasMaterial()


const AbsCoefCurves &AbsorptionDataServer::henke(const string &materialSymbol)
const
{
	PROC

	const_iterator i = find(materialSymbol);
	if (i == end())
		errstr<<"There is no absorption curve for material `"
			  <<materialSymbol<<"'"<<Sas::error("NoSuchData");

	RETURN (*i).second.henke;
} // AbsorptionDataServer::henke()


const AtomicProperties &AbsorptionDataServer::atomicData(const string
															&materialSymbol)
const
{
	PROC

	const_iterator i = find(materialSymbol);
	if (i == end())
		errstr<<"There is no absorption curve for material `"
			  <<materialSymbol<<"'"<<Sas::error("NoSuchData");

	RETURN (*i).second.atomic;
} // AbsorptionDataServer::henke()


AbsorptionDataServer::~AbsorptionDataServer()
{
	PROC

	for (const_iterator i = begin(); i != end(); ++i) {
		const AbsCoefCurves &curve = (*i).second.henke;
		delete curve[0];
		delete curve[1];
	}

	ENDPROC
} // AbsorptionDataServer::~AbsorptionDataServer()




