//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	BadPixelDataServer --- implementation of BadPixel related classes
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1997-07-25: 1.0.0: - skeleton
//		UL, 2001-02-19: 1.1.0: - first version extracted from CalAtomsCommon.h
//							   - added mode dependence of bad pixels
//		UL, 2001-02-20: 1.1.1: - reading of MODE_N/A vector is sensitive
//								 to negative elements which are taken to
//								 end the list
//		UL, 2001-03-22: 1.1.0: - moved BadPixel/BadPixelMap classes to
//								 caloalutils
//		UL, 2001-04-09: 1.2.0: - MODE_N/A column can now be fixed or
//								 variable length
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/BadPixelDataServer.cc,v 1.6 2001/04/10 10:19:43 ulammers Exp $
//

#include <functional>	//	binders
#include <algorithm>
#include <cmath>

#include "CalOalUtility.h"
#include "CalServer.h"

using namespace CalOalUtility;

//
//	------------------------( BadPixelDataServer )-----------------------------
//
BadPixelDataServer::BadPixelDataServer(CalAtomCollection<BadPixelDataServer>
										*parent)
	: CalChild<CalAtomCollection<BadPixelDataServer> >(parent)
{
	PROC
	ENDPROC
} // BadPixelDataServer::BadPixelDataServer()


BadPixelMap BadPixelDataServer::map() const
{
	PROC

	BadPixelMap m;
	unsigned short mode = calServer.state()->mode()->get();
	unsigned N=_badPixels.size();
	for (unsigned n=0; n<N; ++n) {
		vector<short>::const_iterator i =
			find(_badPixels[n].modena.begin(), _badPixels[n].modena.end(),
				 mode);
		if (i != _badPixels[n].modena.end())
			//	pixel is not present in this mode
			continue;
		m[_badPixels[n].loc] = _badPixels[n].bp;
	}

	RETURN m;
} // BadPixelDataServer::map()


BadPixelDataServer::~BadPixelDataServer()
{
	PROC
	ENDPROC
} // BadPixelDataServer::~BadPixelDataServer()


CalAtomCollection<BadPixelDataServer>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	PROC
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs * _nNodes;
	BadPixelDataServer	*null = 0;
	_collection	= new vector<BadPixelDataServer *>(_nEntries, null);
	for (unsigned n=0; n<_nEntries;
		(*_collection)[n++] = new BadPixelDataServer(this));

	CifEntry *e = ccf->entry(instrument, Ccf::BadPix);

	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet, 
								e->queryBlockName(Ccf::toString(Ccf::BadPix)));
	const Table *table = ccftab.block();
	ColumnData *cdccdId	= table->column("CCD_ID")->data(),
			   *cdnodeId = table->column("NODE_ID")->data(),
			   *cdrawx = table->column("RAWX")->data(),
			   *cdrawy = table->column("RAWY")->data(),
			   *cdylen = table->column("YLENGTH")->data(),
			   *cddesc = table->column("DESCRIPTION")->data();
	int8	*ccdId	= cdccdId->int8Data(),
			*nodeId	= cdnodeId->int8Data();
	int16	*rawx	= cdrawx->int16Data(),
			*rawy	= cdrawy->int16Data(),
			*ylen	= cdylen->int16Data();


	bool hasModeCol = table->hasColumn("MODE_N/A");
	//	loop over all bad pixels
	for (unsigned n=0; n<table->rows(); ++n) {
		if (ccdId[n] <= 0 || ccdId[n] > INT(_nCCDs) ||
			nodeId[n] >= INT(_nNodes))
			errstr<<"Illegal chip and/or node number in row #"
				  <<n<<" of CCF file "
				  <<Xmm::toString(instrument->id())<<"/"
				  <<Ccf::toString(Ccf::BadPix)
				  <<Sas::error("IllegalValues");

		unsigned pos = index(ccdId[n], nodeId[n]);

		PixelLocation			l(rawx[n], rawy[n]);
		BadPixel::DefectTypes	t;
		char flagVal = stringColumnRow(cddesc, n)[0];
		switch(toupper(flagVal)) {
			case 'D'	: t = BadPixel::DEAD; break;
			case 'F'	: t = BadPixel::FLICKERING; break;
			case 'H'	: t = BadPixel::HOT; break;
			case 'P'	: t = BadPixel::PIN_HOLE; break;
			default		: t = BadPixel::ANOMALOUS; break;
		}

		BadPixel bp(l, ylen[n], t, islower(flagVal));
		BadPixelOm ombp(bp, 99);
		const BadPixel &badpix = instrument->id() != Xmm::OM
				? bp
					// this is an OM bad pixel - assign severity as well
					// endcoding has not been decided yet - assign 99
					// moment
				: ombp;
		vector<short> modeVec;
		if (hasModeCol) {
			CalOalUtility::fillVector(table, "MODE_N/A", n, modeVec);
			//	erase negative elements which mark the end of the vector
			modeVec.erase(find_if(modeVec.begin(), modeVec.end(),
						  bind2nd(less<short>(), 0)), modeVec.end());
		}

		BadPixelDataServer::LocBadPixModeType c = { l, badpix, modeVec };
		(*_collection)[pos]->_badPixels.push_back(c);
	}

	//	tidy up
	delete cdccdId;
	delete cdnodeId;
	delete cdrawx;
	delete cdrawy;
	delete cdylen;
	delete cddesc;

	dataSetServer->close(dataSet);

	ENDPROC
} // CalAtomCollection<BadPixelDataServer>::CalAtomCollection()


BadPixelDataServer *CalAtomCollection<BadPixelDataServer>::select(
					Ccd chipNr,
					Node nodeNr,
					unsigned) const
{
	PROC
	unsigned n = index(chipNr, nodeNr);

	RETURN (*_collection)[n];
} // CalAtomCollection<BadPixelDataServer>::select


#if 0
void BadPixelMapCollection::mergeNodes()
{
	PROC
	for (unsigned ccd=1; ccd<=_nCCDs; ++ccd) {
		BadPixelMap *m = (*_collection)[index(ccd, 0)];
		if (m)
			for (unsigned node=1; node<_nNodes; ++node) {
				unsigned n = index(ccd, node);
				BadPixelMap *m2 = (*_collection)[n];
				if (m2) {
					//	anachronism: GNU's STL map does not have
					//			     map.insert(iterator, iterator)
					const BadPixelMap::MapType::value_type
											*b = &*m2->_map.begin(),
											*e = &*m2->_map.end();
					m->_map.insert(b, e);
				}
				delete m2;
				(*_collection)[n] = 0;
			}
	}
	ENDPROC
} // BadPixelMapCollection::mergeNodes
#endif


