//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalAtomsBase.h	--- define basic CAL constituents
//
//	DESCRIPTION:
//		This file contains definitions of the basic data constituents
//		of the CAL, like Point-Spread-Functions, Bad-Pixel Map, etc.
//		These definitions are to be used by the CAL engine layer
//		who implement data manipulations using the CalAtoms in here.
//
//	CHANGE HISTORY:
//		UL, 25/10/97: 1.0.0: - skeleton
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/CalAtomsBase.h,v 1.11 2001/02/19 15:39:49 ulammers Exp $
//

#ifndef CAL_ATOMS_BASE_H
#define CAL_ATOMS_BASE_H

#include "common.h"
#include "CalCommon.h"
#include "Xmm.h"
#include "Ccf.h"
#include "Dal.h"
#include "Tracer.h"

#include <map>
#include <vector>

class CalAtom;	//	forward declaration

//
//	a collection of CalAtoms
//
class CalAtomCollectionBase {
public:
	virtual CalAtom *select(Ccd ccdNr, Node nodeNr,
							unsigned modeOffset = 0) const = 0;

	//	return attributes
	Filter filter() const { PROC RETURN _filter; }
	AccuracyLevel accuracy() const { PROC RETURN _accuracy; }

	size_t		entries() const { PROC RETURN _nEntries; }
	unsigned	ccds() const { PROC RETURN _nCCDs; }
	unsigned	nodes() const { PROC RETURN _nNodes; }
	unsigned	modes() const { PROC RETURN _nModes; }

	virtual ~CalAtomCollectionBase() {}

protected:
	CalAtomCollectionBase() {}
	unsigned		index(unsigned modeNr) const;
	unsigned		index(unsigned ccdNr, unsigned nodeNr) const;
	unsigned		indexCcdMode(unsigned ccdNr, unsigned modeNr) const;
	unsigned		index(unsigned ccdNr, unsigned nodeNr, unsigned modeNr)
					const;
	size_t			_nEntries;
	unsigned		_nCCDs, _nNodes, _nModes;
	Filter			_filter;		// filter setting
	AccuracyLevel	_accuracy;		// accuracy of atom collection
};


template<class AtomType>
class CalAtomCollection : public CalAtomCollectionBase {
public:
	//	empty default constructor
	CalAtomCollection() { PROC ENDPROC }

	//	construct from CCF constituent
	CalAtomCollection(const XmmInstrument *instrument,
					  const Ccf *ccf,
					  Filter filter,
					  AccuracyLevel accuracy);

	//	select an atom from the collection
	AtomType *select(Ccd ccd, Node node,
					 unsigned modeOffset = 0) const;

	Ccd ccd() const { PROC RETURN _ccd; }
	Node node() const { PROC RETURN _node; }
	unsigned modeOffset() const { PROC RETURN _modeOffset; }

	~CalAtomCollection();

protected:
	vector<AtomType *>	*_collection;	//	the collection
	mutable Ccd			_ccd;			//	selected CCD
	mutable Node		_node;			//	selected node
	mutable unsigned	_modeOffset;	//	selected modeOffset
};


//
//	collection destructor - always the same
//
template<class AtomType>
CalAtomCollection<AtomType>::~CalAtomCollection()
{
	PROC
	for (unsigned n=0; n<_collection->size(); ++n)
		delete (*_collection)[n];	// delete the atom
	delete _collection;
	ENDPROC
}


class CalAtom {
public:
	virtual ~CalAtom() {}
	virtual void dump(ostream &) {
		errstr<<"Sorry - dumping to an ostream is not implemented for this "
				"CAL atom"<<Sas::error("MethodUimplemented");
	}

protected:
	CalAtom() { _algoid = 0; }

	unsigned short	_algoid;
};

template<class AtomClass>
inline void calAtomUpcast(AtomClass *&ptr, CalAtom *atomPtr)
{
	PROC
	ptr = reinterpret_cast<AtomClass *>(atomPtr);
	ENDPROC
} // calAtomUpcast

#endif
