//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CalDataVector.cc	--- implementation of CalDataVector
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-05-31: 1.0.0: - first version
//		UL, 1999-12-23: 1.1.0: - added CalIntVector<int32>
//		UL, 2000-03-08: 1.2.0: - added CalVector::clone()
//
//	AUTHOR:
//		U. Lammers, Mon May 31 16:45:25 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/CalDataVector.cc,v 1.13 2001/02/26 17:23:25 ulammers Exp $
//

#include <cstring>

#include "errstr.h"
#include "common.h"
#include "Random.h"
#include "CalDataVector.h"
#include "CalServer.h"

template<class ElementType>
CalVector<ElementType>::CalVector(ElementType *ptr, unsigned n)
	: _ptr(ptr), _n(n), _memAllocated(false)
{
	PROC

	if (!ptr && n) {
		_ptr = new ElementType[n];
		_memAllocated = true;
	}

	ENDPROC
} // CalVector::CalVector()


template<class ElementType>
CalVector<ElementType>::CalVector(const CalVector<ElementType> &src)
{
	if (&src != this) {
		_ptr = src._ptr;
		_n = src._n;
		_memAllocated = src._memAllocated;
	}
}

template<class ElementType>
void CalVector<ElementType>::clone(const CalVector<ElementType> &source)
{
	if (this->_ptr == source._ptr)
		return;

	this->free();
	_n				= source._n;
	_ptr 			= new ElementType[_n];
	_memAllocated	= true;
	memcpy(_ptr, source._ptr, _n * sizeof(ElementType));
} // CalVector::clone()



template<class IntType>
CalIntVector<IntType>::CalIntVector(IntType *ptr, unsigned n,
									const CalServer *server)
    : CalVector<IntType>(ptr, n), CalChild<CalServer>(server)
{
	PROC
	ENDPROC
} // CalIntVector<IntType>::CalIntVector()


template<class IntType>
CalRealVector<real32> &CalIntVector<IntType>::toReal(CalRealVector<real32> &rv)
const
{
	PROC

	if (rv.size() != _n)
		errstr<<"Size of vectors do not match"<<Sas::error("SizeMismatch");

	
	const CalServer *p = this->parent();
	if (!p)
		p = &calServer;

	if (p->state()->randomize()->get())
		for (unsigned n=0; n<_n; ++n)
			rv[n] = FLT(_ptr[n]) + randGen.rand();
	else
		for (unsigned n=0; n<_n; ++n)
			rv[n] = FLT(_ptr[n]);

	RETURN rv;
} // CalIntVector<IntType>::toReal()


template<class RealType>
CalRealVector<RealType>::CalRealVector(RealType *ptr, unsigned n,
									   const CalServer *server)
    : CalVector<RealType>(ptr, n), CalChild<CalServer>(server)
{
	PROC
	ENDPROC
} // CalRealVector<RealType>::CalRealVector()



template<class RealType>
template<class ToIntType> CalIntVector<ToIntType>
	&CalRealVector<RealType>::toInt(CalIntVector<ToIntType> &iv) const
{
	PROC

	if (iv.size() != _n)
		errstr<<"Size of vectors do not match"<<Sas::error("SizeMismatch");

	for (unsigned n=0; n<_n; ++n)
		iv[n] = _ptr[n] >= 0. ? S_CAST(_ptr[n] + .5, ToIntType)
							  : S_CAST(_ptr[n] - .5, ToIntType);

	RETURN iv;
} // CalIntVector<RealType>::toInt()


//
//	explicit template instantiation
//
template class CalVector<int8>;
template class CalVector<int16>;
template class CalVector<int32>;
template class CalVector<real32>;
template class CalVector<real64>;

template class CalIntVector<int8>;
template class CalIntVector<int16>;
template class CalIntVector<int32>;
template class CalRealVector<real32>;
template class CalRealVector<real64>;

template CalIntVector<int8>
		&CalRealVector<real32>::toInt(CalIntVector<int8> &) const;
template CalIntVector<int16>
		&CalRealVector<real32>::toInt(CalIntVector<int16> &) const;

template CalIntVector<int8>
		&CalRealVector<real64>::toInt(CalIntVector<int8> &) const;
template CalIntVector<int16>
		&CalRealVector<real64>::toInt(CalIntVector<int16> &) const;
template CalIntVector<int32>
		&CalRealVector<real64>::toInt(CalIntVector<int32> &) const;

template void CalVector<int8>::clone(const CalVector<int8> &);
template void CalVector<int16>::clone(const CalVector<int16> &);
template void CalVector<int32>::clone(const CalVector<int32> &);
template void CalVector<real32>::clone(const CalVector<real32> &);
template void CalVector<real64>::clone(const CalVector<real64> &);

