//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CcdQuantumDataServer.cc --- implementation of CcdQuantumDataServer class
//
//	DESCRIPTION:
//		This implements all methods declared in the CcdQuantumDataServer class
//
//	CHANGE HISTORY:
//		UL, 2000-02-17: 1.0.0: - first version
//		UL, 2000-05-20: 1.1.0: - many internal changes
//		UL, 2000-05-24: 1.2.0: - reading of PATTERNLIB is only done for EMOS1/2
//								 now (fixes SPR 1347)
//		UL, 2000-06-27: 1.3.0: - bug fix in CcdQuantumDataServer::at()
//		UL, 2000-06-30: 1.4.0: - small fix to bring in line with selectlib-4.35
//		UL, 2000-08-03: 1.5.0: - added RgsCcdQuantumDataServer
//		UL, 2000-08-29: 1.6.0: - if a pixel location is outside the FOV
//								 a constant QE=0 is returned; this fixes
//								 SPR: 1717/1736: Cannot get QE for chipy=1
//		UL, 2000-08-31: 1.7.0: - bug fix in RgsQuantumDataServer constructor
//								 (erroneous release of CAL atom)
//
//	AUTHOR:
//		U. Lammers (UL), Thu Feb 17 14:13:14 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/CcdQuantumDataServer.cc,v 1.15 2000/12/20 16:04:50 ulammers Exp $
//

#include <cmath>
#include <algorithm>
#include "CalOalUtility.h"
#include "stringutil.h"
#include "CalServer.h"
#include "PatternLibServer.h"
#include "AbsorptionDataServer.h"
#include "PhysicalConstants.h"
#include "GeometryDataServer.h"
#include "CcdQuantumDataServer.h"

using namespace CalOalUtility;

//
//	---( CcdQuantumEfficiencyCurve )-------------------------------------------
//
CcdQuantumEfficiencyCurve &CcdQuantumEfficiencyCurve::operator+=(
										const CcdQuantumEfficiencyCurve &rhs)
{
	PROC
	Slatec::CubicHermiteInterpolant<double>::operator+=(rhs);
	RETURN *this;
} // CcdQuantumEfficiencyCurve::operator+=()


double CcdQuantumEfficiencyCurve::efficiency(double E) const
{
	PROC
	RETURN (*this)(E);
} // CcdQuantumEfficiencyCurve::Efficiency()


double CcdQuantumEfficiencyCurve::efficiency(double Elo, double Ehi) const
{
	PROC

	if (fabs(Ehi-Elo) < 1.e-8)
		ALTRETURNV(this->efficiency(Elo));

	RETURN this->integral(Elo, Ehi)/(Ehi-Elo);
} // CcdQuantumEfficiencyCurve::Efficiency


//
//	---( CcdQuantumDataServer )------------------------------------------------
//
CcdQuantumDataServer::CcdQuantumDataServer(Ccd ccd,
						 CalAtomCollection<CcdQuantumDataServer> *parent) :
		CalChild<CalAtomCollection<CcdQuantumDataServer> >(parent),
		_ccd(ccd)
{
	PROC
	ENDPROC
} // CcdQuantumDataServer::CcdQuantumDataServer()


CcdQuantumDataServer::~CcdQuantumDataServer()
{
	PROC
	ENDPROC
} // CcdQuantumDataServer::~CcdQuantumDataServer()



//
//	---( EpicCcdQuantumDataServer )-------------------------------------------
//
EpicCcdQuantumDataServer::EpicCcdQuantumDataServer(
										   Ccd ccd,
										   const Table *table,
										   const Proxy<vector<double> >&eGrid,
							 CalAtomCollection<CcdQuantumDataServer> *parent)
	: CalChild<CalAtomCollection<CcdQuantumDataServer> >(parent),
	  CcdQuantumDataServer(ccd, parent), 
	  _eGrid(eGrid), _atCalled(false), _regionIndex(-2)
{
	PROC

	vector<unsigned> pattVec;
	for (unsigned col=0; col<table->columns(); ++col) {
		vector<string> sv = breakup(table->column(col)->name(), "_");
		if (sv.size() != 2 && sv[0] != "QE")
			continue;
		pattVec.push_back(toNumber<unsigned>(sv[1]));
	}
	sort(pattVec.begin(), pattVec.end());

	_patterns = pattVec.size();
	//
	//	construct _gradePatterns vector
	//	_gradePatterns[n] = list of patterns with grade n
	//
	if (calServer.instrument()->id() == Xmm::EMOS1||
		calServer.instrument()->id() == Xmm::EMOS2) {
		//	only EMOS comes with pattern lib
		PatternLibServer *pls = calServer.getAtom(pls, false);
		if (pls)
			_gradePatterns = pls->gradePatternList();
	}

	ColumnData *cdr = table->column("REGION")->data(),
			   *cdqe[_patterns];
	real32	   *qe[_patterns];
	for (unsigned n=0; n<_patterns; ++n) {
		cdqe[n] = table->column("QE_" + toString(pattVec[n]))->data();
		qe[n] = cdqe[n]->real32Data();
	}

	//	loop over rows (=regions)
	unsigned N = table->rows();
	for (unsigned n=0; n<N; ++n) {
		EvalTuple et(newEvalTuple(cdr, n));

		EfficiencyVector effVec(_patterns);
		//	loop over columns (=patterns)
		for (unsigned i=0; i<_patterns; ++i) {
			unsigned patt = pattVec[i];

			vector<double> qeVec(eGrid->size());
			unsigned nElements = cdqe[i]->elements();
			if (nElements != eGrid->size())
				errstr<<"Size of QE efficiency data vector ("
					  <<nElements<<") for pattern "<<patt
					  <<"' in "<<table->qualifiedName()
					  <<" is not identical to size of energy grid ("
					  <<eGrid->size()<<")."
					  <<Sas::error("CcfDataIntegrityError");

			for (unsigned j=0; j<nElements; ++j)
				qeVec[j] = qe[i][cdqe[i]->position(n, j)];

			effVec[i] = new CcdQuantumEfficiencyCurve(*_eGrid, qeVec);
		}

		_localizedEfficiencyVectors.push_back(
										LocalizedEfficiencyVector(et, effVec));
	}

	for (unsigned n=0; n<_patterns; ++n)
		delete cdqe[n];
	delete cdr;

	ENDPROC
} // EpicCcdQuantumDataServer::EpicCcdQuantumDataServer()


void EpicCcdQuantumDataServer::checkAt() const
{
	PROC

	if (!_atCalled) {
		unsigned chipX = calServer.instrument()->maxX()/2,
				 chipY = calServer.instrument()->maxY()/2;
		at(chipX, chipY);
		errstr<<"No chip location for QE specified - will retrieve data "
				"applicable at center ("<<chipX<<'/'<<chipY<<")"
			  <<Sas::message(Msg::Lib, Msg::Verbose);
		_atCalled = false;
	}

	ENDPROC
} // EpicCcdQuantumDataServer::checkAt()


EpicCcdQuantumDataServer::EvalTuple
EpicCcdQuantumDataServer::newEvalTuple(const ColumnData *cdr, unsigned n) const
{
	string region(stringColumnRow(cdr, n));

	if (!Evaluator::isValidExpression(region))
		errstr<<"The region specification `"<<region<<"' in `"
			  <<cdr->parent()->qualifiedName()<<"' is syntactically "
				"incorrect"<<Sas::error("CcfDataIntegrityError");

	EvalTuple et;
	et.evaluator = new Evaluator(region);
	NumericSymbols syms = et.evaluator->numericIdentifiers();
	et.chipX = et.chipY = 0;
	if (syms.name.size() == 2) {
		if (uppercase(syms.name[0]) == "RAWX" &&
			uppercase(syms.name[1]) == "RAWY") {
			et.chipX = syms.valuePtr[0];
			et.chipY = syms.valuePtr[1];
		} else if (uppercase(syms.name[0]) == "RAWY" &&
				   uppercase(syms.name[1]) == "RAWX") {
			et.chipX = syms.valuePtr[1];
			et.chipY = syms.valuePtr[0];
		} else
			errstr<<"Unrecognized symbols "<<syms.name[0]<<'/'
				  <<syms.name[1]<<" in `"
				  <<cdr->parent()->qualifiedName()<<"' - expected RAWX + RAWY"
				  <<Sas::error("CcfDataIntegrityError");
	} else if (syms.name.size() != 0)
		errstr<<"Number of symbols in REGION specification `"
			  <<region<<"' from `"
			  <<cdr->parent()->qualifiedName()<<"' must be two: RAWX + RAWY"
			  <<Sas::error("CcfDataIntegrityError");

	RETURN et;
} // EpicCcdQuantumDataServer::newEvalTuple()


CcdQuantumEfficiencyCurve EpicCcdQuantumDataServer::pattern(
													unsigned patternId) const
{
	PROC

	if (patternId >= _patterns)
		errstr<<"Illegal pattern ID "<<patternId<<" in selecting quantum "
				"efficiency curve - legal range is [0, "<<_patterns-1
			  <<"]"<<Sas::error("IllegalPatternId");

	checkAt();
	assert(_regionIndex >= -1 &&
		   _regionIndex < INT(_localizedEfficiencyVectors.size()));

	if (_regionIndex >= 0) {
		const EfficiencyVector &effVec =
							_localizedEfficiencyVectors[_regionIndex].second;
		if (patternId > effVec.size())
			errstr<<"No QE data for pattern "<<patternId<<" available"
				  <<Sas::error("NoQEdata");

		ALTRETURNV(*effVec[patternId]);
	}

	//
	//	location is outside FOV - return QE=0
	//
	vector<double> eGrid(2); eGrid[0] = 0.; eGrid[1] = 1.e4;
	vector<double> qe0(2); qe0[0] = 0.; qe0[1] = 0.;
	RETURN CcdQuantumEfficiencyCurve(eGrid, qe0); 
} // EpicCcdQuantumDataServer::pattern()


CcdQuantumEfficiencyCurve EpicCcdQuantumDataServer::grade(unsigned gradeNr)
const
{
	PROC

	if (gradeNr < 1 ||
		gradeNr > _gradePatterns.size())
		errstr<<"Invalid grade value "<<gradeNr<<" - must be in interval [1, "
			  <<_gradePatterns.size()-1<<"]"
			  <<Sas::error("IllegalGradeValue");

	checkAt();
	assert(_regionIndex >= -1 &&
		   _regionIndex < INT(_localizedEfficiencyVectors.size()));

	if (_regionIndex >= 0) {
		const EfficiencyVector &effVec =
							_localizedEfficiencyVectors[_regionIndex].second;

		const vector<unsigned short> &patternVec = _gradePatterns[gradeNr-1];
		if (!patternVec.size())
			errstr<<"There are no grade-"<<gradeNr<<" patterns"
				  <<Sas::error("NoSuchPattern");

		CcdQuantumEfficiencyCurve ec = *effVec[patternVec[0]];
		for (unsigned n=1; n<patternVec.size(); ++n)
			ec += *effVec[patternVec[n]];

		ALTRETURNV(ec);
	}

	//
	//	location is outside FOV - return QE=0
	//
	vector<double> eGrid(2); eGrid[0] = 0.; eGrid[1] = 1.e4;
	vector<double> qe0(2); qe0[0] = 0.; qe0[1] = 0.;
	RETURN CcdQuantumEfficiencyCurve(eGrid, qe0); 
} // EpicCcdQuantumEfficiencyCurve::pattern()


const CcdQuantumDataServer &EpicCcdQuantumDataServer::at(unsigned short chipX,
														 unsigned short chipY)
const
{
	PROC

	_atCalled = true;

	//	search for first region applicable to rawX/rawY
	unsigned N = _localizedEfficiencyVectors.size();
	for (unsigned n=0; n<N; ++n) {
		const EvalTuple &et = _localizedEfficiencyVectors[n].first;
		if (et.chipX && et.chipY) {
			*et.chipX = DBL(chipX);
			*et.chipY = DBL(chipY);
		}
		if (et.evaluator->evalLogic()) {
			//	yup - found it
			_regionIndex = n;
			ALTRETURNV(*this);
		}
	}

	//
	//	could not find applicable region - check if the requested location
	//	is outside the FOV
	//
	int16 x = chipX,
		  y = chipY;
	CalInt16Vector chipXv(&x, 1),
				   chipYv(&y, 1);
	GeometryDataServer *ct = calServer.getAtom(ct);
	XmmPixelList pl = ct->chipCoordList(chipXv, chipYv);
	vector<bool> inFov = pl.insideFOV();
	if (inFov[0])
		//
		//	pixel is inside FOV but no applicable QE region could be found -
		//	this is a CCF problem
		//
		errstr<<"Could not find applicable region for rawX/rawY="
			  <<chipX<<'/'<<chipY<<" in quantum efficiency data retrieval"
			  <<Sas::error("NoQEdata");

	//	signal that pixel is outside FOV
	_regionIndex = -1;

	RETURN *this;
} // EpicCcdQuantumDataServer::at()


EpicCcdQuantumDataServer::~EpicCcdQuantumDataServer()
{
	PROC

	for (unsigned n=0; n<_localizedEfficiencyVectors.size(); ++n) {
		delete _localizedEfficiencyVectors[n].first.evaluator;
		EfficiencyVector &effVec = _localizedEfficiencyVectors[n].second;
		unsigned N = effVec.size();
		for (unsigned nn=0; nn<N; delete effVec[nn++]);
		effVec.resize(0);
	}

	ENDPROC
} // EpicCcdQuantumDataServer::~EpicCcdQuantumDataServer()


//
//	RGS
//
RgsCcdQuantumDataServer::RgsCcdQuantumDataServer(
					     Ccd ccd,
						 const Table *materialThicknessTable,
						 const Table *SiThicknessTable,
						 CalAtomCollection<CcdQuantumDataServer> *parent)
	: CalChild<CalAtomCollection<CcdQuantumDataServer> >(parent),
	  CcdQuantumDataServer(ccd, parent), _currentSiThickness(0.)
{
	PROC

	//
	//	get thicknesses of material layers
	//
	ColumnData *cdccdid = materialThicknessTable->column("CCDID")->data(),
			   *cdal = materialThicknessTable->column("D_AL")->data(),
			   *cdmgf2 = materialThicknessTable->column("D_MGF2")->data(),
			   *cdsi = materialThicknessTable->column("D_SI")->data(),
			   *cdsio2 = materialThicknessTable->column("D_SIO2")->data(),
			   *cdalo2 = materialThicknessTable->column("D_ALO2")->data();
	int8 *ccdid = cdccdid->int8Data();
	real32 *al = cdal->real32Data(),
		   *mgf2 = cdmgf2->real32Data(),
		   *si = cdsi->real32Data(),
		   *sio2 = cdsio2->real32Data(),
		   *alo2 = cdalo2->real32Data();

	unsigned N = materialThicknessTable->rows();
	unsigned n;
	for (n=0; n<N; ++n)
		if (ccdid[n] == ccd)
			break;
	if (n>=N)
		errstr<<"No entry for CCD "<<ccd<<" in `"
			  <<materialThicknessTable->qualifiedName()
			  <<Sas::error("CcfIntegrityError");

	double dAl = 1.e-7 * al[n],		//	nm -> cm
		   dMgF2 = 1.e-7 * mgf2[n],
		   dSi = 1.e-7 * si[n],
		   dSiO2 = 1.e-7 * sio2[n],
		   dAl2O3 = 1.e-7 * alo2[n];

	delete cdccdid;
	delete cdal;
	delete cdmgf2;
	delete cdsi;
	delete cdsio2;
	delete cdalo2;

	//
	//	construct Si thickness map
	//
	double dActiveSi = SiThicknessTable->attribute("MTHICK")->asReal();
	if (SiThicknessTable->hasColumn("CHIPX")) {
		ColumnData *cdchipX = SiThicknessTable->column("CHIPX")->data(),
				   *cdchipY = SiThicknessTable->column("CHIPY")->data(),
				   *cdfringes = SiThicknessTable->column("FRINGES")->data();
		_thicknessSi = new ThicknessSiInterpolator(SiThicknessTable->rows(),
												   cdchipX->int16Data(),
												   cdchipY->int16Data(),
												   cdfringes->int16Data(),
												   dActiveSi);
		delete cdchipX;
		delete cdchipY;
		delete cdfringes;
	} else
		_thicknessSi = new ThicknessSiInterpolator(dActiveSi);

	//
	//	construct transmission / tau*henke curves
	//
	AbsorptionDataServer *ads = calServer.getAtom(ads);
	const AbsCoefCurve &f2Al	= ads->henke("Al").f2(),
					   &f2F		= ads->henke("F").f2(),
					   &f2Mg	= ads->henke("Mg").f2(),
					   &f2O		= ads->henke("O").f2(),
					   &f2Si	= ads->henke("Si").f2();
	double	AAl		= ads->atomicData("Al").weight(),			// [g/mol]
			AF		= ads->atomicData("F").weight(),
			AMg		= ads->atomicData("Mg").weight(),
			AO		= ads->atomicData("O").weight(),
			ASi		= ads->atomicData("Si").weight(),

			rhoAl	= ads->atomicData("Al").density() * 1.e-3,	// [g/cm^3]
			rhoSi	= ads->atomicData("Si").density() * 1.e-3;

	double	AMgF2	= AMg + 2. * AF,
			rhoMgF2	= 3.15,	//	[g/cm^3]
			AAl2O3	= 2. * AAl + 3. * AO,
			rhoAl2O3= 3.97,	//	[g/cm^3]
			ASiO2	= ASi + 2. * AO,
			rhoSiO2	= 2.64;	//	[g/cm^3]

	_eGrid = f2Si.abscissa();
	N = _eGrid.size();
	vector<double> trans(N),
				   tau(N);
	for (unsigned n=0; n<N; ++n) {
		double eV = _eGrid[n];								// [eV]
		double f0 = PhysicalConstants::avogadro * 2. *		// [1/mol]
					PhysicalConstants::r0 * 1.e2 *			// [cm]
					PhysicalConstants::eVtimesAo/eV * 1.e-8,// [cm]
			   f1 = // MgF2
						1. * rhoMgF2/AMgF2 * dMgF2 * f2Mg(eV) +
						2. * rhoMgF2/AMgF2 * dMgF2 * f2F(eV) +
					// Al
						1. * rhoAl/AAl * dAl * f2Al(eV) +
					// Al2O3
						3. * rhoAl2O3/AAl2O3 * dAl2O3 * f2O(eV) +
					// SiO2
						1. * rhoSiO2/ASiO2 * dSiO2 * f2Si(eV) +
						2. * rhoSiO2/ASiO2 * dSiO2 * f2O(eV) +
					// Si
						1. * rhoSi/ASi * dSi * f2Si(eV);

		trans[n] = exp(-f0*f1);
		tau[n] = f0 * rhoSi/ASi * f2Si(eV);
	}

	_transmission = InterpolatedCurve(_eGrid, trans);
	_tau = InterpolatedCurve(_eGrid, tau);

	ENDPROC
} // RgsCcdQuantumDataServer::RgsCcdQuantumDataServer()


CcdQuantumEfficiencyCurve RgsCcdQuantumDataServer::pattern(unsigned patternId)
const
{
	PROC

	if (patternId)
		errstr<<"Can only give pattern-0 QE curves"
			  <<Sas::error("NoSuchQEcurve");

	RETURN grade(1);
} // RgsCcdQuantumDataServer::pattern()


CcdQuantumEfficiencyCurve RgsCcdQuantumDataServer::grade(unsigned gradeNr)
const
{
	PROC

	if (gradeNr != 1)
		errstr<<"QE only available for grade-1 events"
			  <<Sas::error("NoSuchQE");

	if (_currentSiThickness == 0.) {
		errstr<<"Computing average Si-layer thickness for QE calculation"
			  <<Sas::message(Msg::Lib, Msg::Verbose);
		averageSiThickness();
	}

	unsigned N = _eGrid.size();
	vector<double> v(N);
	for (unsigned n=0; n<N; ++n) {
		double e = _eGrid[n];
		v[n] = (1. - exp(-_currentSiThickness * _tau(e))) * _transmission(e);
	}

	RETURN CcdQuantumEfficiencyCurve(_eGrid, v);
} // RgsCcdQuantumEfficiencyCurve::grade()


void RgsCcdQuantumDataServer::averageSiThickness() const
{
	PROC

	unsigned short minX = calServer.instrument()->minX(),
				   maxX = calServer.instrument()->maxX(),
				   minY = calServer.instrument()->minY(),
				   maxY = calServer.instrument()->maxY();

	unsigned n=0;
	for (unsigned ix=minX; ix<=maxX; ix+=10)
		for (unsigned iy=minY; iy<=maxY; iy+=10) {
			++n;
			_currentSiThickness += (*_thicknessSi)(ix, iy);
		}

	_currentSiThickness /= DBL(n);

	ENDPROC
} // RgsCcdQuantumDataServer::averageSiThickness()


const CcdQuantumDataServer &RgsCcdQuantumDataServer::at(unsigned short chipX,
														unsigned short chipY)
const
{
	PROC

	_currentSiThickness = (*_thicknessSi)(chipX, chipY);

	RETURN *this;
} // RgsCcdQuantumDataServer::at()


RgsCcdQuantumDataServer::~RgsCcdQuantumDataServer()
{
	PROC

	delete _thicknessSi;

	ENDPROC
} // RgsCcdQuantumDataServer::~RgsCcdQuantumDataServer()



CalAtomCollection<CcdQuantumDataServer>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	PROC

	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs;
	CcdQuantumDataServer *null = 0;
	_collection = new vector<CcdQuantumDataServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::QuantumEf);
	DataSet *dataSet = e->open();

	switch(instrument->id()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2:
		case Xmm::EPN: {
			//	read energy grid
			CcfConstituentTable ccftab(dataSet, "EBINS");
			const Table *table = ccftab.block();
			ColumnData *cde = table->column("ENERGY")->data();
			real32 *energy = cde->real32Data();
			Proxy<vector<double> > eGrid = new vector<double>(energy,
													  energy+cde->elements());
			delete cde;

			for (unsigned n=0; n<_collection->size(); ++n) {
				CcdQuantumDataServer *&qeds = (*_collection)[n];

				string tableName = "QE_CCD" + toString(n+1);
				if (!dataSet->hasBlock(tableName)) {
					errstr<<"No QE data for CCD #"<<n+1<<" in "
						  <<dataSet->name()<<Sas::warning("NoQEdata");
					continue;
				}

				CcfConstituentTable ccftab(dataSet, tableName);
				const Table *table = ccftab.block();
				qeds = new EpicCcdQuantumDataServer(n+1, table, eGrid, this);
			}
			break; }
		case Xmm::RGS1:
		case Xmm::RGS2: {
			CcfConstituentTable ccftab(dataSet, "CCD_DESC");
			const Table *materialThicknessTable = ccftab.block();
			for (unsigned n=0; n<_collection->size(); ++n) {
				CcdQuantumDataServer *&qeds = (*_collection)[n];
				string tableName = "SI" + toString(n+1);
				if (!dataSet->hasBlock(tableName)) {
					errstr<<"No QE data for CCD #"<<n+1<<" in "
						  <<dataSet->name()<<Sas::warning("NoQEdata");
					continue;
				}

				CcfConstituentTable ccftab(dataSet, tableName);
				const Table *table = ccftab.block();
				qeds = new RgsCcdQuantumDataServer(n+1, materialThicknessTable,
												   table, this);
			}			
			break; }
		default:
			break;
	}

	dataSetServer->close(dataSet);

	ENDPROC
} // CalAtomCollection<CcdQuantumDataServer>::CalAtomCollection()



CcdQuantumDataServer *CalAtomCollection<CcdQuantumDataServer>::select(
					Ccd ccd,
					Node,
					unsigned) const
{
	PROC
	RETURN (*_collection)[ccd-1];
} // CalAtomCollection<CcdQuantumDataServer>::select()
