//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	CcdQuantumDataServer --- the CcdQuantumDataServer class
//
//	DESCRIPTION:
//		A class 'CcdQuantumDataServer' is defined which provides access to
//		quantum efficiency data.
//
//	CHANGE HISTORY:
//		UL, 2000-02-17: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers (UL), Thu Feb 17 13:47:35 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Data Access Handbool, C. Erd et al.
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/CcdQuantumDataServer.h,v 1.7 2000/10/05 13:34:05 ulammers Exp $
//

#ifndef CCD_QUANTUM_DATA_SERVER_H
#define CCD_QUANTUM_DATA_SERVER_H

#include "Tracer.h"
#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "slatec++.h"
#include "Evaluator.h"

//
//	---( CcdQuantumEfficiencyCurve )-------------------------------------------
//
class CcdQuantumEfficiencyCurve :
							public Slatec::CubicHermiteInterpolant<double>{
public:
	CcdQuantumEfficiencyCurve() {}
	CcdQuantumEfficiencyCurve(const vector<double> &eGrid,
							  const vector<double> &effvec)
		: Slatec::Interpolant<double>(eGrid),
		  Slatec::CubicHermiteInterpolant<double>(eGrid, effvec)
			{ PROC ENDPROC }


	CcdQuantumEfficiencyCurve &operator+=(const CcdQuantumEfficiencyCurve &);
		//
		//	PURPOSE:
		//		Add a QE curve to the current one
		//


	virtual double efficiency(double E) const;
		//
		//	PURPOSE:
		//		Evaluate CCD quantum efficiency curve at a single energy <E>
		//
		//	ARGUMENTS:
		//		E	: energy in eV
		//
		//	RETURNS:
		//		Value of CCD quantum efficiency curve - unitless number in
		//		range [0, 1]
		//


	virtual double efficiency(double Elo, double Ehi) const;
		//
		//	PURPOSE:
		//		Return weighted integral of CCD quantum efficiency curve
		//		between the two energies
		//
		//	ARGUMENTS:
		//		Elo, Ehi	: energies in eV between which quantum efficiency
		//					  is to be integrated; Elo < Elo
		//
		//	RETURNS:
		//
		//				 Ehi
		//		   1	/
		//		------	| dE qe(E)
		//		Ehi-Elo	/
		//				Elo
		//

	~CcdQuantumEfficiencyCurve() { PROC ENDPROC }
};

//
//	---( CcdQuantumDataServer )----------------------------------------
//
class CcdQuantumDataServer : public CalAtom,
	 public virtual CalChild<CalAtomCollection<CcdQuantumDataServer> > {
public:
	CcdQuantumDataServer(Ccd ccd,
						 CalAtomCollection<CcdQuantumDataServer> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate CcdQuantumDataServer object applicable to CCD
		//		<ccd>
		//


	virtual void dump(ostream &) { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	static unsigned id() { PROC RETURN Ccf::QuantumEf; }
		//
		//	PURPOSE:
		//		Return ID of CalAtom
		//


	virtual unsigned patterns() const { PROC RETURN 0; }
		//
		//	PURPOSE:
		//		Returns maximum number of patterns held by the server
		//


	virtual CcdQuantumEfficiencyCurve pattern(unsigned patternId) const = 0;
	virtual CcdQuantumEfficiencyCurve grade(unsigned gradeNr) const = 0;
		//
		//	PURPOSE:
		//		Select a quantum efficiency curve for a particular
		//		pattern/grade
		//
		//	ARGUMENTS:
		//		gradeNr			: grade (number of pixels above threshold)
		//		patternId		: pattern ID ([0, 31] for EMOS)
		//


	virtual const CcdQuantumDataServer &at(unsigned short chipX,
										   unsigned short chipY) const = 0;
		//
		//	PURPOSE:
		//		Return quantum efficiency curve applicable at CCD pixel
		//		position (chipX, chipY) for pattern specified with preceding
		//		invocation of method pattern()
		//
		//	ARGUMENTS:
		//		chipX, chipY	: pixel position in CHIPCOORD frame at which
		//						  efficiency curve is requested
		//


	virtual ~CcdQuantumDataServer();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//

private:
	//	data members
	Ccd				_ccd;
}; // CcdQuantumDataServer


//
//	instrument specific CcdQuantumDataServer
//	
//	EPIC (EMOS/EPN)
//
class EpicCcdQuantumDataServer : public CcdQuantumDataServer {
public:
	EpicCcdQuantumDataServer(Ccd ccd,
							 const Table *table,
							 const Proxy<vector<double> > &eGrid,
						 CalAtomCollection<CcdQuantumDataServer> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate EpicCcdQuantumDataServer object from energy grid
		//		vector and corresponding Efficiency data
		//


	unsigned patterns() const { PROC RETURN _patterns; }
		//
		//	PURPOSE: see above
		//


	CcdQuantumEfficiencyCurve pattern(unsigned patternId) const;
		//
		//	PURPOSE: see above
		//
	

	CcdQuantumEfficiencyCurve grade(unsigned gradeNr) const;
		//
		//	PURPOSE: see above
		//
	

	const CcdQuantumDataServer &at(unsigned short chipX,
								   unsigned short chipY) const;
		//
		//	PURPOSE: see above
		//


	virtual ~EpicCcdQuantumDataServer();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//

private:
	//	data members
	struct EvalTuple {
		const Evaluator	*evaluator;
		double			*chipX,
						*chipY;
	};
	typedef vector<CcdQuantumEfficiencyCurve *>	EfficiencyVector;
	typedef pair<EvalTuple, EfficiencyVector>	LocalizedEfficiencyVector;
	vector<LocalizedEfficiencyVector>			_localizedEfficiencyVectors;

	Proxy<vector<double> >			_eGrid;

	mutable bool					_atCalled;
	mutable int						_regionIndex;
	unsigned						_patterns;		//	max number of patterns
	vector<vector<unsigned short> >	_gradePatterns;	//	_gradePatterns[i] =
													//	list of all grade-i
													//	patterns

	//	methods
	EvalTuple newEvalTuple(const ColumnData *cdr, unsigned n) const;
	void	  checkAt() const;
}; // EpicCcdQuantumDataServer


//
//	RGS
//
class RgsCcdQuantumDataServer : public CcdQuantumDataServer {
public:
	RgsCcdQuantumDataServer(Ccd ccd,
						 const Table *materialThicknessTable,
						 const Table *SiThicknessTable,
						 CalAtomCollection<CcdQuantumDataServer> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate RgsCcdQuantumDataServer object
		//


	CcdQuantumEfficiencyCurve pattern(unsigned patternId) const;
	CcdQuantumEfficiencyCurve grade(unsigned gradeNr) const;
		//
		//	PURPOSE:
		//		Select a quantum efficiency curve for a particular
		//		pattern/grade
		//
		//	ARGUMENTS:
		//		gradeNr			: grade (number of pixels above threshold)
		//		patternId		: pattern ID ([0, 31] for Rgs)
		//


	const CcdQuantumDataServer &at(unsigned short chipX,
								   unsigned short chipY) const;
		//
		//	PURPOSE:
		//		Return quantum efficiency curve applicable at CCD pixel
		//		position (chipX, chipY) for pattern specified with preceding
		//		invocation of method pattern()
		//
		//	ARGUMENTS:
		//		chipX, chipY	: pixel position in CHIPCOORD frame at which
		//						  efficiency curve is requested
		//


	virtual ~RgsCcdQuantumDataServer();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//

private:
	class ThicknessSiInterpolator {
	public:
		ThicknessSiInterpolator(double thickness)
			: _thickness(1.e-4*thickness), _thicknessMap(0) { PROC ENDPROC }
		ThicknessSiInterpolator(unsigned N, int16 *chipX, int16 *chipY,
								int16 *fringes, double d)
		{
			PROC
			vector<float> x(N), y(N), v(N);
			for (unsigned n=0; n<N; ++n) {
				x[n] = FLT(chipX[n]);
				y[n] = FLT(chipY[n]);
				v[n] = 1.e-4*(d - FLT(fringes[n])/7.); // [2], Sect. 3.5.22.1.1
			}
			_thicknessMap = new Slatec::Bivariate2DInterpolantScattered<float>(
											x, y, v);
			ENDPROC
		}
		double operator()(int16 chipX, int16 chipY)
		{
			PROC
			RETURN _thicknessMap ? (*_thicknessMap)(FLT(chipX), FLT(chipY))
								 : _thickness;
		}
		~ThicknessSiInterpolator() { PROC delete _thicknessMap;  ENDPROC }
	private:
		double											_thickness;
		Slatec::Bivariate2DInterpolantScattered<float>  *_thicknessMap;
	};

	ThicknessSiInterpolator				*_thicknessSi;
	vector<double>						_eGrid;
	InterpolatedCurve					_transmission,
										_tau;

	mutable double						_currentSiThickness;

	//	methods
	void averageSiThickness() const;
}; // RgsCcdQuantumDataServer

#endif
