//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Ccf.h	--- interface between the CAL and the CCF
//
//	DESCRIPTION:
//		This defines the interface between the CCF and the CAL
//
//	CHANGE HISTORY:
//		UL, 1997-10-10: 1.0.0: - skeleton
//		UL, 1999-03-09: 1.1.0: - some methods in CifEntry now return
//								 constant references to data members instead
//								 of references
//		UL, 1999-08-10: 1.2.0: - added Ccf::cifName()
//		UL, 2000-10-05: 2.0.0: - added Ccf::cifDirName() / Ccf::ccfSearchPath()
//							   - added CifEntry::open()
//		UL, 2000-12-18: 2.1.0: - added CcfConstituentBlock
//
//	AUTHOR:
//		U. Lammers, Fri Oct 10 12:43:51 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Indexing of the XMM Current Calibration File, U. Lammers,
//			 15/07/97
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/Ccf.h,v 1.22 2000/12/20 16:04:50 ulammers Exp $
//

#ifndef CCF_H
#define CCF_H

#include <string>
#include <set>
#include <utility>

#include "Dal.h"
#include "Tracer.h"
#include "Xmm.h"
#include "STime.h"
#include "CalChild.h"
#include "StringIntMap.h"

class CifEntry;	// forward declaration
class CalServer;

class Ccf : public CalChild<CalServer> {
public:
	Ccf(const string cifFileName = "", const CalServer *calServerP = 0);
		// open a CIF defining the CCF

	enum CalDataTypes { NONE = -1,
		AbsCoefs = 0,
		ADUConv,
		AstroMet,
		Background,
		BadPix,
		BoreSight,
		CCDrmf,
		CTI,
		CalIndex,
		CalSourceData,
		ClockPatterns,
		ColorTrans,
		Contam,
		CrossPSF,
		DarkCurrent,
		DarkFrame,
		DiffuseGala,
		Efficiency,
		EvtSizeDist,
		EXAFS,
		FilterTransV,
		FilterTransX,
		HKParmInt,
		LeapSeconds,
		LightLeak,
		LinCoord,
		LineSpreadFunc,
		LargeScaleSens,
		MiscData,
		ModeParam,
		PatternLib,
		PixToPixSens,
		PhotToNat,
		PhotVsLam,
		PixToLam,
		PSF1DRB,
		PSFPileUp,
		QuantumEf,
		QuickMag,
		Redist,
		RedistPileUp,
		StrayLight,
		TimeCorr,
		VAreaEf,
		VPSF,
		VSLAreaEf,
		VSLDist,
		XAreaEf,
		XEncirEn,
		XPSF,
		XSLAreaEf,
		XSLDist,
		Zodiacal,
		DiagnosticOffset,
		GratingData
	};

	static const unsigned N_CALDATATYPES = GratingData+1;

	#define TYPE_STR \
		"ABSCOEFS", "ADUCONV", "ASTROMET", "BACKGROUND", "BADPIX", \
		"BORESIGHT", "CCDRMF", "CTI", "CALINDEX", "CALSOURCEDATA", \
		"CLOCKPATTERNS", \
		"COLORTRANS", "CONTAM", "CROSSPSF", "DARKCURRENT", "DARKFRAME", \
		"DIFFUSEGALA", "EFFICIENCY", "EVTSIZEDIST", "EXAFS", "FILTERTRANSV", \
		"FILTERTRANSX", "HKPARMINT", "LEAPSECONDS", "LIGHTLEAK", "LINCOORD", \
		"LINESPREADFUNC", "LARGESCALESENS", "MISCDATA", "MODEPARAM", \
		"PATTERNLIB", "PIXTOPIXSENS", "PHOTTONAT", "PHOTVSLAM", "PIXTOLAM", \
		"PSF1DRB", "PSFPILEUP", "QUANTUMEF", "QUICKMAG", "REDIST", \
		"REDISTPILEUP", "STRAYLIGHT", "TIMECORR", "VAREAEF", "VPSF", \
		"VSLAREAEF", "VSLDIST", "XAREAEF", "XENCIREN", "XPSF", "XSLAREAEF", \
		"XSLDIST", "ZODIACAL", "DIAGNOSTIC_OFFSETS", "GRATING_DATA"

	//
	//	Ccf::toTypeId("ZODIACAL") == Ccf::ZODICAL
	//	Ccf::toString(Ccf::ZODIACAL) == "ZODIACAL"
	//
	static CalDataTypes toTypeId(const string &typeString);
	static const string &toString(CalDataTypes dataTypeId);
	static const string &toString(unsigned dataTypeId);


	static const char *const SAS_CCF			= "SAS_CCF";
		//	environment variable pointing to block "CALINDEX" containing
		//	the CIF; can also be the name of a directory in which case the
		//	CIF is supposed to have a default name SAS_CIFDFNAME

	static const char *const SAS_DEFAULT_CIFNAME= "ccf.cif";
		//	default name of the CIF

	static const char *const SAS_CCFPATH		= "SAS_CCFPATH";
		//	environment variable containing colon separated list of
		//	directories; this defines the search path for locating
		//	CCF constituents

	//
	//	name of CIF / directory name of CIF / CCF constituent search path
	//
	const string &cifName() const { PROC RETURN _cifFileName; }
	const string &cifDirName() const { PROC RETURN _cifDirName; }
	const string &ccfSearchPath() const { PROC RETURN _ccfSearchPath; }

	//	set the search path for CCF constituents; must be a colon
	//	separated list of directory names
	void ccfSearchPath(const string &path) { PROC _ccfSearchPath = path;
											 ENDPROC }

	//
	//	register new CifEntry
	//
	virtual void registerNewCifEntry(const CifEntry &) = 0;

	//	query for a calibration file of a particular type corresponding to an
	//	instrument
	virtual bool hasEntry(const XmmInstrument *, Ccf::CalDataTypes) const =0;
	virtual CifEntry *entry(const XmmInstrument *, Ccf::CalDataTypes,
							bool generateError = true) const = 0;

	virtual void print(ostream &) const;
	friend ostream &operator<<(ostream &, const Ccf *);

	virtual ~Ccf();						// destructor

	typedef pair<Xmm::Instrument, Ccf::CalDataTypes> CcfConstituent;
	typedef map<CcfConstituent, CifEntry *, less<CcfConstituent> > Cif;

protected:
	string _cifFileName,				//	name of current CIF (full path)
		   _cifDirName,					//	directory
		   _ccfSearchPath;				//	search path for constituents
			
private:
	static StringIntMap<Ccf::CalDataTypes>	_calDataTypes;

	friend class CalviewGui;
	friend class ShowCcf;
	virtual const Cif &cif() const = 0;
}; // Ccf

//
//	print CCF to a stream
//
ostream &operator<<(ostream &, const Ccf *);


template<class T>
class CcfConstituentBlock {
public:
	CcfConstituentBlock(const DataSet *set, const string &name,
						unsigned acceptAlgoIdsLessEqThan = 0);
	CcfConstituentBlock(const DataSet *set, unsigned nr,
						unsigned acceptAlgoIdsLessEqThan = 0);
		//
		//	PURPOSE:
		//		Construct object; a CcfConstituentBlock is a block
		//		(table or array) in a CCF constituent
		//
		//	ARGUMENTS:
		//		set		: data set containing the block
		//		name	: name of the block to access
		//		number	: number of block to access
		//		acceptAlgoIdsLessEqThan:
		//				  each block is associated with a dedicated calibration
		//				  algorithm identified through its ID (>=0);
		//				  the ID is encoded in attribute
		//					ALGOID
		//				  in the table; a missing attribute is taken as
		//				  ALGOID=0; all IDs less than or equal to
		//				  <acceptAlgoIdsLessEqThan> will be accepted otherwise
		//				  an error condition is raised
		//


	const T *block() const { PROC RETURN _block; }
		//
		//	PURPOSE:
		//		Access block associated with this object
		//


	unsigned algoid() const { PROC RETURN _algoid; }
		//
		//	PURPOSE:
		//		Return ALGOID of the block
		//


	~CcfConstituentBlock() { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Destroy object
		//

private:
	unsigned	_algoid;
	const T		*_block;
};


typedef CcfConstituentBlock<Table>	CcfConstituentTable;
typedef CcfConstituentBlock<Array>	CcfConstituentArray;


class CifEntry : public CalChild<Ccf> {
public:
	//	construct from data tuple
	CifEntry(Xmm::Instrument instrument, Ccf::CalDataTypes calDataType,
			 int issue, const string &fileName,
			 const string &valDate, const string &subDate,
			 const string &extKey, const string &extSeq,
			 const Ccf *ccf = 0)
			: CalChild<Ccf>(ccf),
			 _instrument(instrument),
			 _calDataType(calDataType),
			 _issue(issue),
			 _fileName(fileName),
			 _validityDate(valDate),
			 _submissionDate(subDate),
			 _extensionKey(extKey),
			 _extensionSequence(extSeq)
		   {}

	//	construct from calibration file name
	CifEntry(const string calFileName, const Ccf *ccf = 0);

	//	copy constructor
	CifEntry &operator=(const CifEntry &entry);
	CifEntry(const CifEntry &entry);

	//	query block number/name
	bool hasBlock(const string &truncatedBlockName) const;
	string queryBlockName(const string &truncatedBlockName,
						  unsigned *blockNr = 0) const;

	Xmm::Instrument		instrument() const { PROC RETURN _instrument; }
	Ccf::CalDataTypes	calDataType() const { PROC RETURN _calDataType; }
	int					issue() const { PROC RETURN _issue; }
	string				fileName() const { PROC RETURN _fileName; }
	const STime			&validityDate() const { PROC RETURN _validityDate; }
	const STime			&submissionDate() const { PROC RETURN _submissionDate;}
	const string		&extensionKey() const { PROC RETURN _extensionKey; }
	const string		&extensionSequence() const { PROC
												RETURN _extensionSequence; }

	DataSet *open(DataSetServer::MemoryModel memModel = DataSetServer::Low,
				  bool logAccess = true) const;
		//
		//	PURPOSE:
		//		Try to open the CCF constituent in Read-Mode using the
		//		file name specified in the CIF. If this fails and SAS_CCFPATH
		//		is defined try to locate the file along the path(s); if
		//		SAS_CCFPATH	is undefined raise an error condition
		//
		//	ARGUMENTS:
		//		memModel	: DAL memory model to open the data set with
		//		logAccess	: log this access in global CcfAccessLog?
		//

	//	destructor
	~CifEntry();

	friend ostream &operator<<(ostream &, const CifEntry &);

private:
	//	default constructor
	explicit CifEntry() : CalChild<Ccf>(0) {}

	// data members
	Xmm::Instrument		_instrument;			// instrument name
	Ccf::CalDataTypes	_calDataType;			// file type
	int					_issue;					// file issue number
	mutable string		_fileName;				// name on local file system
	STime				_validityDate,			// date of validity
						_submissionDate;		// date of submission
	string				_extensionKey,			// key of type of extensions
						_extensionSequence;		// truncated extension names
};

ostream &operator<<(ostream &, const CifEntry &);

//
//	CIF : a Cif is a set of CifEntries
//
class Cif : public set<CifEntry> {};

//
//	CcfAccessLog: class used to log CCF accesses
//
class CcfAccessLog : public set<CifEntry> {
public:
	CcfAccessLog() : set<CifEntry>() { PROC ENDPROC }

	void reset() { PROC erase(begin(), end()); ENDPROC }

	~CcfAccessLog() { PROC ENDPROC }

	static CcfAccessLog &instance() { PROC RETURN _instance; }

private:
	static CcfAccessLog	_instance;
};

ostream &operator<<(ostream &, const CcfAccessLog &);

#endif
