//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	EffectiveAreaDataServer.cc --- implementation of EffectiveAreaDataServer class
//
//	DESCRIPTION:
//		This implements all methods declared in the EffectiveAreaDataServer class
//
//	CHANGE HISTORY:
//		UL, 2000-02-01: 1.0.0: - first version
//		UL, 2000-04-25: 1.1.0: - temporarily disabled instantiation of
//								 low-accuracy EffectiveAreaDataServer
//								 (numerical problems with 2-D polynomial fit
//								 from IDL)
//		UL, 2000-05-09: 1.2.0: - bug fix in
//									EffectiveAreaDataServer::vignettingCurve
//								 for RGS
//		UL, 2000-09-15: 1.2.1: - theta>0 was implicitely assumed in
//									effectiveAreaCurve()
//
//	AUTHOR:
//		U. Lammers (UL), Tue Feb  1 18:24:59 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, C. Erd, et al.
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/EffectiveAreaDataServer.cc,v 1.3 2000/12/20 16:04:51 ulammers Exp $
//

#include <cmath>
#include "CalOalUtility.h"
#include "CalServer.h"
#include "EffectiveAreaDataServer.h"
#include "GratingDataServer.h"

using namespace CalOalUtility;

//
//	---( EffectiveAreaCurve )-----------------------------------------------
//
double EffectiveAreaCurve::area(double E) const
{
	PROC
	RETURN (*this)(E);
} // EffectiveAreaCurve::area()


double EffectiveAreaCurve::area(double Elo, double Ehi) const
{
	PROC

	if (fabs(Ehi-Elo) < 1.e-8)
		ALTRETURNV(this->area(Elo));

	RETURN this->integral(Elo, Ehi)/(Ehi-Elo);
} // EffectiveAreaCurve::area()


//
//	---( EffectiveAreaDataServer )-------------------------------------------
//
EffectiveAreaDataServer::EffectiveAreaDataServer(
							CalAtomCollection<EffectiveAreaDataServer> *parent)
	: CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent)
{
	PROC
	ENDPROC
} // EffectiveAreaDataServer::EffectiveAreaDataServer()


void EffectiveAreaDataServer::dump(ostream &)
{
	PROC

	errstr<<"Sorry dumping EffectiveAreaDataServer not yet implemented"
		  <<Sas::warning("NotYetImplemented");

	ENDPROC
} // EffectiveAreaDataServer::dump()


EffectiveAreaCurve EffectiveAreaDataServer::effectiveAreaCurveFull(
															double theta,
															double phi) const
{
	PROC
	RETURN effectiveAreaCurve(theta, phi);
} //  EffectiveAreaDataServer::effectiveAreaCurveFull()


EffectiveAreaCurve EffectiveAreaDataServer::effectiveAreaCurve(double theta,
														double phi) const
{
	PROC

	errstr<<"Generic EffectiveAreaDataServer cannot provide effective area "
			"curve"<<Sas::error("MeaningLessRequest");

	RETURN effectiveAreaCurve(theta, phi);
} //  EffectiveAreaDataServer::effectiveAreaCurveFull()


EffectiveAreaCurve EffectiveAreaDataServer::effectiveAreaCurve(int,
														double theta,
														double phi) const
{
	PROC

	errstr<<"Generic EffectiveAreaDataServer cannot provide RGS-specific "
			"effective area curve"<<Sas::error("MeaninglessRequest");

	RETURN effectiveAreaCurve(theta, phi);
} // EffectiveAreaDataServer::effectiveAreaCurve()


EffectiveAreaCurve EffectiveAreaDataServer::intrinsicEffectiveAreaCurve(
														int order,
														double theta,
														double phi) const
{
	PROC
	RETURN effectiveAreaCurve(order, theta, phi);
} // EffectiveAreaDataServer::effectiveAreaCurve()


VignettingCurve EffectiveAreaDataServer::vignettingCurve(double theta,
												  double phi) const
{
	PROC

	VignettingCurve vc(effectiveAreaCurve(theta, phi));
	vc /= effectiveAreaCurve(0., 0.);

	RETURN vc;
} // EffectiveAreaDataServer::vignettingCurve()


VignettingCurve EffectiveAreaDataServer::vignettingCurve(int spectralOrder,
												  double theta,
												  double phi) const
{
	PROC

	VignettingCurve vc(effectiveAreaCurve(spectralOrder, theta, phi));
	vc /= effectiveAreaCurve(spectralOrder, 0., 0.);

	RETURN vc;
} // EffectiveAreaDataServer::vignettingCurve()


EffectiveAreaDataServer::~EffectiveAreaDataServer()
{
	PROC
	ENDPROC
} // EffectiveAreaDataServer::~EffectiveAreaDataServer()


//
//	---( LowAccuracyEffectiveAreaDataServer )----------------------------------
//
LowAccuracyEffectiveAreaDataServer::LowAccuracyEffectiveAreaDataServer(
								Proxy<EffectiveAreaCurve>
									onAxisEffectiveAreaCurve,
								const vector<double> &eGrid,
								const vector<DImage> &coefs,
								double deltaF,
								double deltaY,
								double deltaZ,
							CalAtomCollection<EffectiveAreaDataServer> *parent)
	: CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent),
	  EffectiveAreaDataServer(parent),
	  _onAxisEffectiveAreaCurve(onAxisEffectiveAreaCurve),
	  _eGrid(eGrid),
	  _deltaF(deltaF), _deltaY(deltaY), _deltaZ(deltaZ)
{
	PROC

	MiscDataServer *mds = calServer.getAtom(mds);
	_focalLength = mds->get("FOCAL_LENGTH");

	//	compute coefficient image
	unsigned N = _eGrid.size();
	if (N != coefs.size())
		errstr<<"Size mismatch between energy grid and coefficient image "
				"vector"
			  <<Sas::error("SizeMismatch");
	if (N < 2)
		errstr<<"Need at least two elements in energy grid vector"
			  <<Sas::error("TooLittleDataPoints");

	unsigned jMax = coefs[0].xExtent(),
			 kMax = coefs[0].yExtent();

	_coefsCurve = Image<Slatec::CubicHermiteInterpolant<double> *>(jMax, kMax);
	for (unsigned j=0; j<jMax; ++j)
		for (unsigned k=0; k<kMax; ++k) {
			vector<double> v(N);
			for (unsigned n=0; n<N; ++n)
				v[n] = coefs[n][j][k];
			_coefsCurve[j][k] =
					new Slatec::CubicHermiteInterpolant<double>(_eGrid, v);
		}

	ENDPROC
} // LowAccuracyEffectiveAreaDataServer::LowAccuracyEffectiveAreaDataServer()


EffectiveAreaCurve LowAccuracyEffectiveAreaDataServer::effectiveAreaCurve(
										double theta,	//	[arcsec]
										double phi)		//	[rad]
const
{
	PROC

	double	c = _focalLength*(1+_deltaF/_focalLength),
			tanTheta = tan(theta/3600. * DEG_TO_RAD),
			y = c * tanTheta * cos(phi) + _deltaY,
			z = c * tanTheta * sin(phi) + _deltaZ;

	EffectiveAreaCurve ae(*_onAxisEffectiveAreaCurve);
	const vector<double> &E = ae.abscissa();
	unsigned N = E.size();
	vector<double> v(N);
	unsigned jMax = _coefsCurve.xExtent(),
			 kMax = _coefsCurve.yExtent();

	for (unsigned n=0; n<N; ++n) {
		double ypow = 1.,
			   zpow = 1.;
		v[n] = 0.;
	
		for (unsigned j=0; j<jMax; ++j) {
			for (unsigned k=0; k<kMax; ++k) {
				v[n] += (*_coefsCurve[j][k])(E[n]) * ypow * zpow;
				ypow *= y;
			}
			zpow *= z;
		}
	}

	RETURN ae*Slatec::CubicHermiteInterpolant<double>(E, v);
} // LowAccuracyDataServer::effectiveAreaCurve()


LowAccuracyEffectiveAreaDataServer::~LowAccuracyEffectiveAreaDataServer()
{
	PROC

	unsigned jMax = _coefsCurve.xExtent(),
			 kMax = _coefsCurve.yExtent();
	for (unsigned j=0; j<jMax; ++j)
		for (unsigned k=0; k<kMax; ++k)
			delete _coefsCurve[j][k];

	ENDPROC
} // LowAccuracyEffectiveAreaDataServer::~LowAccuracyEffectiveAreaDataServer()



MediumAccuracyEffectiveAreaDataServer::MediumAccuracyEffectiveAreaDataServer(
							Proxy<EffectiveAreaCurve>
								onAxisEffectiveAreaCurve,
							const Table *vignTable,
							CalAtomCollection<EffectiveAreaDataServer> *parent)
	: CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent),
	  EffectiveAreaDataServer(parent),
	  _onAxisEffectiveAreaCurve(onAxisEffectiveAreaCurve)
{
	PROC

	const ColumnData *cdE = vignTable->column("ENERGY")->data(),
				 *cdVignFac = vignTable->column("VIGNETTING_FACTOR")->data();
	real32 *E = cdE->real32Data(),
		   *vignFac = cdVignFac->real32Data();

	unsigned Ne = vignTable->rows(),
			 Ntheta = cdVignFac->elements();
	_eGrid.resize(Ne);
	vector<double> thetaGrid(Ntheta);
	double deltaTheta = vignTable->attribute("D_THETA")->asReal() // [deg]
						* 3600.; // [deg->arcsec]
	for (unsigned n=0; n<Ntheta; ++n)
		thetaGrid[n] = DBL(n) * deltaTheta;
	thetaGrid.push_back(1.e4);	//	one data point at extreme angle to avoid
								//	extrapolation artefacts

	for (unsigned ne=0; ne<Ne; ++ne) {
		_eGrid[ne] = E[ne];
		vector<double> fac(Ntheta);
		for (unsigned nt=0; nt<Ntheta; ++nt)
			fac[nt] = vignFac[cdVignFac->position(ne, nt)];
		fac.push_back(0.);	//	vignetting 0 at theta=1.e4"

		_vignetting.push_back(new Slatec::CubicHermiteInterpolant<double>
									(thetaGrid, fac));
	}

	delete cdE;
	delete cdVignFac;

#if 0
	unsigned N = eGrid.size();
	if (flatFields.size() != N)
		errstr<<"Vector with flatfields and energy grid vector have different "
				"sizes"<<Sas::error("SizeMismatch");

	//
	//	construct the interpolation maps
	//
	for (unsigned n=0; n<N; ++n) {
		const SImage &map = flatFields[n];
		int xExt = map.xExtent(),
			yExt = map.yExtent();
		vector<float> x(xExt),
					  y(yExt),
					  z(map.data(), map.data()+ xExt * yExt);
		for (int i=0; i<xExt; ++i)
			x[i] = DEG_TO_RAD*((i-INT(refPixXY)+1)*deltaXY + refValXY);
		for (int j=0; j<yExt; ++j)
			y[j] = DEG_TO_RAD*((j-INT(refPixXY)+1)*deltaXY + refValXY);

		_flatFields.push_back(
			new Slatec::Bivariate2DInterpolantRectangular<float>(x, y, z));
	}
#endif

	ENDPROC
} // MediumAccuracyEffectiveAreaDataServer::MediumAccuracyEffectiveAreaDataSercer()


EffectiveAreaCurve MediumAccuracyEffectiveAreaDataServer::effectiveAreaCurve(
											double theta,
											double) const
{
	PROC

	unsigned N = _eGrid.size();
	vector<double> vignetting(N);
	double abstheta = fabs(theta);
	for (unsigned n=0; n<N; ++n)
		vignetting[n] = (*_vignetting[n])(abstheta);

	Slatec::CubicHermiteInterpolant<double> vignFunc(_eGrid, vignetting),
										onAxisAe(*_onAxisEffectiveAreaCurve);

	RETURN onAxisAe * vignFunc;
} // MediumAccuracyEffectiveAreaDataServer::effectiveAreaCurve()


MediumAccuracyEffectiveAreaDataServer::~MediumAccuracyEffectiveAreaDataServer()
{
	PROC

	unsigned N = _vignetting.size();
	for (unsigned n=0; n<N; ++n)
		delete _vignetting[n];

	ENDPROC
} // MediumAccuracyEffectiveAreaDataServer::~MediumAccuracyEffectiveAreaDataServer()


//
//	-----( EMOS/RGS EffectiveAreaDataServer )---------------------------------
//
EmosEffectiveAreaDataServer::EmosEffectiveAreaDataServer(
						Proxy<EffectiveAreaDataServer> mds,
						CalAtomCollection<EffectiveAreaDataServer> *parent)
		: CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent),
		  EffectiveAreaDataServer(parent),
		  _unvignetted(mds)
{
	PROC
	ENDPROC
} // EmosEffectiveAreaDataServer::EmosEffectiveAreaDataServer()


EffectiveAreaCurve EmosEffectiveAreaDataServer::effectiveAreaCurveFull(
																double theta,
																double phi)
const
{
	PROC
	RETURN _unvignetted->effectiveAreaCurve(theta, phi);
} // EmosEffectiveAreaDataServer::effectiveAreaCurveFull()


EffectiveAreaCurve EmosEffectiveAreaDataServer::effectiveAreaCurve(
															double theta,
															double phi) const
{
	PROC

	GratingDataServer *gds = calServer.getAtom(gds);
	InterpolatedCurve oef = gds->obscurationEfficiency(theta, phi);

	EffectiveAreaCurve ae(_unvignetted->effectiveAreaCurve(theta, phi));
	ae *= (1. - oef);

	RETURN ae;
} // EmosEffectiveAreaDataServer::effectiveAreaCurve()


RgsEffectiveAreaDataServer::RgsEffectiveAreaDataServer(
						Proxy<EffectiveAreaDataServer> mds,
						CalAtomCollection<EffectiveAreaDataServer> *parent)
		: CalChild<CalAtomCollection<EffectiveAreaDataServer> >(parent),
		  EffectiveAreaDataServer(parent),
		  _unvignetted(mds)
{
	PROC
	ENDPROC
} // RgsEffectiveAreaDataServer::RgsEffectiveAreaDataServer()


EffectiveAreaCurve RgsEffectiveAreaDataServer::effectiveAreaCurveFull(
																double theta,
																double phi)
const
{
	PROC
	RETURN _unvignetted->effectiveAreaCurve(theta, phi);
} // RgsEffectiveAreaDataServer::effectiveAreaCurveFull()


EffectiveAreaCurve RgsEffectiveAreaDataServer::effectiveAreaCurve(
															int spectralOrder,
															double theta,
															double phi) const
{
	PROC

	EffectiveAreaCurve ae(intrinsicEffectiveAreaCurve(spectralOrder,
													  theta, phi));

	GratingDataServer *gds = calServer.getAtom(gds);
	ae *= gds->selfVignettingEfficiency(ae.abscissa(), theta, phi,
										spectralOrder);

	//
	//	[2]: missing are last two factors:
	//		CAL_rgsCrossPsf x f_xdspsel
	//		CAL_rgsLSF x f_prdersel
	//

	RETURN ae;
} // RgsEffectiveAreaDataServer::effectiveAreaCurve()


EffectiveAreaCurve RgsEffectiveAreaDataServer::intrinsicEffectiveAreaCurve(
														   int spectralOrder,
														   double theta,
														   double phi) const
{
	PROC

	EffectiveAreaCurve ae(_unvignetted->effectiveAreaCurve(theta, phi));

	GratingDataServer *gds = calServer.getAtom(gds);
	ae *= gds->interceptEfficiency(theta, phi);
	ae *= gds->reflectionEfficiency(theta, phi, spectralOrder);

	RETURN ae;
} // RgsEffectiveAreaDataServer::intrinsicEffectiveAreaCurve()



CalAtomCollection<EffectiveAreaDataServer>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	PROC

	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= 1;
	EffectiveAreaDataServer *null = 0;
	_collection = new vector<EffectiveAreaDataServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::XAreaEf);
	DataSet *set = e->open();

	//
	//	read on-axis effective area data
	//
	CcfConstituentTable ccftab(set, "ONAXISXAREAEF");
	const Table *table = ccftab.block();
	unsigned N = table->rows();
	vector<double> Evec(N),
				   onAxisAe(N);
	ColumnData *cdE = table->column("ENERGY")->data(),
			   *cdAe = table->column("AREA")->data();
	real32 *E = cdE->real32Data(),
		   *area = cdAe->real32Data();
	for (unsigned n=0; n<N; ++n) {
		Evec[n] = E[n];
		onAxisAe[n] = area[n];
	}
	delete cdE;
	delete cdAe;

	EffectiveAreaDataServer *mds;
	Proxy<EffectiveAreaCurve> ae = new EffectiveAreaCurve(Evec, onAxisAe);

	AccuracyLevel acc = MEDIUM;	//	for the moment disable low-accuracy
								//	EffectiveArea data server; numerical
								//	problems with polynomial fit from IDL
								//	= _accuracy
	switch (acc) {
		case LOW: {
			CcfConstituentTable ccftab(set, "FIT_PARAM");
			const Table *table = ccftab.block();
			const ColumnData *cdE = table->column("ENERGY")->data(),
							 *cdCoef = table->column("COEF_MAPS")->data();
			real32 *E = cdE->real32Data();
			real64 *coef = cdCoef->real64Data();

			unsigned N = table->rows();
			vector<double> eGrid(N);
			vector<DImage> coefMaps;
			vector<unsigned long> exts = cdCoef->size();
			for (unsigned n=0; n<N; ++n) {
				eGrid[n] = E[n];
				coefMaps.push_back(DImage(exts[0], exts[1]));
				double *map = coefMaps[n].data();
				for (unsigned i=0; i<exts[0]*exts[1]; ++i)
					map[i] = coef[cdCoef->position(n, i)];
			}
			delete cdE;
			delete cdCoef;

			mds = new LowAccuracyEffectiveAreaDataServer(
							ae,
						    eGrid,
						    coefMaps,
							table->attribute("DELTA_F")->asReal(),
							table->attribute("DELTA_Y")->asReal(),
							table->attribute("DELTA_Z")->asReal(),
							this);

			break;
			}
		case HIGH:
			errstr<<"High-accuracy version of EffectiveArea effective area is "
					"not available - reverting to normal accuracy"
				  <<Sas::warning("NotAvailable");
		case MEDIUM: {
			CcfConstituentTable ccftab(set, "VIGNETTING");
			mds = new MediumAccuracyEffectiveAreaDataServer(ae,
													 ccftab.block(),
													 this);
			break;
			}
		default:
			errstr<<"Unknown accuracy"<<Sas::error("ThisShouldNotHappen");
			break;
	}

	dataSetServer->close(set);

	//
	//	now instantiate right EffectiveArea data server
	//
	switch (instrument->id()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2:
			(*_collection)[0] =
				new EmosEffectiveAreaDataServer(
									Proxy<EffectiveAreaDataServer>(mds), this);
			break;
		case Xmm::RGS1:
		case Xmm::RGS2:
			(*_collection)[0] =			
				new RgsEffectiveAreaDataServer(
									Proxy<EffectiveAreaDataServer>(mds), this);
			break;
		default:
			//	for PN the Ae is the full one
			(*_collection)[0] = mds;
	}

	ENDPROC
} // CalAtomCollection<EffectiveAreaDataServer>::CalAtomCollection()



EffectiveAreaDataServer *CalAtomCollection<EffectiveAreaDataServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	PROC
	RETURN (*_collection)[0];
} // CalAtomCollection<EffectiveAreaDataServer>::select()
