//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	EncircledEnergyDataServer --- the EncircledEnergyDataServer class
//
//	DESCRIPTION:
//		A class 'EncircledEnergyDataServer' is defined which provides access
//		to the encircled energy function in the CCF.
//
//	CHANGE HISTORY:
//		UL, 2000-01-31: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers (UL), Wed Jan 31 14:59:56 MET 2001
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Data Access Handbook, C. Erd et al.
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/EncircledEnergyDataServer.h,v 1.2 2001/03/12 16:06:32 ulammers Exp $
//

#ifndef ENCIRCLED_ENERGY_DATA_SERVER_H
#define ENCIRCLED_ENERGY_DATA_SERVER_H

#include <map>

#include "Tracer.h"
#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "CalCommon.h"
#include "slatec++.h"
#include "Proxy.h"

//
//	---( EncircledEnergyFunction )---------------------------------------------
//
class EncircledEnergyFunction : public InterpolatedCurve {
public:
	double E() const { PROC RETURN _E; }
	double theta() const { PROC RETURN _theta; }
	double phi() const { PROC RETURN _phi; }
		//
		//	PURPOSE:
		//		Return characteristic parameters this EncircledEnergyFunction
		//		is valid for:
		//			E		: energy [eV]
		//			theta	: off-axis angle [arcsec]
		//			phi		: azimuth angle [rad]
		//


	virtual double operator()(double r) const {
		PROC
		RETURN InterpolatedCurve::operator()(r);
	}
		//
		//	PURPOSE:
		//		Compute fraction of PSF enclosed by circle of given radius.
		//
		//	ARGUMENTS:
		//		r	: radius of circle [arcsec]; must be >= 0 otherwise
		//			  error condition is raised
		//
		//	RETURNS:
		//		PSF fraction in circle in range [0, 1)
		//


	virtual ~EncircledEnergyFunction() { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Destroy EncircledEnergyFunction object constructed with below
		//		constructor(s)
		//

private:
	//	data
	double	_E, _theta, _phi;

	//	methods
	friend class EncircledEnergyDataServer;
	EncircledEnergyFunction(double E, double theta, double phi,
							const vector<double> &r,
							const vector<double> &ee);
		//
		//	PURPOSE:
		//		Construct EncircledEnergyFunction
		//
		//	ARGUMENTS:
		//		E/theta/phi	: energy [eV], off-axis [arcsec],
		//					  azimuth angle [rad]
		//		r			: vector with radius values [arcsec]; has to be
		//					  strictly monotonically increasing
		//		ee			: vector with EE values in range [0, 1); has to
		//				      be strictly monotonically increasing
		//
};


class EncircledEnergy {
	typedef Interval<double> ValidityInterval;
};


//
//	---( EncircledEnergyDataServer )------------------------------------------
//
class EncircledEnergyDataServer : public CalAtom,
	 public virtual CalChild<CalAtomCollection<EncircledEnergyDataServer> > {
public:
	EncircledEnergyDataServer(const Table *eeTab,
				  CalAtomCollection<EncircledEnergyDataServer> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate EncircledEnergyDataServer object
		//


	virtual void dump(ostream &);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		os:	stream to print on
		//


	static unsigned id() { PROC RETURN Ccf::XEncirEn; }
		//
		//	PURPOSE:
		//		Return ID of CalAtom
		//


	virtual const EncircledEnergy::ValidityInterval &eValidity() const;
	virtual const EncircledEnergy::ValidityInterval &thetaValidity() const;
	virtual const EncircledEnergy::ValidityInterval &phiValidity() const;
		//
		//	PURPOSE:
		//		Inquire energy/theta/phi validity limits of this
		//		EncircledEnergyDataServer
		//


	virtual EncircledEnergyFunction &encircledEnergy(double E,
													 double	theta = 0.,
													 double phi = 0.);
		//
		//	PURPOSE:
		//		Return reference to EncircledEnergyFunction instance
		//		applicable at given	input parameters; requesting
		//		EncircledEnergyFunction outside its validity
		//		limits might lead to an error condition
		//
		//	ARGUMENTS:
		//		E		: energy [eV]
		//		theta	: off-axis field angle (TELCOORD) [arcsec]
		//		phi		: azimuth field angle (TELCOORD) [rad]
		//


	virtual ~EncircledEnergyDataServer();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//

	private:
		//	data
		EncircledEnergyFunction		*_eeFunc;
		double						_E, _theta, _phi;
		mutable	EncircledEnergy::ValidityInterval
									_eVal, _thetaVal, _phiVal;

		typedef map<double, const InterpolatedCurve *, less<double> >
																MapTypeInner;
		typedef MapTypeInner::iterator							MIterInner;
		typedef map<double, MapTypeInner, less<double> >		MapTypeOuter;
		typedef MapTypeOuter::iterator							MIterOuter;
		MapTypeOuter											_eeCurves;	
		//	Energy [eV] theta["] eeCuve

		class ThetaScale {
		public:
			double _linC, _quadC, _c;
			void setTheta(double theta) { _c = 1. + _linC*theta +
													_quadC*theta*theta; }
			double operator()(double r) { return r*_c; };
		} _thetaScale;

		//	methods
		void setParameters(double E, double theta, double phi);

}; // EncircledEnergyDataServer

#endif

