//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	EnergyCorrector --- A class EnergyCorrector is defined
//
//	DESCRIPTION:
//		Instances of class EnergyCorrector can
//			a) offset correct PHA channels
//			b) gain correct offset-corrected PHA channels (PHA->PI conversion)
//
//	CHANGE HISTORY:
//		UL, 1999-05-25: 1.0.0: - first version from CalAtomsCommon
//		UL, 1999-07-09: 1.1.0: - added RgsEnergyCorrector which implements
//								 RGS-specific offset correction
//      UL, 1999-08-13: 1.2.0: - added piTokeV()
//		UL, 1999-09-13: 1.3.0: - changed piTokeV() -> piToeV
//		UL, 1999-12-16: 1.4.0: - added special EmosEnergyCorrector
//		UL, 2000-02-02: 1.5.0: - added qSplitRatio() to retrieve split ratios
//								 for RGS
//		UL, 2000-02-29: 1.6.0: - brought RGS offset I/F in line with [2]
//		UL, 2000-05-09: 1.7.0: - new EMOS gain correction (depends on pattern)
//		UL, 2001-04-03: 1.7.1: - couple of transparent changes to implement
//								 ALGOID-1 RGS offset correction
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, C. Erd et al.
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/EnergyCorrector.h,v 1.14 2001/04/03 15:04:02 ulammers Exp $
//

#ifndef ENERGY_CORRECTOR_H
#define ENERGY_CORRECTOR_H

#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "CalDataVector.h"
#include "slatec++.h"
#include "STime.h"

#include <vector>

//
//	---( EnergyCorrector )--------------------------------------------------
//
class EnergyCorrector : public CalAtom,
			  public virtual CalChild<CalAtomCollection<EnergyCorrector> > {
public:
	EnergyCorrector(Ccd ccd,
					const CcfConstituentTable &aduConv,
					CalAtomCollection<EnergyCorrector> *parent = 0);
		//
		//	PURPOSE:
		//		Construct EnergyCorrector object
		//
		//	ARGUMENTS:
		//		aduConv	: CCF table containing data for generic offset/gain
		//				  correction
		//		ccd		: CCD identifier
		//


	static unsigned id() { PROC RETURN Ccf::ADUConv; }
		//
		//	PURPOSE:
		//		Return a unique ID which for mapping Cal atoms onto
		//		CCF constituents
		//


	virtual void dump(ostream &os);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	float offsetValue(Node node) const;
		//
		//	RETURN:
		//		Fixed offset value for node <node>
		//


	virtual CalReal32Vector &offsetCorrect(const CalReal64Vector &time,
										   const CalReal32Vector &eIn,
										   CalReal32Vector &eOut) const;
	virtual CalReal32Vector &offsetCorrect(const CalReal64Vector &time,
										   const CalInt8Vector &nodes,
										   const CalReal32Vector &eIn,
										   CalReal32Vector &eOut) const;
		//
		//	PURPOSE:
		//		Correct a vector of PHA channels for offset (time dependency
		//		currently ignored):
		//			eOut = eIn - offset
		//
		//	ARGUMENTS:
		//		In:
		//			time	: vector with time tags in TIMCOORD frame
		//					  (currently ignored)
		//			eIn		: vector with PHA channel value to correct
		//			nodes	: node vector with nodes identifiers
		//					  in the range [0, MAX_NODES] signifying the
		//					  readout node which gave the pha channel
		//					  value passed via `eIn'; if omitted by using
		//					  the first form, the node is assumed to be
		//					  fixed and taken from the corresponding CAL
		//					  state variable
		//		Out:
		//			eOut	: offset corrected PHA channel vector
		//


	virtual CalReal32Vector &gainCorrect(const CalReal64Vector &time,
										 const CalReal32Vector &eIn,
										 CalReal32Vector &eOut) const;
	virtual CalReal32Vector &gainCorrect(const CalReal64Vector &time,
										 const CalInt8Vector &nodes,
										 const CalReal32Vector &eIn,
										 CalReal32Vector &eOut) const;
		//
		//	PURPOSE:
		//		Convert offset corrected PHA channels to PI, i.e., apply
		//		(time-dependent) gain correction
		//
		//	ARGUMENTS:
		//		In:
		//			time	: vector with time tags in TIMCOORD frame
		//					  (currently ignored)
		//			nodes	: see offsetCorrect() above
		//			eIn		: vector with pha channels to correct
		//		Out:
		//			eOut	: vector with gain corrected PI channels
		//
		//	RETURNS:
		//		vector with gain corrected PI channels
		//


	virtual CalReal32Vector &piToeV(const CalReal32Vector &eIn,
									CalReal32Vector &eOut) const;
		//
		//	PURPOSE:
        //      Convert a vector of PI channels to eV:
		//			eOut = eIn * EV_PER_PI + EV_0
		//
		//	ARGUMENTS:
		//		In:
		//			eIn		: vector with PI channel values to convert
		//		Out:
		//			eOut	: vector with eV values
		//


	virtual ~EnergyCorrector() { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Destruct EnergyCorrector object
		//


	//
	//	--- I/F for specialized versions of above generic operations
	//

	//
	//	EMOS
	//
	virtual CalReal32Vector &gainCorrect(const CalReal64Vector &,
										 const CalReal32Vector &,
										 const CalInt8Vector &,
										 CalReal32Vector &eOut) const
		{ PROC
			errstr<<"Generic energy corrector cannot perform EMOS gain "
					"correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
		  RETURN eOut;
		}

	virtual vector<float> e1Coefficients() const
		{ PROC RETURN vector<float>(); }
	virtual vector<float> e2Coefficients() const
		{ PROC RETURN vector<float>(); }
	virtual vector<float> e3Coefficients() const
		{ PROC RETURN vector<float>(); }
	virtual vector<float> e4Coefficients() const
		{ PROC RETURN vector<float>(); }


	//
	//	EPN
	//
	virtual CalReal32Vector &gainCorrect(const CalReal64Vector &,
										 const CalInt16Vector &,
										 const CalInt16Vector &,
										 const CalReal32Vector &,
										 CalReal32Vector &eOut) const
		{
			PROC
			errstr<<"Generic energy corrector does not implement EPN-specific "
					"gain correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
			RETURN eOut;
		}

	virtual double reEmissionThreshold(double, int16) const
		{
			PROC
			errstr<<"Generic energy corrector does not implement re-emission "
					"threshold algorithm"
				  <<Sas::error("UnsuitableEnergyCorrector");
			RETURN 0.;
		}


	//
	//	RGS
	//
	virtual CalReal32Vector &offsetCorrect(const CalReal64Vector &, //time
										   const CalInt16Vector &,	//rawX
										   const CalInt16Vector &,	//rawY
										   const CalInt8Vector &,	//shape
										   const CalInt16Vector &,	//eIn
										   CalReal32Vector &eOut) const
		{ PROC
			errstr<<"Generic energy corrector cannot perform RGS offset "
					"correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
		  RETURN eOut;
		}

	virtual CalReal32Vector &offsetCorrect(const CalReal64Vector &,	//time
										   const CalInt16Vector &,	//rawX
										   const CalInt16Vector &,	//rawY
										   const CalInt8Vector &,	//shape
										   const CalInt8Vector &,	//nodes
										   const CalInt16Vector &,	//eIn
										   CalReal32Vector &eOut) const
		{ PROC
			errstr<<"Generic energy corrector cannot perform RGS offset "
					"correction"
				  <<Sas::error("UnsuitableEnergyCorrector");
		  RETURN eOut;
		}


	virtual double rejectionThreshold(unsigned) const
		{ PROC
			errstr<<"Generic energy corrector has no knowledge about "
					"rejection/acceptance thresholds"
				  <<Sas::error("UnsuitableEnergyCorrector");
		  RETURN 0.;
		}
	virtual double rejectionThreshold(unsigned, Node) const
		{ PROC RETURN rejectionThreshold(0); }

	virtual double acceptanceThreshold(unsigned) const
		{ PROC RETURN rejectionThreshold(0); }
	virtual double acceptanceThreshold(unsigned, Node) const
		{ PROC RETURN rejectionThreshold(0); }

	virtual unsigned nGrades() const { PROC RETURN 0; }
	virtual vector<float> qSplitRatios() const
		{ PROC RETURN vector<float>(); }

protected:
	//	data members
	vector<float>	_offsetConst, _offsetSlope, _gain;
    double	        _eVperPi, _eV0; //  eV per PI channel value
	STime			_offsetRefTime;
};



//
//	a version of the EnergyCorrector specialized for EMOS
// 
class EmosEnergyCorrector : public virtual EnergyCorrector {
public:
	EmosEnergyCorrector(Ccd ccd,
						const CcfConstituentTable &aduConv,
						const Table *eCombine,
						CalAtomCollection<EnergyCorrector> *parent = 0);
		//
		//	PURPOSE:
		//		Construct EmosEnergyCorrector object
		//
		//	ARGUMENTS:
		//		Ccd			: CCD number this EnergyCorrector is applicable to
		//		aduConv		: CCF AduConv table from which to initialize object
		//		eCombine	: CCF table with energy combination coefficients
		//		parent		: parent of the object
		//


	CalReal32Vector &gainCorrect(const CalReal64Vector &time,
								 const CalReal32Vector &eIn,
								 const CalInt8Vector &pattern,
								 CalReal32Vector &eOut) const;
		//
		//	PURPOSE:
		//		Perform EMOS gain correction
		//
		//	ARGUMENTS:
		//		time	: vector with time tags
		//		pattern	: vector with pattern ID [0, 31]
		//		eIn		: vector with uncorrected raw ADUs
		//		eOut	: gain corrected ADUs
		//

	vector<float> e1Coefficients() const { PROC RETURN _e1Coef; }
	vector<float> e2Coefficients() const { PROC RETURN _e2Coef; }
	vector<float> e3Coefficients() const { PROC RETURN _e3Coef; }
	vector<float> e4Coefficients() const { PROC RETURN _e4Coef; }
		//
		//	PURPOSE:
		//		Return coefficient vectors driving the energy combination
		//		algorithm
		//


	~EmosEnergyCorrector() { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Destroy object.
		//

private:
	static vector<float>	_e1Coef, _e2Coef, _e3Coef, _e4Coef;
	static bool				_coefRead;
	float					_offsetGain[3][4][2];
	float					_fastG;
};


//
//	a version of the EnergyCorrector specialized for EPN
// 
class EpnEnergyCorrector : public virtual EnergyCorrector {
public:
	EpnEnergyCorrector(Ccd ccd,
					   const CcfConstituentTable &aduConv,
					   const Table *gainHigh,
					   const Table *reEmission,
					   CalAtomCollection<EnergyCorrector> *parent = 0);
		//
		//	PURPOSE:
		//		Construct EpnEnergyCorrector object
		//
		//	ARGUMENTS:
		//		ccd		: CCD number this EnergyCorrector is applicable to
		//		aduConv	: CCF AduConv table from which to initialize object
		//		gainHigh: table with data driving high-accuracy gain algorithm
		//		reEmission: table with re-emission parameters
		//


	CalReal32Vector &gainCorrect(const CalReal64Vector &time,
								 const CalInt16Vector &rawX,
								 const CalInt16Vector &rawY,
								 const CalReal32Vector &eIn,
								 CalReal32Vector &eOut) const;
		//
		//	PURPOSE:
		//		Perform high-accuracy gain correction
		//
		//	ARGUMENTS:
		//		time	: vector with time tags
		//		rawX	: vector with rawX coordinates of events
		//		rawY	: vector with rawY coordinates of events
		//		eIn		: vector with uncorrected energies in ADU
		//		eOut	: vector with gain corrected energies in ADU
		//


	double reEmissionThreshold(double E, int16 rawY) const;
		//
		//	PURPOSE:
		//		Compute re-emission threshold value
		//
		//	ARGUMENTS:
		//		E	: energy [ADU]
		//		rawY: rawY coordinate of event
		//


	~EpnEnergyCorrector();
		//
		//	PURPOSE:
		//		Destroy object
		//

private:
	typedef vector<float *> EpnGainParameters;
	static const unsigned N_PARAMETERS = 10;
		//	this is a vector with the elements [0..9]:
		//		0    1    2    3    4     5     6   7     8   9
		//		pp0, pp1, pp2, pp3, dum1, dum2, aa, dum3, bb, dum

	static bool					_gainParamsSet;
	static EpnGainParameters	_gainParams;
	static vector<float>		_gainParamsSuppl;

	static double				_Acoef[2], _Bcoef[2], _adu_eV, _rawy0;
};


class RgsEnergyCorrector : public virtual EnergyCorrector {
public:
	RgsEnergyCorrector(Ccd ccd,
					   const CcfConstituentTable &aduConv,
					   const Table *darkFrameTable,
					   const Table *offsetOverrideTable,
					   CalAtomCollection<EnergyCorrector> *parent = 0);
		//
		//	PURPOSE:
		//		Construct RgsEnergyCorrector object
		//


	CalReal32Vector &offsetCorrect(const CalReal64Vector &time,
								   const CalInt16Vector &rawX,
								   const CalInt16Vector &rawY,
								   const CalInt8Vector &shape,
								   const CalInt16Vector &eIn,
								   CalReal32Vector &eOut) const;
	CalReal32Vector &offsetCorrect(const CalReal64Vector &time,
								   const CalInt16Vector &rawX,
								   const CalInt16Vector &rawY,
								   const CalInt8Vector &shape,
								   const CalInt8Vector &nodes,
								   const CalInt16Vector &eIn,
								   CalReal32Vector &eOut) const;
		//
		//	PURPOSE:
		//		Correct a vector of PHA channels for the effect of offset
		//
		//	ARGUMENTS:
		//		time		: vector with time tags
		//		rawX/rawY	: pixel coordinates in PIXCOORD1
		//		shape		: number of pixels combined during on-board split
		//					  event reconstruction
		//		nodes		: node vector; for the first from, the node
		//					  is assumed to be fixed and taken from the
		//					  corresponding CAL state variable
		//		eIn			: to be corrected PHA values; conversion to
		//					  floating point type is done prior to perform
		//					  the correction; this optionally takes into
		//					  account randomization; see [2] for details
		//		eOut		: corrected values to be stored in this vector
		//


	float darkCurrent(int16 rawX, int16 rawY, unsigned node) const;
		//
		//	PURPOSE:
		//		Compute dark current at pixel position (rawX/rawY) for node
		//		<node>
		//

	static const unsigned NGRADES = 4;
	unsigned nGrades() const { PROC RETURN NGRADES; }
	vector<float> qSplitRatios() const;
	double rejectionThreshold(unsigned grade) const;
	double rejectionThreshold(unsigned grade, Node node) const;
	double acceptanceThreshold(unsigned grade) const;
	double acceptanceThreshold(unsigned grade, Node node) const;
		//
		//	PURPOSE:	
		//		Return rejection/acceptance threshold values in PI channels
		//		corresponding to a specified grade/node; in the second forms
		//		the node argument overrides the setting of the corresponding
		//		state variable
		//
		//	ARGUMENTS:
		//		grade:	number of pixels the event consists of
		//					1 <= grade <= 4
		//		node :	readout node identifier
		//					0 <= node <= 1


	~RgsEnergyCorrector();
		//
		//	PURPOSE:
		//		Destroy RGS energy corrector object
		//

protected:
	//	data members
	vector<double>	_rejectionThresholds,
					_acceptanceThresholds,
					_a0, _a1, _a2, _a3,
					_b0, _b1, _b2, _b3;			//	dark frame parameters
	double			_qSplitR[NGRADES];

	bool			_hasDiagnosticOffsetData;
	double			_eRandMin, _eRandMax;
	Slatec::CubicHermiteInterpolant<double>	*_offsetCurves[2];

private:
	void flatOffsetCurve(unsigned node);
	void flatOffsetCurves();
	void randomize(const CalInt16Vector &eIn, CalReal32Vector &eOut) const;
	unsigned short shapeToGrade(unsigned short shape) const;
};

#endif
