//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	FilterDataServer.cc --- implementation of FilterDataServer class
//
//	DESCRIPTION:
//		This implements all methods declared in the FilterDataServer class
//
//	CHANGE HISTORY:
//		UL, 2000-02-01: 1.0.0: - first version
//		UL, 2000-06-30: 1.1.0: - small fix to bring in line with selectlib-4.35
//		UL, 2001-03-21: 1.2.0: - better error message if given point is outside
//								 FOV
//
//	AUTHOR:
//		U. Lammers (UL), Tue Feb  1 18:24:59 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/FilterDataServer.cc,v 1.10 2001/03/22 18:01:34 ulammers Exp $
//

#include <cmath>
#include "CalOalUtility.h"
#include "CalServer.h"
#include "FilterDataServer.h"

using namespace CalOalUtility;

//
//	---( FilterTransmissionCurve )-------------------------------------------
//
double FilterTransmissionCurve::transmission(double E) const
{
	PROC
	RETURN (*this)(E);
} // FilterTransmissionCurve::transmission()


double FilterTransmissionCurve::transmission(double Elo, double Ehi) const
{
	PROC

	if (fabs(Ehi-Elo) < 1.e-8)
		ALTRETURNV(this->transmission(Elo));

	RETURN this->integral(Elo, Ehi)/(Ehi-Elo);
} // FilterTransmissionCurve::transmission


//
//	---( FilterDataServer )--------------------------------------------------
//
FilterDataServer::FilterDataServer(
					 const vector<double> &eGrid,
					 CalAtomCollection<FilterDataServer> *parent)
	: CalChild<CalAtomCollection<FilterDataServer> >(parent),
	  _eGrid(eGrid)
{
	PROC
	ENDPROC
} // FilterDataServer::FilterDataServer()



void FilterDataServer::addTransmissionVector(const string &region,
											 const vector<double> transmission)
{
	PROC

	EvalTuple et;
	if (!Evaluator::isValidExpression(region))
		errstr<<"The region specification `"<<region<<"' is syntactically "
				"incorrect"<<Sas::error("SyntaxError");

	et.evaluator = new Evaluator(region);
	NumericSymbols syms = et.evaluator->numericIdentifiers();
	et.theta = et.phi = 0;
	if (syms.name.size() == 2) {
		if (uppercase(syms.name[0]) == "THETA" &&
			uppercase(syms.name[1]) == "PHI") {
			et.theta = syms.valuePtr[0];
			et.phi = syms.valuePtr[1];
		} else if (uppercase(syms.name[0]) == "PHI" &&
				   uppercase(syms.name[1]) == "THETA") {
			et.theta = syms.valuePtr[1];
			et.phi = syms.valuePtr[0];
		} else
			errstr<<"Unrecognized symbols "<<syms.name[0]<<'/'
				  <<syms.name[1]<<" in registration of filter transmission "
					"vector - expected THETA + PHI"
				  <<Sas::error("UnknownSymbols");
	} else if (syms.name.size() != 0)
		errstr<<"Number of symbols in REGION specification `"
			  <<region<<"' must be two: THETA + PHI"
			  <<Sas::error("IllegalNumberOfSymbols");

	_transmissionData.push_back(
			Pair(et, new FilterTransmissionCurve(_eGrid, transmission)));

	ENDPROC
} // FilterDataServer::addTransmissionVector()



const FilterTransmissionCurve &FilterDataServer::at(double theta,
													double phi) const
{
	PROC

	//	search for first region applicable to theta/phi
	unsigned N = _transmissionData.size();
	for (unsigned n=0; n<N; ++n) {
		const EvalTuple &et = _transmissionData[n].first;
		if (et.theta && et.phi) {
			*et.theta = theta;
			*et.phi = phi;
		}
		if (et.evaluator->evalLogic())
			//	yup - found it
			ALTRETURNV(*_transmissionData[n].second);
	}

	errstr<<"Could not find applicable region for theta/phi="
		  <<theta<<'/'<<phi<<" in filter transmission data retrieval - the "
			"given point is probably outside the FOV"
		  <<Sas::error("NoData");

	RETURN *_transmissionData[0].second;
} // FilterDataServer::at()


FilterDataServer::~FilterDataServer()
{
	PROC

	for (unsigned n=0; n<_transmissionData.size(); ++n) {
		delete _transmissionData[n].first.evaluator;
		delete _transmissionData[n].second;
	}

	ENDPROC
} // FilterDataServer::~FilterDataServer()


CalAtomCollection<FilterDataServer>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	PROC

	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= 1;
	FilterDataServer *null = 0;
	_collection = new vector<FilterDataServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::FilterTransX);
	DataSet *dataSet = e->open();

	//	read energy grid
	CcfConstituentTable ccftab(dataSet, "EBINS");
	const Table *table = ccftab.block();
	ColumnData *cde = table->column("ENERGY")->data();
	real32 *energy = cde->real32Data();
	vector<double>	eGrid(energy, energy+cde->elements());
	delete cde;

	FilterDataServer *&fds = (*_collection)[0];
	fds = new FilterDataServer(eGrid, this);

	//	read transmission vector
	string filterString = uppercase(Xmm::toString(filter)),
		   tableName = "FILTER-" + filterString;

	CcfConstituentTable ccftab2(dataSet, tableName);
	table = ccftab2.block();
	ColumnData *cdr = table->column("REGION")->data(),
			   *cdtr = table->column("TRANSMISSION")->data();

	real32 *tr = cdtr->real32Data();
	unsigned N = table->rows();
	for (unsigned n=0; n<N; ++n) {
		vector<double> trans(eGrid.size());
			unsigned nElements = cdtr->elements();
			if (nElements != eGrid.size())
				errstr<<"Size of transmission data vector ("
					  <<nElements<<") for filter `"<<filterString
					  <<"' in "<<table->qualifiedName()
					  <<" is not identical to size of energy grid ("
					  <<eGrid.size()<<")."
					  <<Sas::error("VectorLengthMismatch");
			for (unsigned i=0; i<nElements; ++i)
				trans[i] = tr[cdtr->position(n, i)];

		fds->addTransmissionVector(stringColumnRow(cdr, n), trans);
		

	}

	delete cdtr;
	delete cdr;

	dataSetServer->close(dataSet);

	ENDPROC
} // CalAtomCollection<FilterDataServer>::CalAtomCollection()



FilterDataServer *CalAtomCollection<FilterDataServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	PROC
	RETURN (*_collection)[0];
} // CalAtomCollection<FilterDataServer>::select()

