//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	FilterDataServer --- the FilterDataServer class
//
//	DESCRIPTION:
//		A class 'FilterDataServer' is defined which provides access to
//		filter transmission data for the EPIC MOS/PN filters.
//
//	CHANGE HISTORY:
//		UL, 2000-02-01: 1.0.0: - first version
//
//	AUTHOR:
//		U. Lammers (UL), Tue Feb  1 17:20:27 MET 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Data Access Handbool, C. Erd et al.
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/FilterDataServer.h,v 1.3 2000/03/08 12:03:50 ulammers Exp $
//

#ifndef FILTERDATASERVER_H
#define FILTERDATASERVER_H

#include "Tracer.h"
#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "slatec++.h"
#include "Evaluator.h"

//
//	---( FilterTransmissionCurve )-------------------------------------------
//
class FilterTransmissionCurve : public Slatec::CubicHermiteInterpolant<double>{
public:
	FilterTransmissionCurve(const vector<double> &eGrid,
							const vector<double> &trans)
		: Slatec::CubicHermiteInterpolant<double>(eGrid, trans)
			{ PROC ENDPROC }

	virtual double transmission(double E) const;
		//
		//	PURPOSE:
		//		Evaluate filter transmission curve at a single energy <E>
		//
		//	ARGUMENTS:
		//		E	: energy in eV
		//
		//	RETURNS:
		//		Value of filter transmission curve - unitless number in
		//		range [0, 1]
		//


	virtual double transmission(double Elo, double Ehi) const;
		//
		//	PURPOSE:
		//		Return weighted integral of filter transmission curve between
		//		the	two energies
		//
		//	ARGUMENTS:
		//		Elo, Ehi	: energies in eV between which filter transmission
		//					  is to be integrated; Elo < Elo
		//
		//	RETURNS:
		//						 Ehi
		//                 1    /
		//			F = ------- |   dE f(E)
		//              Ehi-Elo /
		//						Elo
		//

	~FilterTransmissionCurve() { PROC ENDPROC }
};

//
//	---( FilterDataServer )----------------------------------------
//
class FilterDataServer : public CalAtom,
	 public virtual CalChild<CalAtomCollection<FilterDataServer> > {
public:
	FilterDataServer(const vector<double> &eGrid,
			  CalAtomCollection<FilterDataServer> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate FilterDataServer object from energy grid vector
		//		and corresponding transmission data
		//


	void addTransmissionVector(const string &region,
							   const vector<double> transmission);
		//
		//	PURPOSE:
		//		Make new transmission vector know to object
		//
		//	ARGUMENTS:
		//		region	: selectlib compliant spatial region expression
		//				  at which transmission data is applicable, e.g.
		//					(THETA*cos(PHI), THETA*sin(PHI)) in
		//						circle(0, 0, 30)
		//				  THETA is the off-axis angle in arcsecs, PHI is
		//				  azimuth angle in radians
		//		transmission:
		//				  vector with filter transmission values in the
		//				  range [0, 1] applicable in the region specified
		//				  via <region>
		//


	virtual void dump(ostream &) { PROC ENDPROC }
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	static unsigned id() { PROC RETURN Ccf::FilterTransX; }
		//
		//	PURPOSE:
		//		Return ID of CalAtom
		//


	virtual const FilterTransmissionCurve &at(double theta = 0.,
											  double phi = 0.) const;
		//
		//	PURPOSE:
		//		Return filter transmission curve applicable at detector
		//		location (theta, phi)
		//
		//	ARGUMENTS:
		//		theta/phi	: detector location in TELCOORD frame at which
		//					  transmission curve is requested
		//


	virtual ~FilterDataServer();
		//
		//	PURPOSE:
		//		Destroy object constructed with above constructor(s).
		//

private:
	//	data members
	struct EvalTuple {
		const Evaluator	*evaluator;
		double			*theta,
						*phi;
	};
	typedef pair<EvalTuple, const FilterTransmissionCurve *> Pair;
	vector<Pair>	_transmissionData;

	vector<double>	_eGrid;
}; // FilterDataServer

#endif
