//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	HKvalRangesServer.cc --- implementation of HKvalRangesServer class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-10-19: 1.0.0: - first incarnation
//
//	AUTHOR:
//		U. Lammers, Tue Oct 19 12:12:00 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/HKvalRangesServer.cc,v 1.6 2000/12/20 16:04:51 ulammers Exp $
//

#include <algorithm>

#include "stringutil.h"
#include "CalOalUtility.h"
#include "HKvalRangesServer.h"

using namespace CalOalUtility;

//
//	---( HKvalRanges )-------------------------------------------------------
//
HKvalRanges::MapType::const_iterator HKvalRanges::has(const string &key) const
{
	PROC
	RETURN find(uppercase(key));
} // HKvalRanges::has()


string HKvalRanges::get(MapType::const_iterator where,
						const string &key) const
{
	PROC

	if (where != end())
		ALTRETURNV ((*where).second);

	errstr<<"There is no validity range associated with parameter `"
		  <<key<<"'"
		  <<Sas::error("NoSuchData");

	RETURN "";
} // HKvalRanges::get()


string HKvalRanges::get(const string &key) const
{
	PROC
	RETURN get(find(uppercase(key)), key);
} // HKvalRanges::get()


string HKvalRanges::getExpanded(MapType::const_iterator where,
								const string &key) const
{
	PROC
	RETURN expand(get(where, key), (*where).first);
} // HKvalRanges::getExpanded()


string HKvalRanges::getExpanded(const string &key) const
{
	PROC
	RETURN expand(get(key), key);
} // HKvalRanges::getExpanded()


string HKvalRanges::expand(const string &expr, const string &key) const
{
	PROC

	string s(expr);
	string::size_type p;
	while ((p = s.find('@')) != string::npos)
		s.replace(p, 1, key);

	RETURN s;
} // HKvalRanges::expand()


HKvalRanges HKvalRanges::reduceWithShortNames(const vector<string> &
																shortHKnames)
{
	PROC

	HKvalRanges rlist;
	for (MIter i = begin(); i != end(); ++i) {
		string parName = (*i).first;
		MCIter j = _translationMap.find(parName);
		if (j != _translationMap.end())
			//	yes, a long name for which there is a short name replacement
			parName = (*j).second;

		//	see if the name occurs in the the list of HK column names
		//	if yes it has to be copied in the the new list
		vector<string>::const_iterator p = ::find(shortHKnames.begin(),
												  shortHKnames.end(),
												  parName);
		if (p != shortHKnames.end()) {
			//	yup, found one
			rlist[parName] = (*i).second;
			erase(i);
		}
	}

	RETURN rlist;
} // HKvalRanges::reduceWithShortNames()


string HKvalRanges::selectionExpression(bool translateLongToShort) const
{
	PROC

	string s;
	MapType::const_iterator i = begin();
	while (i != end()) {
		s += '(' + getExpanded(i++) + ')';
		if (i != end())
			s += "&&";
	}

	//	translate long names into short names
	if (translateLongToShort)
		for (MapType::const_iterator i = _translationMap.begin();
			 i != _translationMap.end(); ++i) {
			string::size_type p;
			while ((p = s.find((*i).first)) != string::npos)
				s.replace(p, (*i).first.length(), (*i).second);
		}

	RETURN s;
} // HKvalRanges::selectionExpression()


//
//	---( HKvalRangesServer )---------------------------------------------------
//
CalAtomCollection<HKvalRangesServer>::CalAtomCollection(
					const XmmInstrument *instrument,
					const Ccf *ccf,
					Xmm::Filter filter,
					AccuracyLevel accuracy)
{
	PROC

	_filter		= filter;
	_accuracy	= accuracy;
	_collection	= new vector<HKvalRangesServer *>;
	_collection->push_back(new HKvalRangesServer);
	HKvalRangesServer &hkval = *(*_collection)[0];

	CifEntry *e = ccf->entry(instrument, Ccf::HKParmInt);
	DataSet *dataSet = e->open();
	CcfConstituentTable ccftab(dataSet,
							e->queryBlockName(Ccf::toString(Ccf::HKParmInt)));
	const Table *table = ccftab.block();

	errstr<<"Reading HK validity ranges from file "<<e->fileName()
		  <<Sas::message(Msg::Lib, Msg::Verbose);


	//	read all parameters corresponding to the current instrument, the
	//	relevant telescope and the satellite
	ColumnData *parmName = table->column("PARM_NAME")->data(),
			   *parmRange = table->column("PARM_RANGE")->data();
	for (unsigned n=0; n<table->rows(); ++n) {
		string name = uppercase(stringColumnRow(parmName, n)),
			   range = uppercase(stringColumnRow(parmRange, n));
		hkval[name] = range;

		if (errHandler.verbosity(Msg::Lib) >= Msg::Verbose)
			errstr<<name<<" : "<<range<<endl
				  <<Sas::message(Msg::Lib, Msg::Verbose);
	}

	delete parmName;
	delete parmRange;

	//
	//	read name translation table if present
	//
	if (dataSet->hasBlock("TRANSLATION")) {
		CcfConstituentTable ccftab(dataSet, "TRANSLATION");
		const Table *table = ccftab.block();

		ColumnData *shortName = table->column("SHORT_NAME")->data(),
				   *longName = table->column("LONG_NAME")->data();

		HKvalRanges::MapType map;
		for (unsigned n=0; n<table->rows(); ++n) {
			string shortHK = uppercase(stringColumnRow(shortName, n)),
				   longHK = uppercase(stringColumnRow(longName, n));

			map[longHK] = shortHK;

			if (errHandler.verbosity(Msg::Lib) >= Msg::Verbose)
				errstr<<shortHK<<" => "<<longHK<<endl
					  <<Sas::message(Msg::Lib, Msg::Verbose);
		}
		hkval.translationMap(map);
	}

	dataSetServer->close(dataSet);

	ENDPROC
} // CalAtomCollection<HKvalRangesServer>::CalAtomCollection()


void HKvalRangesServer::dump(ostream &os)
{
	PROC

	os<<"HKvalRangesServer{\n";
	for (const_iterator i = begin(); i != end(); ++i)
		os<<(*i).first<<"="<<(*i).second<<endl;
	os<<"}";

	ENDPROC
} // HKvalRangesServer::dump()


HKvalRangesServer *CalAtomCollection<HKvalRangesServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	PROC
	RETURN (*_collection)[0];
} // HKvalRangesCollection<HKvalRangesServer>::select()


HKvalRanges &HKvalRangesServer::valRanges()
{
	PROC
	RETURN *this;
} // HKvalRangesCollection::valRanges;

