//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	PsfDataServer.cc --- implementation of PsfDataServer class
//
//	DESCRIPTION:
//		This implements all methods declared in the PsfDataServer class
//
//	CHANGE HISTORY:
//		UL, 2000-09-05: 1.0.0: - first version
//		UL, 2000-09-25: 1.1.0: - bug fix: _xOffset/_yOffset were not
//								 initialized
//		UL, 2000-10-17: 1.2.0: - added yet another flavor of
//									XmmPsf::enboxedEnergy()
//								 that simply integrates Psf in given window
//							   - all enboxed/encircledEnergy() variants now
//								 make use of spatial origin set via
//								 atOrigin()
//		UL, 2000-10-19: 1.3.0: - bug fix in XmmPsf::integratedPsfImage()
//								 (image was not normalized)
//		UL, 2000-11-22: 1.4.0: - added dispersionFigureDistribution()
//		UL, 2000-12-15: 1.5.0: - fixed failing reading of PSF maps when
//								 BETAPSF table is present
//		UL, 2000-12-20: 1.6.0: - dispersionFigureDistribution: removed
//								 constraint: 0<=inChannelOffset<1
//		UL, 2001-01-31: 1.7.0: - changed integration limits in
//								 dispersionFigureDistribution back to
//								 [n-.5*delta, n+.5*delta] (they were
//							     [n, n+delta] in cal>=3.92)
//		UL, 2001-05-11: 1.8.0: - fixed SPR 2027: Normalization of dispersed
//								 PSF distribution wrong
//		UL, 2001-05-15: 1.8.1: - commented out accidentally left over
//								 diagnostic output statements
//
//	AUTHOR:
//		U. Lammers (UL), Wed Sep  6 11:14:59 MET DST 2000
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.,
//		[2]: Calibration Access and Data Handbook, C. Erd, et al.
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/PsfDataServer.cc,v 1.13 2001/05/15 07:37:56 ulammers Exp $
//

#include <cmath>
#include <algorithm>
#include "CalOalUtility.h"
#include "CalServer.h"
#include "PsfDataServer.h"

using namespace CalOalUtility;

//
//	---( PSF )----------------------------------------------------------
//
Psf::Psf() : _E(0.), _theta(0.), _phi(0.),
			 _eVal(ValidityInterval(100., 15000.)),
			 _thetaVal(ValidityInterval(0., 900.)),
			 _phiVal(ValidityInterval(0., M_TWO_PI))
{
	PROC
	ENDPROC
} // Psf::Psf()


const Psf &Psf::setParameters(double E, double theta, double phi)
{
	PROC

	_E = E;
	if (E<_eVal.lo() || E>_eVal.hi())
		errstr<<"Request PSF at energy E="<<E<<" eV which is outside the "
				"validity range ["<<_eVal.lo()<<", "<<_eVal.hi()<<"]"
			  <<Sas::warning("EnergyOutsideValidityRange");

	_theta = theta/3600. * DEG_TO_RAD;
	if (theta<_thetaVal.lo() || theta>_thetaVal.hi())
		errstr<<"Request PSF at off-angle theta="<<theta<<" arcsec which is "
				"outside the validity range ["<<_thetaVal.lo()<<", "
			  <<_thetaVal.hi()<<"]"
			  <<Sas::warning("OffaxisAngleOutsideValidityRange");

	_phi = fmod(phi, M_TWO_PI);
	if (_phi < 0.)
		_phi += M_TWO_PI;
	if (_phi<_phiVal.lo() || phi>_phiVal.hi())
		errstr<<"Request PSF at azimuth-angle phi="<<phi*RAD_TO_DEG<<" deg "
				"which is outside the validity range ["
			  <<_phiVal.lo()*RAD_TO_DEG<<", "<<_phiVal.hi()*RAD_TO_DEG<<"]"
			  <<Sas::warning("AzimuthAngleOutsideValidityRange");

	RETURN *this;
} // Psf::setParameters()


Psf::BinSize Psf::binSize() const
{
	PROC
	RETURN _binSize;
} // Psf::binSize()


const Psf &Psf::binSize(const Psf::BinSize &binSize) const
{
	PROC

	if (binSizeChangeable())
		_binSize = binSize;
	else
		errstr<<"Bin size of this PSF implementation is not changeable"
			  <<Sas::error("BinSizeUnchangeable");

	RETURN *this;
} // Psf::binSize()


Psf::ValidityInterval Psf::eValidity() const { PROC RETURN _eVal; }
Psf::ValidityInterval Psf::thetaValidity() const { PROC RETURN _thetaVal; }
Psf::ValidityInterval Psf::phiValidity() const { PROC RETURN _phiVal; }

const Psf &Psf::atOrigin(double x, double y) const
{
	PROC

	_xOffset = x;
	_yOffset = y;

	RETURN *this;
} // Psf::atOrigin()

template<class T>
static T value(const Block *blk, const string attName, T dftVal)
{
	PROC

	if (blk->hasAttribute(attName))
		ALTRETURNV(blk->attribute(attName)->asReal());

	RETURN dftVal;
} // value()

//
//	XmmPsf
//
XmmPsf::XmmPsf(const Table *alignmentParamTable,
			   const Block *binningParameterBlock) : Psf()
{
	PROC

    MiscDataServer *mds = calServer.getAtom(mds);
    _F = mds->get("FOCAL_LENGTH");
	_px = binningParameterBlock->attribute("Y_PIXSZ")->asReal();
	_py = binningParameterBlock->attribute("Z_PIXSZ")->asReal();
	_nx = binningParameterBlock->attribute("Y_EXTENT")->asReal();
	_ny = binningParameterBlock->attribute("Z_EXTENT")->asReal();

	eValidity(Psf::ValidityInterval(
				value(binningParameterBlock, "E_MIN", 100.),
				value(binningParameterBlock, "E_MAX", 15000.)));
	thetaValidity(Psf::ValidityInterval(
				value(binningParameterBlock, "THET_MIN", 0.) * 60.,
				value(binningParameterBlock, "THET_MAX", 15.) * 60.));
	phiValidity(Psf::ValidityInterval(
				value(binningParameterBlock, "PHI_MIN", 0.),
				value(binningParameterBlock, "PHI_MAX", M_TWO_PI)));

	//
	//	init PSF sampling bin size to intrinsic one
	//
	binSize(BinSize(_px, _py));

	ColumnData
		*cdTheta = alignmentParamTable->column("THETA")->data(),
		*cdDecenterY = alignmentParamTable->column("DECENTER_Y")->data(),
		*cdDecenterZ = alignmentParamTable->column("DECENTER_Z")->data(),
		*cdDefocus = alignmentParamTable->column("DEFOCUS")->data();
	real32 *theta = cdTheta->real32Data(),
		   *dY = cdDecenterY->real32Data(),
		   *dZ = cdDecenterZ->real32Data(),
		   *dF = cdDefocus->real32Data();

	unsigned N = alignmentParamTable->rows();
	vector<double> thetav(N), dYv(N), dZv(N), dFv(N);
	for (unsigned n=0; n<N; ++n) {
		thetav[n] = theta[n];
		dYv[n] = dY[n];
		dZv[n] = dZ[n];
		dFv[n] = dF[n];
	}
	delete cdTheta;
	delete cdDecenterY;
	delete cdDecenterZ;
	delete cdDefocus;

	_deltaYInter = InterpolatedCurve(thetav, dYv);
	_deltaZInter = InterpolatedCurve(thetav, dZv);
	_deltaFInter = InterpolatedCurve(thetav, dFv);

	ENDPROC
} // XmmPsf::XmmPsf()


const Psf &XmmPsf::setParameters(double E, double theta, double phi)
{
	PROC

	Psf::setParameters(E, theta, phi);

	_oneOverPx	= 1./_px;
	_oneOverPy	= 1./_py;
	_sinPhi		= sin(this->phi());
	_cosPhi		= cos(this->phi());
	_Fbar		= (_F + _deltaFInter(this->theta())) * tan(this->theta());
	_deltaX		= _deltaYInter(this->theta());
	_deltaY		= _deltaZInter(this->theta());

	FocalPlanePosition fp = barycenter();
	_baryX = fp.first;
	_baryY = fp.second;
	atOrigin(0., 0.);

	RETURN *this;
} // XmmPsf::setParameters()



inline pair<Image<float>, double> XmmPsf::integratedPsfImage(
												const PsfWindow &window) const
{
	PROC

	double binX = binSize().binX(),
		   binY = binSize().binY();
	int Nx = UNT((window.halfX()-.5*binX)/binX + .5),
		Ny = UNT((window.halfY()-.5*binY)/binY + .5);
	Image<float> map(2*Nx+1, 2*Ny+1);

	double tot = 0.;
	for (int nx=-Nx; nx<=Nx; ++nx)
		for (int ny=-Ny; ny<=Ny; ++ny) {
			double x = DBL(nx)*binX + _baryX + _xOffset,
				   y = DBL(ny)*binY + _baryY + _yOffset;
			map[nx+Nx][ny+Ny] = (*this)(x, y);
			tot += map[nx+Nx][ny+Ny];
		}

	tot *= binX*binY;

	SAS_DEBUG(cout<<"binX/binY: "<<binX<<'/'<<binY<<endl;)
	SAS_DEBUG(cout<<"Total integrated: "<<tot<<' '<<tot<<endl;)

	RETURN PsfPair(map, tot);
} // XmmPsf::integratedPsfImage()


XmmPsf::PsfWindow XmmPsf::enboxedEnergy(double fraction,
										double aspectRatioYoverX) const
{
	PROC

	if (fraction >= 1.)
		errstr<<"Cannot determine PSF bounding box that includes "
			  <<100.*fraction<<"% of the energy"
			  <<Sas::error("InvalidInput");

	_binX = binSize().binX();
	_binY = binSize().binY();
	_binArea = _binX * _binY;

	int	maxX = 0,
		maxY = 0;
	double totPsf = intPsf(0, 0);
	while (totPsf < fraction) {
		//
		//	expand box in X by one pixel
		//
		++maxX;
		for (int iy=-maxY; iy<=maxY; ++iy)
			totPsf += intPsf(maxX, iy) + intPsf(-maxX, iy);

		//
		//	expand box in Y by aspectRatioYoverX pixels
		//
		int newMaxY = INT(.5*(DBL(2*maxX+1) * aspectRatioYoverX - 1.) + .5);
		for (int iy=maxY+1; iy<=newMaxY; ++iy)
			for (int ix=maxX; ix>=-maxX; --ix)
				totPsf += intPsf(ix, iy) + intPsf(ix, -iy);
		maxY = newMaxY;
	}

	RETURN PsfWindow(maxX * _binX, maxY * _binY);
} // XmmPsf::encircledEnergyInBox()


double XmmPsf::encircledEnergy(double fraction) const
{
	PROC

	//	in 0'th approximation just return radius of circle with same
	//	area as corresponding square box

	RETURN 2.*enboxedEnergy(fraction).halfX()/sqrt(M_PI);
} // XmmPsf::encircledEnergy()


inline double XmmPsf::intPsf(int ix, int iy) const
{
	PROC
	RETURN _binArea * (*this)(DBL(ix)*_binX + _baryX + _xOffset,
							  DBL(iy)*_binY + _baryY + _yOffset);
} // XmmPsf::intPsf()


double XmmPsf::xCc2(double xPsf, double yPsf) const
{
	PROC

	double xb = xPsf - .5*_px*(_nx-1.),
		   yb = yPsf - .5*_py*(_ny-1.);

	RETURN (xb + _Fbar) * _cosPhi + yb * _sinPhi + _deltaX;
} // XmmPsf::xCc2()


double XmmPsf::yCc2(double xPsf, double yPsf) const
{
	PROC

	double xb = _px*(xPsf - .5*(_nx-1.)),
		   yb = _py*(yPsf - .5*(_ny-1.));

	RETURN (-xb + _Fbar) * _sinPhi + yb * _cosPhi + _deltaY;
} // XmmPsf::yCc2()


FocalPlanePosition XmmPsf::barycenter() const
{
	PROC

#if 0	
	float theta = this->theta() * RAD_TO_DEG * 3600.,
		  phi = this->phi();

	GeometryDataServer *cdf = calServer.getAtom(cdf);
	CalReal32Vector thetav(&theta, 1),
					phiv(&phi, 1);

	float baryX, baryY, baryZ;
	XmmCartesianPointList pl =
		cdf->telCoordList(thetav, phiv).toCamCoord2(&baryX, &baryY, &baryZ);
#endif

	double phi = this->phi();
	RETURN FocalPlanePosition(_Fbar*cos(phi), _Fbar*sin(phi));
} // XmmPsf::barycenter()


double XmmPsf::xPsf(double xCc2, double yCc2) const
{
	PROC

	double x = xCc2 - _deltaX,
		   y = yCc2 - _deltaY,

		   xb = x*_cosPhi - y*_sinPhi +_Fbar*(_sinPhi*_sinPhi-_cosPhi*_cosPhi);

	RETURN (xb + .5*_px*(_nx-1.))/_px;
} // XmmPsf::xPsf()


double XmmPsf::yPsf(double xCc2, double yCc2) const
{
	PROC

	double x = xCc2 - _deltaX,
		   y = yCc2 - _deltaY,
		   yb = x*_sinPhi + y*_cosPhi - 2.*_Fbar*_sinPhi*_cosPhi;

	RETURN (yb + .5*_py*(_ny-1.))/_py;
} // XmmPsf::yPsf()


//
//	CslAnalyticPsfLow
//
CslAnalyticPsfLow::CslAnalyticPsfLow(const Table *alignmentParamTable,
									 const Table *fitParamTable)
	: XmmPsf(alignmentParamTable, fitParamTable)
{
	PROC

	ColumnData *cdE = fitParamTable->column("ENERGY")->data(),
			   *cdTheta = fitParamTable->column("THETA")->data(),
			   *cdParam = fitParamTable->column("PARAMS")->data();

	real32 *E = cdE->real32Data(),
		   *theta = cdTheta->real32Data(),
		   *param = cdParam->real32Data();

	unsigned N = fitParamTable->rows();
	vector<float> Ev(E, E+N), thetav(theta, theta+N);
	vector<float>::iterator vi = unique(Ev.begin(), Ev.end());
	Ev.erase(vi, Ev.end());

	sort(thetav.begin(), thetav.end());
	vi = unique(thetav.begin(), thetav.end());
	thetav.erase(vi, thetav.end());

	vector<float> p(N);
	assert(Ev.size() * thetav.size() == p.size());
	unsigned nParam = cdParam->elements();
	for (unsigned i=0; i<nParam; ++i) {
		for (unsigned j=0; j<N; ++j)
			p[j] = param[cdParam->position(j, i)];

		_paramInter.push_back(new InterpolatedSurface(thetav, Ev, p));

		SAS_DEBUG(
			cout<<"i="<<i<<endl;
			for (unsigned ii=0; ii<Ev.size(); ++ii) {
				for (unsigned jj=0; jj<thetav.size(); ++jj)
					cout<<(*_paramInter[i])(thetav[jj], Ev[ii])<<' ';
				cout<<endl;
			});
	}

	delete cdE;
	delete cdTheta;
	delete cdParam;
	_param.resize(nParam);
	_norm = 1./px()/py();

	ENDPROC
} // CslAnalyticPsfLow::CslAnalyticPsfLow()


const Psf &CslAnalyticPsfLow::setParameters(double E, double theta, double phi)
{
	PROC

	XmmPsf::setParameters(E, theta, phi);
	unsigned N = _param.size();
	double e = this->E(),
		   t = this->theta();
	for (unsigned n=0; n<N; ++n)
		_param[n] = (*_paramInter[n])(t, e);

	if (N != 6)
		errstr<<"Expected 6 parameters to drive CslAnalyticPsfLow but there "
				"are only "<<N<<" - this is a CCF problem"
			  <<Sas::error("TooFewParameters");
	_norm = 1./(M_TWO_PI*_param[2]*_param[3]*_param[1]*px()*py());

	RETURN *this;
} // CslAnalyticPsfLow::setParameters()


double CslAnalyticPsfLow::pointValue(double x, double y) const
{
	PROC
//	RETURN _param[0] +	ignore
	RETURN _norm * _param[1] *
				exp(-.5*(CalOalUtility::SQR((x-_param[4])/_param[2]) +
						 CalOalUtility::SQR((y-_param[5])/_param[3])));
} // CslAnalyticPsfLow::pointValue()


CslAnalyticPsfLow::~CslAnalyticPsfLow()
{
	PROC

	for (unsigned n=0; n<_paramInter.size(); delete _paramInter[n++]);

	ENDPROC
} // CslAnalyticLow::~CslAnalyticPsfLow()


//
//	CslAnalyticPsfMedium
//
CslAnalyticPsfMedium::CslAnalyticPsfMedium(const Table *alignmentParamTable,
										   const Table *fitParamTable)
	: XmmPsf(alignmentParamTable, fitParamTable)
{
	PROC

	ColumnData *cdParam = fitParamTable->column("PARAMS")->data();
	real32 *param = cdParam->real32Data();
	unsigned Np = cdParam->elements();
	_Ap.resize(Np);
	_Bp.resize(Np);
	_Cp.resize(Np);

	assert(fitParamTable->rows()==3);
	for (unsigned n=0; n<Np; ++n) {
		_Ap[n] = param[cdParam->position(0, n)];
		_Bp[n] = param[cdParam->position(1, n)];
		_Cp[n] = param[cdParam->position(2, n)];
	}

	delete cdParam;

	_norm = 1./fitParamTable->attribute("NORM")->asReal()/px()/py();

	ENDPROC
} // CslAnalyticPsfMedium::CslAnalyticPsfMedium()


const Psf &CslAnalyticPsfMedium::setParameters(double E,
											   double theta, double phi)
{
	PROC

	XmmPsf::setParameters(E, theta, phi);

	RETURN *this;
} // CslAnalytic::setParameters()


double CslAnalyticPsfMedium::pointValue(double x, double y) const
{
	PROC

	double phi = atan2(x-_Ap[1], y-_Ap[2]);
	RETURN (_Ap[0] * exp(-(CalOalUtility::SQR(x-_Ap[1])/_Ap[3] +
						 CalOalUtility::SQR(y-_Ap[2])/_Ap[4])) *
					(1.+_Ap[5] * cos(3.*phi+_Ap[6])) +
		   _Bp[0] * exp(-(CalOalUtility::SQR(x-_Bp[1])/_Bp[3] +
						 CalOalUtility::SQR(y-_Bp[2])/_Bp[4])) *
					(1.+_Bp[5] * cos(3.*phi+_Bp[6])) +
		   _Cp[0] * exp(-(CalOalUtility::SQR(x-_Cp[1])/_Cp[3] +
						 CalOalUtility::SQR(y-_Cp[2])/_Cp[4])) *
					(1.+_Cp[5] * cos(16.*phi+_Cp[6])))*_norm;
} // CslAnalyticPsfMedium::pointValue()


CslAnalyticPsfMedium::~CslAnalyticPsfMedium()
{
	PROC
	ENDPROC
} // CslAnalyticMedium::~CslAnalyticPsfMedium()


//
//	---( SmallMapsBasedPsf )----------------------------------------------
//
SmallMapsBasedPsf::PsfMap::PsfMap(Array *a)
{
	PROC
	a>>*this;
	ENDPROC
} // PsfMap::PsfMap()


Array *operator>>(Array *a, SmallMapsBasedPsf::PsfMap &psfMap)
{
	PROC
	Image<float> &i = psfMap;
	a>>i;
	psfMap._id = a->blockNumber();
	psfMap._E = a->attribute("ENERGY")->asReal();
	psfMap._theta = a->attribute("THETA")->asReal() * 60.; // arcmin->arcsec

	RETURN a;
} // operator>>


SmallMapsBasedPsf::SmallMapsBasedPsf(const Table *alignmentParamTable,
									 const CifEntry *cifEntry,
									 const Array *firstSmallMap)
	: XmmPsf(alignmentParamTable, firstSmallMap)
{
	PROC

	_psfSet = cifEntry->open();
	_xExtent = firstSmallMap->size()[0];
	_yExtent = firstSmallMap->size()[1];
	unsigned fromArray = firstSmallMap->blockNumber(),
			 toArray = _psfSet->blocks();
	for (unsigned na=fromArray; na<toArray; ++na) {
		if (_psfSet->block(na)->type() == Block::TableT)
			//	skip tables
			continue;
		CcfConstituentArray ccfarr(_psfSet, na);
		const Array *a = ccfarr.block();
		if (INT(a->size()[0]) != _xExtent ||
			INT(a->size()[1]) != _yExtent)
			errstr<<"Size of array `"<<a->qualifiedName()<<"' is "
				  <<a->size()[0]<<'x'<<a->size()[1]<<" but expected "
				  <<_xExtent<<'x'<<_yExtent
				  <<Sas::error("CcfIntegrityError");

		double E = a->attribute("ENERGY")->asReal(),
			   theta = a->attribute("THETA")->asReal() * 60.; // arcmin->arcsec
		MIterOuter i = _idMap.find(E);
		if (i != _idMap.end()) {
			MIterInner j = i->second.find(theta);
			if (j != i->second.end())
				errstr<<"Have read PSF map for E/theta="<<E<<'/'<<theta
					  <<" already"<<Sas::error("CcfIntegrityError");
			i->second[theta] = na;
		} else {
			MapTypeInner innerMap;
			innerMap[theta] = na;
			_idMap[E] = innerMap;
		}
	}
	for (unsigned n=0; n<NUM_OF(_maps); _maps[n++]=0);
	eValidity(
		ValidityInterval(_idMap.begin()->first, (--_idMap.end())->first));
	thetaValidity(
		ValidityInterval(0., (--_idMap.begin()->second.end())->first));
	phiValidity(ValidityInterval(0., M_TWO_PI));

	ENDPROC
} // SmallMapsBasedPsf::SmallMapsBasedPsf()


const Psf &SmallMapsBasedPsf::setParameters(double E,
											double theta,
											double phi)
{
	PROC

	XmmPsf::setParameters(E, theta, phi);

	MIterOuter ior = _idMap.upper_bound(E);
	if (ior == _idMap.end() || ior == _idMap.begin())
		errstr<<"PSF energy can only be chosen in the interval ["
			  <<_idMap.begin()->first<<", "<<(--_idMap.end())->first
			  <<"] - requested energy is: "<<E
			  <<Sas::error("EnergyOutOfValidBounds");
	MIterOuter iol = ior; --iol;

	MIterInner iiur = ior->second.upper_bound(theta);
	if (iiur == ior->second.end() || iiur == ior->second.begin())
		errstr<<"PSF off-axis angle can only be chosen in the interval ["
			  <<ior->second.begin()->first<<", "<<(--ior->second.end())->first
			  <<"] - requested angle is: "<<theta
			  <<Sas::error("OffAxisAngleOutOfValidBounds");
	MIterInner iilr = iiur; --iilr;

	MIterInner iiul = iol->second.upper_bound(theta);
	assert(iiul != iol->second.begin() && iiul != iol->second.end());
	MIterInner iill = iiul; --iill;

	MIterInner	itervec[] = { iill, iilr, iiur, iiul };

	unsigned N = NUM_OF(itervec);
	for (unsigned n=0; n<N; ++n) {
		unsigned id = itervec[n]->second;
		if (!_maps[n])
			_maps[n] = new PsfMap(_psfSet->array(id));
		else
			if (_maps[n]->id() != id) {
				unsigned i = n+1;
				for (; i<N; ++i)
					if (_maps[i]->id() == id) {
						PsfMap *tmp = _maps[n];
						_maps[n] = _maps[i];
						_maps[i] = tmp;
						break;
					}
				if (i>=N) {
					delete _maps[n];
					_maps[n] = new PsfMap(_psfSet->array(id));
				}
			}
	}

	double E0 = _maps[0]->E(),
		   E1 = _maps[1]->E(),
		   theta0 = _maps[0]->theta(),
		   theta1 = _maps[2]->theta();

	_tE = (E-E0)/(E1-E0);
	_ut = (theta-theta0)/(theta1-theta0);

	RETURN *this;
} // SmallMapsBasedPsf::setParameters()


double SmallMapsBasedPsf::pointValue(double x, double y) const
{
	PROC

	int ix = INT(x+.5),
		iy = INT(y+.5);

	if (ix<0. || ix>=_xExtent ||
		iy<0. || iy>=_yExtent)
		ALTRETURNV(0.);
//		errstr<<"Requested PSF at point ("<<ix<<','<<iy<<") which is out "
//				"of valid bounds [0, "<<_xExtent-1<<"] x [0, "<<_yExtent-1
//			  <<"]"<<Sas::error("positionOutOfBounds");


	PsfMap &psfLL = *_maps[0],
		   &psfLR = *_maps[1],
		   &psfUR = *_maps[2],
		   &psfUL = *_maps[3];

	RETURN (1.-_tE)*(1.-_ut) * psfLL[ix][iy] +
		   _tE*(1.-_ut) * psfLR[ix][iy] +
		   _tE*_ut * psfUR[ix][iy] +
		   (1.-_tE)*_ut * psfUL[ix][iy];
} // SmallMapsBasedPsf::pointValue()


SmallMapsBasedPsf::~SmallMapsBasedPsf()
{
	PROC

	dataSetServer->close(_psfSet);
	for (unsigned n=0; n<NUM_OF(_maps); delete _maps[n++]);

	ENDPROC
} // SmallMapsBasedPsf::~SmallMapsBasedPsf()


//
//	---( PsfDataServer )--------------------------------------------------
//
PsfDataServer::PsfDataServer(Psf *psf,
							 const Table *betaPsfTab,
							 CalAtomCollection<PsfDataServer> *parent)
	: CalChild<CalAtomCollection<PsfDataServer> >(parent), _psf(psf)
{
	PROC

	_hasBetaPsf = betaPsfTab;
	if (!_hasBetaPsf) {
		if (calServer.instrument()->behindTelescope() != Xmm::XRT3)
		errstr<<"CCF::XRT?_PSF does not contain a BETAPSF table - "
				"PsfDataServer::dispersionFigureDistribution() will not work"
			  <<Sas::warning("NoSuchTable");
		ALTRETURN;
	}

	vector<double>	norm, center, width;
	vector<string>	type;
	CalOalUtility::fillVector(betaPsfTab, "TYPE", type);
	CalOalUtility::fillVector(betaPsfTab, "NORM", norm);
	CalOalUtility::fillVector(betaPsfTab, "CENTER", center);	// [rad]
	CalOalUtility::fillVector(betaPsfTab, "WIDTH", width);		// [rad]


	unsigned N = type.size();
	for (unsigned n=0; n<N; n++) {
		unsigned i = type[n][0] == 'G' ? 0 : 1;
		_norm[i].push_back(norm[n]);
		_center[i].push_back(center[n]);
		_width[i].push_back(width[n]);
	}

	ENDPROC
} // PsfDataServer::PsfDataServer()


void PsfDataServer::dump(ostream &)
{
	PROC

	errstr<<"Sorry dumping PsfDataServer not yet implemented"
		  <<Sas::warning("NotYetImplemented");

	ENDPROC
} // PsfDataServer::dump()


Psf::ValidityInterval PsfDataServer::eValidity() const
{
	PROC
	assert(_psf);
	RETURN _psf->eValidity();
} // PsfDataServer::eValidity()


Psf::ValidityInterval PsfDataServer::thetaValidity() const
{
	PROC
	assert(_psf);
	RETURN _psf->thetaValidity();
} // PsfDataServer::thetaValidity()


Psf::ValidityInterval PsfDataServer::phiValidity() const
{
	PROC
	assert(_psf);
	RETURN _psf->phiValidity();
} // PsfDataServer::phiValidity()


Psf &PsfDataServer::psf(double E, double theta, double phi)
{
	PROC

	_psf->setParameters(E, theta, phi);	

	RETURN *_psf;	
} // PsfDataServer::psf()


Spectrum PsfDataServer::dispersionFigureDistribution(
											  double nBetaHWHM,
											  double betaBinWidth,
											  double chromaticMagnification,
											  double inChannelOffset,
											  bool analytical)
{
	PROC

	if (!_hasBetaPsf)
		errstr<<"CCF does not contain BETAPSF table - cannot continue"
			  <<Sas::error("NoSuchTable");

	if (!analytical)
		errstr<<"Numerical integration of intrinsic PSF in cross-dispersion "
				"direction not implemented yet - reverting to analytical "
				"model"<<Sas::warning("NotImplementedYet");

	//	copy function parameters
	vector<double> norm[2] = { _norm[0], _norm[1] },
				   center[2] = { _center[0], _center[1] },
				   width[2] = { _width[0], _width[1] };
	//	index 0: Gaussian
	//		  1: Lorentzian

	unsigned N[2] = { norm[0].size(), norm[1].size() };
	for (unsigned n=0; n<N[0]; ++n) {
		norm[0][n] *= M_SQRT_TWO_PI * width[0][n];
	}

	//	total normalizations
	double tot[2];
	for (unsigned i=0; i<2; ++i) {
		tot[i] = 0.;
		for (unsigned n=0; n<N[i]; tot[i] += norm[i][n++]);
	}
	double totNorm = tot[0] + tot[1];

	//	calculate first moment and renormalize individual amplitudes
	double firstMoment[2];	//	in beta space
	for (unsigned i=0; i<2; ++i) {
		firstMoment[i] = 0.;
		for (unsigned n=0; n<N[i]; ++n) {
			norm[i][n] /= totNorm;
			firstMoment[i] += norm[i][n]*center[i][n];
		}
	}

	double firstMom =
				(tot[0]*firstMoment[0]+tot[1]*firstMoment[1])/(tot[0]+tot[1]);

	//	apply chromatic magnification correction to centers and sigmas
	double maxWidth = 0;
	for (unsigned i=0; i<2; ++i)
		for (unsigned n=0; n<N[i]; ++n) {
			center[i][n] = chromaticMagnification*(center[i][n]-firstMom)+
																	firstMom;
			width[i][n] *= chromaticMagnification;
			if (width[i][n] > maxWidth)
				maxWidth = width[i][n];
		}

	//	go to channel space - also adjust the centers such that the
	//	first moment lies at <inChannelOffset> within its channel
	unsigned nBetaBins = 2 * UNT(maxWidth/betaBinWidth * nBetaHWHM + .5);
	double betaBins = DBL(nBetaBins),
		   betaSpan = betaBins * betaBinWidth,
		   chanPerRad = betaBins/betaSpan,
		   firstMomChan = firstMom * chanPerRad;
	double dum,
		   deltaCenter = inChannelOffset - modf(firstMomChan, &dum);

	for (unsigned i=0; i<2; ++i)
		for (unsigned n=0; n<N[i]; ++n) {
			center[i][n] = center[i][n] * chanPerRad + deltaCenter;
			width[i][n] *= chanPerRad;
	}

	//
	//	fill the spectrum
	//
	Spectrum spec(nBetaBins, 0.);
	int iBetaBins = nBetaBins/2;
	for (int ii=1; ii>=-1; ii-=2)
		for (int i=0; i<iBetaBins; ++i) {
			int indx = ii*i;
			spec[indx+iBetaBins] = 0.;
			//	add up Gaussians
			for (unsigned n=0; n<N[0]; ++n) {
				double oneOversq2sigma = 1./(M_SQRT_2*width[0][n]),
					   argPlus = DBL(indx)+.5-center[0][n],
					   argMinus = DBL(indx)-.5-center[0][n];
				spec[indx+iBetaBins] += .5*norm[0][n] *
							fabs(Slatec::erf(oneOversq2sigma * argPlus) -
								 Slatec::erf(oneOversq2sigma * argMinus));
			}
			//	add up Lorentzians
			for (unsigned n=0; n<N[1]; ++n) {
				double normOverPi = norm[1][n]/M_PI,
					   argPlus = DBL(indx)+.5-center[1][n],
					   argMinus = DBL(indx)-.5-center[1][n];
				spec[indx+iBetaBins] += normOverPi *
											fabs(atan2(argPlus, width[1][n]) -
												 atan2(argMinus, width[1][n]));
			}
		}

	//	set first moment
	spec.firstMoment(firstMomChan+deltaCenter + .5*betaBins);

	SAS_DEBUG({
		unsigned NN = spec.size();
		double ntot = 0.;
		for (unsigned n=0; n<NN; ntot+=spec[n++]);
		cout<<NN<<' '<<ntot<<' '<<spec.firstMoment()<<endl;
	})

	RETURN spec;
} // PsfDataServer::dispersionFigureDistribution()


PsfDataServer::~PsfDataServer()
{
	PROC

	delete _psf;

	ENDPROC
} // PsfDataServer::~PsfDataServer()



CalAtomCollection<PsfDataServer>::CalAtomCollection(
							const XmmInstrument *instrument,
							const Ccf *ccf,
							Xmm::Filter filter,
							AccuracyLevel accuracy)
{
	PROC

	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= 1;
	PsfDataServer *null = 0;
	_collection = new vector<PsfDataServer *>(_nEntries, null);

	CifEntry *e = ccf->entry(instrument, Ccf::XPSF);
	DataSet *set = e->open();
	const Table *betaPsfTab = 0;
	if (set->hasBlock("BETAPSF")) {
		CcfConstituentTable ccftab(set, "BETAPSF");
		betaPsfTab = ccftab.block();
	}

	CcfConstituentTable ccftab0(set, "ALIGNMENT_PARAMS");
	switch (accuracy) {
		case LOW: {
			CcfConstituentTable ccftab1(set, "SINGLE_GAUSS_PARAMS");
			(*_collection)[0] = new PsfDataServer(
									new CslAnalyticPsfLow(
										ccftab0.block(),
										ccftab1.block()),
									betaPsfTab, this);
			break;
		}
		case MEDIUM: {
			CcfConstituentArray ccfarr1(set, 3);
			(*_collection)[0] = new PsfDataServer(
									new SmallMapsBasedPsf(
										ccftab0.block(), e,
										ccfarr1.block()), betaPsfTab, this);
			break;
		}
		case HIGH: {
			CcfConstituentTable ccftab1(set, "TRIPLE_GAUSS_PARAMS");
			(*_collection)[0] = new PsfDataServer(
									new CslAnalyticPsfMedium(
										ccftab0.block(),
										ccftab1.block()),
									betaPsfTab, this);
			break;
		}
		default:
			break;
	}

	dataSetServer->close(set);

	ENDPROC
} // CalAtomCollection<PsfDataServer>::CalAtomCollection()



PsfDataServer *CalAtomCollection<PsfDataServer>::select(
					Ccd,
					Node,
					unsigned) const
{
	PROC
	RETURN (*_collection)[0];
} // CalAtomCollection<PsfDataServer>::select()
