//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Redistribution.cc --- implementation of redistribtion classes
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-09-13: 1.0.0: - first incarnation (from CalAtomsCommon.cc)
//		UL, 1999-09-15: 1.1.0: - fixed problem in construction of RGS
//								 redistribution (convolution of noise-free
//								 spectrum with noise had the wrap-around
//								 problem - see [3], p. 429)
//		UL, 1999-10-27: 2.0.0: - reintroduced eBins()
//		UL, 1999-12-10: 2.0.1: - fixed minor problem in SAS_DEBUG statement
//								 (fixes SPR 1017)
//		UL, 2000-01-04: 2.0.2: - fixed normalization problem of total RGS
//								 redistribution function
//		GV, 2000-01-12: 2.0.3: - remove gain fudge in RgsRedist
//							   - add code to correctly normalize
//								 RgsRedist (integral is now QE)
//							   - RgsRedist does not invok spec.norm() any
//								 longer.
//		UL, 2000-02-22: 2.0.4: - bug fix in RgsRedist::absorptionLength()
//		UL, 2000-04-04: 2.1.0: - implemented EmosRedist according to [4]
//		UL, 2000-05-22: 2.2.0: - bug fix in CanonicalRedist constructor
//								 (caused SEGV when Redist+AborptionDataServer
//								  were used together)
//		UL, 2000-05-23: 2.3.0: - fixed a number of problem with the new
//								 MOS redistribution model
//							   - compared output against rmfgen code
//								 from Steve Sembay, Leceister
//		UL, 2000-05-30: 2.4.0: - added binnedPIeBounds()
//							   - EmosRedist::spectrum(E) evaluates
//								 redistribution model now on binnedPIeBounds()
//								 axis
//		UL, 2000-09-01: 2.5.0: - added support for EPN; uses redistribution
//								 code in package 'calpnalgo'
//		UL, 2000-11-01: 2.6.0: - added reading of ABSORPTION data in
//								 EPN::REDIST (if present)
//		UL, 2000-01-11: 2.7.0: - brought RgsRedist::noiseFreeSpectrum() in
//								 line with latest model update in [2]
//		UL, 2001-02-02: 2.8.0: - RgsRedist::spectrum(): return null-spectrum
//								 and print warning if input energy is
//								 outside energy axis bounds
//		UL, 2001-03-15: 2.9.0: - implemented I/F to new EMOS redistribution
//								 code in calmosalgo (activated with ALGOID=1
//								 in EMOS?_REDIST)
//		UL, 2001-04-23: 2.10.0:- set new parameter 'cmode' of CALPNALGO_resp()
//							   - bug fix in EpnRedist::spectrum(): caused
//								 erroneous message about null-spectrum
//								 being returned
//		UL, 2001-05-01: 2.11.0:- added gmax variable to CALMOSALGO_initRmfRow()
//								 call
//
//	AUTHOR:
//		U. Lammers, Mon Sep 13 18:13:08 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al.
//		[2]: Calibration Access and Data Handbook, C. Erd et al.
//		[3]: Numerical Recipes in C, W.H. Press et al.
//		[4]: MOS CCD RMF Description, S. Sembay + P. Bennie, EPIC-LUX-RE-170
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/Redistribution.cc,v 1.35 2001/05/01 12:41:22 ulammers Exp $
//

#include <cmath>
#include <cstring>
#include <numeric>

#include "CalOalUtility.h"
#include "PhysicalConstants.h"
#include "CalServer.h"
#include "AbsorptionDataServer.h"
#include "MiscDataServer.h"
#include "Redistribution.h"
#include "CalPnAlgo.h"
#include "CalMosAlgoRmf.h"
#include "Xmm.h"

#include "slatec++.h"


//
//	---( CanonicalRedist )---------------------------------------------------
//
CanonicalRedist::CanonicalRedist(unsigned channels,
								 double gain,
			 					 const Proxy<BinnedAxis> &eBounds,
								 const Proxy<BinnedAxis> &eBins,
								 const Proxy<BinnedAxis> &binnedPIeBounds,
								 CalAtomCollection<CanonicalRedist> *parent)
	: CalChild<CalAtomCollection<CanonicalRedist> >(parent),
		_eBounds(eBounds), _eBins(eBins), _binnedPIeBounds(binnedPIeBounds),
		_gain(gain), _channels(channels),
		_chipX(0), _chipY(0), _failSafe(false)
{
	PROC

	AbsorptionDataServer *ads = calServer.getAtom(ads);
	const AtomicProperties &ap = ads->atomicData("Si");
	_SiA = ap.weight();
	_SiDensity = ap.density();

	ENDPROC
} // CanonicalRedist::CanonicalRedist()


void CanonicalRedist::dump(ostream &)
{
	PROC
	ENDPROC
} // CanonicalRedist::dump()


const BinnedAxis &CanonicalRedist::eBounds() const
{
	PROC
	RETURN *_eBounds;
} // CanonicalRedist::eBins()


const BinnedAxis &CanonicalRedist::eBins() const
{
	PROC
	RETURN *_eBins;
} // CanonicalRedist::eBins()


const BinnedAxis &CanonicalRedist::binnedPIeBounds() const
{
	PROC
	RETURN *_binnedPIeBounds;
} // CanonicalRedist::binnedPIeBounds()


const CanonicalRedist &CanonicalRedist::failSafe(bool suppressError) const
{
	PROC

	_failSafe = suppressError;

	RETURN *this;
} // CanonicalRedist::failSafe()


const CanonicalRedist &CanonicalRedist::at(unsigned short chipY) const
{
	PROC

	_chipY = chipY;

	RETURN *this;
} // CanonicalRedist::at()


const CanonicalRedist &CanonicalRedist::at(unsigned short chipX,
										   unsigned short chipY) const
{
	PROC

	_chipX = chipX;
	_chipY = chipY;

	RETURN *this;
} // CanonicalRedist::at()


Spectrum CanonicalRedist::noiseFreeSpectrum(unsigned index,
											unsigned eventGrade) const
{
	PROC
	RETURN noiseFreeSpectrum(_eBins->bin(index).mid(), *_eBins, eventGrade);
} // CanonicalRedist::noiseFreeSpectrum()


Spectrum CanonicalRedist::spectrum(unsigned index,
								   unsigned eventGrade) const
{
	PROC
	RETURN spectrum(_eBins->bin(index).mid(), *_eBins, eventGrade);
} // CanonicalRedist::spectrum()


Spectrum CanonicalRedist::spectrum(double E, unsigned eventGrade) const
{
	PROC
	RETURN spectrum(E, *_binnedPIeBounds, eventGrade);
} // CanonicalRedist::spectrum()


Spectrum CanonicalRedist::noiseFreeSpectrum(double E, unsigned eventGrade)
const
{
	PROC
	RETURN noiseFreeSpectrum(E, *_eBins, eventGrade);
} // CanonicalRedist::noiseFreeSpectrum()


//
//	the following is a canonical model spectrum
//
Spectrum CanonicalRedist::spectrum(double E,
								   const BinnedAxis &eAxis,
								   unsigned) const
{
	PROC

	double fano = .165,	// Fano factor
		   epsi = 3.6,	// energy required for one e-/h pair [eV]
		   r	= 5.,	// electronic noise [e- rms]
	 	   a	= .05,	// fraction of signal in the tail
		   b	= 1.-a;	// fraction in the Gaussian

	double ne = fano * E / epsi,	// number of e- generated by X-ray
		   mean = ne * _gain,		// corresponding PHA channel number
		   sigma = sqrt(ne + r * r) * _gain,
		   A = 2./mean,
		   B = -2./mean/mean;

	Spectrum spec(eAxis.bins());
	//	initialize vector	
	spec = 0.;

	//	calculate response only around mean channel
	unsigned fromChannel = UNT(max(0., mean - 3. * sigma)),
		     toChannel = UNT(min(DBL(_channels), mean + 3. * sigma));
	for (unsigned i=fromChannel; i<toChannel; ++i) {
		double tail = i <= UNT(mean) ? tail = a * (A + B * i) : 0.,
			   c = (DBL(i) - mean)/sigma;
		spec[i] = tail + b * exp(-.5 * c * c) /	sqrt(2.*acos(-1.))/sigma;
	}

	RETURN spec;
} // CanonicalRedist::spectrum()


Spectrum CanonicalRedist::noiseFreeSpectrum(double E,
											const BinnedAxis &eAxis,
											unsigned eventGrade) const
{
	PROC
	RETURN spectrum(E, eAxis, eventGrade);
} // CanonicalRedist::noiseFreeSpectrum


double CanonicalRedist::absorptionLength(double f2, double E) const
{
	PROC

	double nSi = 1000. * _SiDensity *
					PhysicalConstants::avogadro/ _SiA,				// [1/m^3]
		   lambda = PhysicalConstants::h * PhysicalConstants::c/
						(E * PhysicalConstants::eVtoJ);		// [m]

	RETURN 1./(2. * nSi * PhysicalConstants::r0 * lambda * f2);
} // CanonicalRedist::absorptionLength()


double CanonicalRedist::absorptionLengthInSi(double E) const
{
	PROC

	AbsorptionDataServer *ads = calServer.getAtom(ads);
    const AbsCoefCurve &f2Si	= ads->henke("Si").f2();

	RETURN absorptionLength(f2Si(E), E);
} // CanonicalRedist::absorptionLengthInSi()


CanonicalRedist::~CanonicalRedist()
{
	PROC
	ENDPROC
} // CanonicalRedist::~CanonicalRedist()



CalAtomCollection<CanonicalRedist>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	PROC
	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs      = instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nEntries	= _nCCDs;
	CanonicalRedist *null = 0;
	_collection	= new vector<CanonicalRedist *>(_nEntries, null);

	//
	//	get parameters and EBOUNDS/EBINS/BINNEDPI_EBOUNDS from CCF
	//
	CifEntry *e = ccf->entry(instrument, Ccf::Redist);
	DataSet *set = e->open();

	CcfConstituentTable ccftab(set, set->hasBlock("PHA_EBOUNDS")
								? "PHA_EBOUNDS" : "EBOUNDS");
	const Table *eboundsTable = ccftab.block();
	ColumnData *cdemin = eboundsTable->column("E_MIN")->data(),
			   *cdemax = eboundsTable->column("E_MAX")->data();
	CalReal32Vector	eminV(cdemin->real32Data(), eboundsTable->rows());

	Proxy<BinnedAxis> eBounds(new NonUniformilyBinnedAxis(eminV, true,
							  cdemax->real32Data()[eboundsTable->rows()-1]));
	delete cdemin;
	delete cdemax;

	Proxy<BinnedAxis> eBins = eBounds,
					  binnedPIeBounds = eBounds;

	if (set->hasBlock("EBINS")) {
		CcfConstituentTable ccftab(set, "EBINS");
		const Table *tab = ccftab.block();
		ColumnData *cdemin = tab->column("E_MIN")->data(),
				   *cdemax = tab->column("E_MAX")->data();
		CalReal32Vector	eminV(cdemin->real32Data(), tab->rows());

		eBins =
			Proxy<BinnedAxis>(new NonUniformilyBinnedAxis(eminV, true,
							  cdemax->real32Data()[tab->rows()-1]));
		delete cdemin;
		delete cdemax;
	} else
		errstr<<"There is no EBINS table in `"<<set->qualifiedName()
			  <<"' - using EBOUNDS as EBINS"
			  <<Sas::message(Msg::Lib, Msg::Verbose);

	if (set->hasBlock("BINNEDPI_EBOUNDS")) {
		CcfConstituentTable ccftab(set, "BINNEDPI_EBOUNDS");
		const Table *tab = ccftab.block();
		ColumnData *cdemin = tab->column("E_MIN")->data(),
				   *cdemax = tab->column("E_MAX")->data();
		CalReal32Vector	eminV(cdemin->real32Data(), tab->rows());

		binnedPIeBounds =
			Proxy<BinnedAxis>(new NonUniformilyBinnedAxis(eminV, true,
							  cdemax->real32Data()[tab->rows()-1]));
		delete cdemin;
		delete cdemax;
	} else
		errstr<<"There is no BINNEDPI_EBOUNDS table in `"<<set->qualifiedName()
			  <<"' - using EBOUNDS as BINNEDPI_EBOUNDS"
			  <<Sas::message(Msg::Lib, Msg::Verbose);

	unsigned channels = instrument->eChannels();	
	for (unsigned ccd=1; ccd<=_nCCDs; ++ccd) {
		CanonicalRedist *&cr = (*_collection)[ccd-1];
		switch(instrument->id()) {
			case Xmm::EMOS1:
			case Xmm::EMOS2:
				cr = new EmosRedist(channels, 1., ccd, eBounds, eBins,
									binnedPIeBounds, set, this);
				break;
			case Xmm::EPN:
				cr = new EpnRedist(channels, 1., ccd, eBounds, eBins,
								   binnedPIeBounds, set, this);
				break;
			case Xmm::RGS1:
			case Xmm::RGS2:
				cr = new RgsRedist(channels, 1.0, ccd, eBounds, eBins,
								   binnedPIeBounds, set, this);
				break;
			default:
				errstr<<"No Redistribution information available for "
						"instrument "
					  <<Xmm::toString(instrument->id())
			  <<Sas::error("NoSuchCcfData");
						break;
		}
	}

	dataSetServer->close(set);

	ENDPROC
} // CalAtomCollection<CanonicalRmf>::CalAtomCollection()


CanonicalRedist *CalAtomCollection<CanonicalRedist>::select(Ccd ccd, Node,
													  unsigned) const
{
	PROC
	RETURN (*_collection)[ccd-1];
} // CalAtomCollection<CanonicalRedist>::select()

#if 0
	//	Ebounds
	switch(instrument->id()) {
		case Xmm::EMOS1:
		case Xmm::EMOS2:
				nChan = instrument->eChannels();
				ePerChan = 12000./3750.;
				break;
		case Xmm::EPN:
				nChan = instrument->eChannels();
				ePerChan = 16000./4000.;
				break;
		case Xmm::RGS1:
		case Xmm::RGS2:
				nChan = instrument->eChannels();
				ePerChan = 4500./3800.;
				break;
#endif





//
//	---( EmosRedist )--------------------------------------------------------
//
EmosRedist::EmosRedist(unsigned channels, double gain, Ccd ccd,
					   const Proxy<BinnedAxis> &eBounds,
					   const Proxy<BinnedAxis> &eBins,
					   const Proxy<BinnedAxis> &binnedPIeBounds,
					   const DataSet *set,
					   CalAtomCollection<CanonicalRedist> *parent)
	: CanonicalRedist(channels, gain, eBounds, eBins, binnedPIeBounds),
	  CalChild<CalAtomCollection<CanonicalRedist> >(parent),
	  _atCalled(false),
	  _nepts1(0), _grades1(0), _models1(0), _epar1(0), _param1(0)
{
	PROC

	MiscDataServer *miscData = calServer.getAtom(miscData);
	_SI_K_EDGE = miscData->get("SI_K_EDGE"),
	_SI_K_ALPHA = miscData->get("SI_FLUORESCENT_PEAK");

	CcfConstituentTable ccftab(set, "CCD_REDISTRIBUTION-" + toString(ccd),
							   1);	//	we know about ALGOID 0+1
	const Table *table = ccftab.block();
	_algoid = ccftab.algoid();

	if (_algoid == 0)
		constructForAlgo0(table);
	else if (_algoid == 1)
		constructForAlgo1(table);
	else
		errstr<<"Unknown ALGOID "<<_algoid<< "in EmosRedist construction"
			  <<Sas::error("UnknownAlgoId");

	ENDPROC
} // EmosRedist::EmosRedist()


void EmosRedist::constructForAlgo0(const Table *table)
{
	//	construct spatial grid
	_depth = LogarithmicDoubleGrid(table->attribute("DP_GSIZE")->asInt(),
								   table->attribute("DP_DEPTH")->asReal());

	//
	//	read parameter data
	//
	if (table->rows() != EVENT_TYPES)
		errstr<<"Expected "<<EVENT_TYPES<<" parameter sets but there are "
				"only "<<table->rows()<<" in `"<<table->qualifiedName()
			  <<"'"<<Sas::error("CcfDataIntegrityError");

	_models.resize(EVENT_TYPES);
	_grades.resize(EVENT_TYPES);
	unsigned nParamPerEventType = table->columns() - 2;
									//	ignore GRADE and MODEL column
	for (unsigned n=0; n<EVENT_TYPES; ++n)
		_param[n].resize(nParamPerEventType);

	vector<ColumnData *> cdp(nParamPerEventType);
	ColumnData *cdGrade = table->column("GRADE")->data(),
			   *cdModel = table->column("MODEL")->data();
	int8	*grade = cdGrade->int8Data(),
			*model = cdModel->int8Data();
	vector<real32 *>	p(nParamPerEventType);
	for (unsigned n=0; n<nParamPerEventType; ++n) {
		cdp[n] = table->column("P" + toString(n+1))->data();
		if (cdp[n]->elements() != Parameter::N_PARAM)
		errstr<<"Expected "<<Parameter::N_PARAM<<" parameters for `"
			  <<cdp[n]->parent()->qualifiedName()<<"' but saw only "
			  <<cdp[n]->elements()<<" in `"<<table->qualifiedName()
			  <<"'"<<Sas::error("CcfDataIntegrityError");
		p[n] = cdp[n]->real32Data();
	}

	_maxGrades = UNT(log(DBL(ALL_GRADES+1))/log(2.)+.5);
	for (unsigned n=0; n<EVENT_TYPES; ++n) {
		unsigned gradeVal = grade[n];
		if (gradeVal<1 || gradeVal>_maxGrades)
			errstr<<"Illegal GRADE value ("<<gradeVal<<") seen in "
				  <<table->qualifiedName()<<"'"
				  <<Sas::error("CcdDataIntegrityError");
		_grades[n] = gradeVal;
		_models[n] = model[n];
		unsigned pos = cdp[0]->position(n, 0);
		for (unsigned m=0; m<nParamPerEventType; ++m) {
			string pNameMin = "P" + toString(m+1) + "_MIN",
				   pNameMax = "P" + toString(m+1) + "_MAX";
			const Attributable *att = table->parent();
			_param[n][m] = Parameter(p[m]+pos, _SI_K_EDGE,
									 att->hasAttribute(pNameMin) ?
									 att->attribute(pNameMin)->asReal() :
									 -1.e38,
									 att->hasAttribute(pNameMax) ?
									 att->attribute(pNameMax)->asReal() :
									 +1.e38);
		}
	}

	delete cdGrade;
	delete cdModel;
	for (unsigned n=0; n<cdp.size(); delete cdp[n++]);

	ENDPROC
} // EmosRedist::constructForAlgo0()


void EmosRedist::constructForAlgo1(const Table *table)
{
	PROC

	_maxGrades = CALMOSALGO::MAXGRADE;

	_threshold = table->attribute("THRESHLD")->asReal(); // [eV]
	_backSigma = table->attribute("BACKSIGM")->asReal(); // [eV]

	vector<unsigned> eventType, grade, model;
	CalOalUtility::fillVector(table, "EVTYPE", eventType);
	CalOalUtility::fillVector(table, "GRADE", grade);
	CalOalUtility::fillVector(table, "MODEL", model);
	vector<float> E;
	CalOalUtility::fillVector(table, "ENERGY", E);

	unsigned N = eventType.size();
	assert(N>0);
	//	a few consistency checks
	if (eventType[N-1] != CALMOSALGO::NALL)
		errstr<<"Expected "<<CALMOSALGO::NALL<<" different event types in `"
			  <<table->qualifiedName()<<"' but there are "
			  <<eventType[N-1]
			  <<Sas::error("UnexpectedCcfData");
	vector<float> P;
	CalOalUtility::fillVector(table, "PARAM", 0, P);
	if (P.size() != CALMOSALGO::NPARAM)
		errstr<<"Expected to find "<<CALMOSALGO::NPARAM<<" parameters in `"
			  <<table->qualifiedName()<<"' but there are "
			  <<P.size()
			  <<Sas::error("UnexpectedCcfData");

	unsigned nE = N/eventType[N-1];
	if (nE > CALMOSALGO::NEMAX)
		errstr<<"Expected not more than "<<CALMOSALGO::NEMAX<<" different "
				"energies in `"
			  <<table->qualifiedName()<<"' but there are "
			  <<nE
			  <<Sas::error("UnexpectedCcfData");

	//	nullify vectors
	_nepts1 = new int[CALMOSALGO::NEMAX];
	memset(_nepts1, 0, sizeof(int)*CALMOSALGO::NEMAX);
	_grades1 = new int[CALMOSALGO::NALL];
	memset(_grades1, 0, sizeof(int)*CALMOSALGO::NALL);
	_models1 = new int[CALMOSALGO::NALL];
	memset(_models1, 0, sizeof(int)*CALMOSALGO::NALL);
	_epar1 = new float[CALMOSALGO::NEMAX];
	memset(_epar1, 0, sizeof(int)*CALMOSALGO::NEMAX);
	unsigned NP = CALMOSALGO::NALL*CALMOSALGO::NEMAX*CALMOSALGO::NPARAM;
	_param1 = new float[NP];
	//	note: reverse order w.r.t to Fortran declaration
	memset(_param1, 0, NP*sizeof(float));

	//
	//	assign energy data
	//
	for (unsigned n=0; n<nE; ++n) {
		_nepts1[n] = nE;
		_epar1[n] = E[n];
	}

	//
	//	read and assign parameters data
	//
	for (unsigned et=0; et<CALMOSALGO::NALL; ++et) {
		_grades1[et] = grade[et*nE];
		_models1[et] = model[et*nE];
		for(unsigned i=0; i<nE; ++i) {
			_epar1[i] = E[i];	//	yes, ugly
			CalOalUtility::fillVector(table, "PARAM", et*nE+i, P);
			for (unsigned k=0; k<CALMOSALGO::NPARAM; ++k) {
				unsigned indx = CALMOSALGO::NPARAM*CALMOSALGO::NEMAX*et
								+ i*CALMOSALGO::NPARAM + k;
				_param1[indx] = P[k];
			}
		}
	}

	ENDPROC
} // EmosRedist::constructForAlgo1()


void EmosRedist::checkAt() const
{
    PROC

    if (!_atCalled) {
        unsigned chipX = calServer.instrument()->maxX()/2,
                 chipY = calServer.instrument()->maxY()/2;
        at(chipX, chipY);
        errstr<<"No chip location for redistribuion spectum specified - "
				"will retrieve data applicable at center ("
			  <<chipX<<'/'<<chipY<<")"
              <<Sas::message(Msg::Lib, Msg::Verbose);
        _atCalled = false;
    }

    ENDPROC
} // EmosRedist::checkAt()


const CanonicalRedist &EmosRedist::at(unsigned short chipX,
									  unsigned short chipY) const
{
	PROC

	_atCalled = true;
	CanonicalRedist::at(chipX, chipY);

	//	(re-)initialize _gradeProbablities vector
	CcdQuantumDataServer *qds = calServer.getAtom(qds);
	_gradeProbabilities.resize(0);
	for (unsigned gradeNr=1; gradeNr<=_maxGrades; ++gradeNr)
		_gradeProbabilities.push_back(qds->grade(gradeNr));

	RETURN *this;
} // EmosRedist::at()


//
//	the parameterization of the EMOS RMF after D. Lumb, 1999-03-15
//
//	2000-03-30: replaced by new model of [3]
//
#if 0
Spectrum EmosRedist::spectrum(double E, unsigned pixels) const
{
	PROC

	MiscDataServer *miscData = calServer.getAtom(miscData);
	double SI_K_EDGE = miscData->get("SI_K_EDGE"),
		   SI_FLUORESCENT_PEAK = miscData->get("SI_FLUORESCENT_PEAK");

	if (E < 0.)
		errstr<<"Cannot construct spectrum for negative energy "<<E
			  <<Sas::error("InvalidInput");

	double E2 = E * E,
		   E3 = E2 * E,
		   E4 = E2 * E2,
		   log10E = log10(E);
	const double ln10 = log(10.);

	Spectrum spec(_channels);
	//	initialize vector
	spec = 0.;

	//	fitting must be done piecewise at either side of the Si K edge
	if (E < SI_K_EDGE) {
		//	calculate the 3 coefficients for the exponential fit part
		double exp1 = exp(ln10 * (_param[0] + _param[1] * log10E)),
			   exp2 = _param[2] + _param[3] * E,
			   exp3 = exp(ln10 * (_param[4] + _param[5] * log10E));

		//	calculate the 3 coefficients for the main gaussian fit part
		double g1 = _param[6] + _param[7] * E +_param[8] * E2 + _param[9] * E3+
					_param[10] * E4,
			   g2 = _param[11] + _param[12] * E + _param[13] * E2,
			   g3 = _param[14] + _param[15] * E + _param[16] * E2;

		//	the rmf array starts at channel 20 usually ..., and we should
		//	calculate it up to about the end of the gaussian section
		//	(i.e. +/- 5 sigma) - but don't propagate the exponential part
		//	beyond the gaussian peak (=g2)
		unsigned maxChan = UNT(g2);
		double lnexp2 = log(exp2);
		for (unsigned n=20; n<=maxChan; ++n)
			spec[n] += exp1 * exp(DBL(n) * lnexp2) + exp3;

		int ilo = INT(g2) - 5*INT(g3),
			ihi = INT(g2) + 5*INT(g3);

		if (ilo < 0)
			ilo = 0;

		double ginv = 1./g3;
		for (int i=ilo; i<=ihi; ++i) {
			double z = ginv * (DBL(i)-g2);
			spec[i] += g1 * exp(-.5* z*z);
		}

		if (pixels & PIXEL_DOUBLES) {
			//	calculate the 3 coefficients for main  peak  gaussian fit part
			//	bi-pixel patterns
			double pg1 = _param[17] + _param[18] * E + _param[19] * E2,
				   pg2 = _param[20] + _param[21] * E,
				   pg3 = _param[22] + _param[23] * E;

			int ilo = INT(pg2) - 5*INT(pg3);
				ihi = INT(pg2) + 5*INT(pg3);

			if (ilo < 0)
				ilo = 0;
			double pginv = 1./pg3;
			for (int i=ilo; i<=ihi; ++i) {
				double z = pginv * (DBL(i) - pg2);
				spec[i] += pg1 * exp(-.5 * z*z);
			}
		}
	} else {
		//	the energy is above Si K edge
		double exp1 = exp(ln10 * (_param[24] + _param[25] * log10E +
						  _param[26] * log10E * log10E)),
			   exp2 = _param[27] + _param[28] * E + _param[29] * E2,
			   exp3 = exp(ln10 * (_param[30] + _param[31] * log10E));

		//	calculate the 3 coefficients for the main gaussian fit part
		double g1 = _param[32] + _param[33] * E + _param[34] * E2,
			   g2 = _param[35] + _param[36] * E + _param[37] * E2,
			   g3 = _param[38] + _param[39] * E + _param[40] * E2;

		double escapeE = E - SI_FLUORESCENT_PEAK;

		//	calculate the 3 coefficients for the escape peak  gaussian fit part
		double g4 = g1 * (0.03 - .0000025 * E),
			   g5 = _param[35] + _param[36] * escapeE +
						_param[37] * escapeE * escapeE,
			   g6 = _param[38] + _param[39] * escapeE +
						_param[40] * escapeE * escapeE;

		//	the fluorescent peak appeaers at 1740. ev - calaculate its
		//	gaussian coefficients
		double g7 = g1 * (0.01 - .0000045 * E),
			   g8 = _param[35] + _param[36] * SI_FLUORESCENT_PEAK +
						_param[37] * SI_FLUORESCENT_PEAK * SI_FLUORESCENT_PEAK,
			   g9 = _param[38] + _param[39] * SI_FLUORESCENT_PEAK +
						_param[40] * SI_FLUORESCENT_PEAK * SI_FLUORESCENT_PEAK;

		//	the rmf array starts at channel 20 usually..., and we should
		//	calculate it up to about the end of the gaussian section
		//	(i.e. +/- 5 SIGMA) - but don't propagate the exponential
		//	part beyond the gaussian peak (=g2)?
		unsigned maxChan = UNT(g2);
		double lnexp2 = log(exp2);

		for (unsigned n=20; n<=maxChan; ++n)
			spec[n] += exp1 * exp(DBL(n) * lnexp2) + exp3;

		int ilo = INT(g2) - 5*INT(g3),
			ihi = INT(g2) + 5*INT(g3);
		if (ilo < 0)
			ilo = 0;
		double pginv = 1./g3;
		for (int i=ilo; i<=ihi; ++i) {
			double z = pginv * (DBL(i) - g2);
			spec[i] += g1 * exp(-.5 * z*z);
		}

		ilo = INT(g5) - 5*INT(g6);
		ihi = INT(g5) + 5*INT(g6);
		if (ilo < 0)
			ilo = 0;
		pginv = 1./g6;
		for (int i=ilo; i<=ihi; ++i) {
			double z = pginv * (DBL(i) - g5);
			spec[i] += g4 * exp(-.5 * z*z);
		}

		ilo = INT(g8) - 5*INT(g9);
		ihi = INT(g8) + 5*INT(g9);
		if (ilo < 0)
			ilo = 0;
		pginv = 1./g9;
		for (int i=ilo; i<=ihi; ++i) {
			double z = pginv * (DBL(i) - g8);
			spec[i] += g7 * exp(-.5 * z*z);
		}

		if (pixels & PIXEL_DOUBLES) {
			//	calculate the 3 coefficients for main  peak  gaussian fit part
			//	for bi-pixel patterns
			double pg1 = _param[41] + _param[42] * E + _param[43]* E2 +
						 _param[44] * E3,
				   pg2 = _param[45] + _param[46] * E + _param[47]* E2,
				   pg3 = _param[48] + _param[49] * E + _param[50]* E2;

			int ilo = INT(pg2) - 5*INT(pg3);
				ihi = INT(pg2) + 5*INT(pg3);

			if (ilo < 0)
				ilo = 0;
			double pginv = 1./pg3;
			for (int i=ilo; i<=ihi; ++i) {
				double z = pginv * (DBL(i) - pg2);
				spec[i] += pg1 * exp(-.5 * z*z);
			}
		}
	}

	//	normalize
	spec.normalize();

	RETURN spec;
} // EmosRedist::spectrum()
#endif


bool EmosRedist::validateInput(double E, unsigned eventGrade) const
{
	PROC

	if (E < _eBounds->lower() || E > _eBounds->upper()) {
		if (!_failSafe)
			errstr<<"Input energy for EMOS redistribution ("<<E<<"eV) is out "
					"of valid bounds: "<<_eBounds->lower()<<"<=E<="
				  <<_eBounds->upper()
				  <<Sas::error("InvalidInput");
		ALTRETURNV(false);
	}

	if (eventGrade == 0 || eventGrade > ALL_GRADES) {
		if (!_failSafe)
			errstr<<"Event grade "<<eventGrade<<" is out of valid bounds: [0, "
				  <<UNT(ALL_GRADES)<<"]"
				  <<Sas::error("InvalidInput");
		ALTRETURNV(false);
	}

	RETURN true;
} // EmosRedist::validateInput()


EmosRedist::FractionalChargeLossGrid::FractionalChargeLossGrid(
								double alpha, double beta,
								const LogarithmicDoubleGrid &depthGrid)
	: _alpha(alpha), _beta(beta)
{
	PROC

	unsigned N = depthGrid.size();
	this->resize(N);
	vector<double> &v = *this;
	v[0] = 0.;
	double valOld = loss(0.);
	for (unsigned n=1; n<N; ++n) {
		double val = loss(depthGrid[n]);
		v[n] = .5 * (val + valOld);
		valOld = val;
	}

	ENDPROC
} // FractionalChargeLossGrid::FractionalChargeLossGrid


EmosRedist::AbsorptionDensityGrid::AbsorptionDensityGrid(
									  const EmosRedist *emosRedist, double E,
									  const LogarithmicDoubleGrid &depthGrid)
{
	PROC

	double absorptionL = 1.e6 * emosRedist->absorptionLengthInSi(E); // [um]

	unsigned N = depthGrid.size();
	this->resize(N);
	vector<double> &v = *this;
	v[0] = 0.;
	for (unsigned n=1; n<N; ++n)
		v[n] = exp(-depthGrid[n-1]/absorptionL) -
			   exp(-depthGrid[n]/absorptionL);

	ENDPROC
} // AbsorptionDensityGrid::AbsorptionDensityGrid()


EmosRedist::ModelParameterSet::ModelParameterSet(double E,
												 double SiKalpha,
												 const vector<Parameter> &p,
												 bool evenModel)
{
	PROC

	_mainNorm	= p[0].value(E);
	_mainE		= evenModel ? p[1].value(E) : E;
	_mainSigma	= p[2].value(E);
	_sLossAlpha	= p[3].value(E);
	_sLossBeta	= p[4].value(E);
	_redSigma	= p[5].value(E);
	_KGaussGamma= 0.;	//	King component disabled
	_KGaussDelta= 0.;
	_shelfNorm	= p[6].value(E);
	_floNorm	= p[7].value(E);
	_floSigma	= p[2].value(SiKalpha, true);
	_escE		= E - SiKalpha;
	_escNorm	= p[8].value(E);
	_escSigma	= p[2].value(_escE, true);

	ENDPROC
} // ModelParameterSet::ModelParameterSet()


//
//	ALGOID=0 spectrum construction
//	parameterization of MOS redistribution model after [4]
//
Spectrum EmosRedist::spectrum0(double eIn, const BinnedAxis &eAxis,
							   unsigned eventGrade) const
{
	PROC

	if (!validateInput(eIn, eventGrade)) {
		ALTRETURNV(Spectrum(0));
	}
	checkAt();

	Spectrum spec[_maxGrades];	//	partial redist spectra
	for (unsigned n=0; n<_maxGrades; ++n) {
		spec[n].resize(eAxis.bins());
		spec[n] = 0.;
	}

	//	construct absorption density grid
	AbsorptionDensityGrid absorptionDensity(this, eIn, _depth);

	//	loop over event types
	for (unsigned eventType=GRADE_1_MAIN; eventType<EVENT_TYPES;
		 ++eventType) {

		//	construct model parameter set
		ModelParameterSet modelParamSet(eIn, _SI_K_ALPHA, _param[eventType],
										_models[eventType] % 2 == 0);

		//	construct fractional charge loss grid
		FractionalChargeLossGrid fracChargeLoss(modelParamSet.sLossAlpha(),
												modelParamSet.sLossBeta(),
												_depth);

		//
		//	loop over the eAxis bins
		//
		double redistOld = SiResponse(_models[eventType],
									  eAxis.bin(UNT(0)).lo(),
									  fracChargeLoss, absorptionDensity,
									  modelParamSet);
		unsigned grade = toGrade(eventType),
				 N = eAxis.bins();
		for (unsigned n=0; n<N; ++n) {
			double E = eAxis.bin(n).hi(),
				   redist = SiResponse(_models[eventType], E,
									   fracChargeLoss, absorptionDensity,
									   modelParamSet);
			spec[grade-1][n] += .5 * (redistOld + redist);
			redistOld = redist;
		}
	}

	//	normalize
	for (unsigned n=0; n<_maxGrades; ++n) {
		double prob = _gradeProbabilities[n](eIn);
		spec[n].normalize(prob);
	}

	//	construct final spectrum
	Spectrum resSpec(eAxis.bins());
	resSpec = 0;
	for (unsigned n=0; n<_maxGrades; ++n)
		if (2>>n & eventGrade)
			//	consider this grade in final spectrum
			resSpec += spec[n];

	resSpec.normalize();

	RETURN resSpec;
} // EmosRedist::spectrum0()


//
//	returns the response at energy E
//
inline double EmosRedist::SiResponse(unsigned short model, double eOut,
								const FractionalChargeLossGrid &fracLossGrid,
								const AbsorptionDensityGrid &absDensityGrid,
								const ModelParameterSet &modelParamSet) const
{
	PROC

	double response = 0.;
	unsigned N = _depth.size();
	switch (model) {
		case 0:
		case 1:	{	//	Gauss model modified by surface loss with additional
					//	and shelf component
			double gContrib = 0.,
				   respMain = 0.,
				   siFlo = 0.,
				   siEsc = 0.;
			unsigned n;
			if (modelParamSet.mainE() > _SI_K_EDGE)
				for (n=1; n<N; ++n) {
					double fracLoss = fracLossGrid[n],
						   E0 = modelParamSet.mainE() * fracLoss;
					respMain = gauss(eOut, E0, modelParamSet.mainSigma());

					double floE0 = _SI_K_ALPHA * fracLoss;
					siFlo = modelParamSet.floNorm() *
								gauss(eOut, floE0, modelParamSet.floSigma());
					double escE0 = E0 - _SI_K_ALPHA;
					siEsc = modelParamSet.escNorm() *
								gauss(eOut, escE0, modelParamSet.escSigma());
					gContrib += absDensityGrid[n] * (respMain + siFlo + siEsc);
					if (fracLoss >= 1.)
						break;
				}
			else
				for (n=1; n<N; ++n) {
					double fracLoss = fracLossGrid[n],
						   E0 = modelParamSet.mainE() * fracLoss;
					respMain = gauss(eOut, E0, modelParamSet.mainSigma());
					gContrib += absDensityGrid[n] * respMain;
					if (fracLoss >= 1.)
						break;
				}

			double sum = respMain + siFlo + siEsc;
			for (unsigned nn=n+1; n<N; ++n)
				gContrib += absDensityGrid[nn] * sum;

			double shelfContrib = 0.;
			if (modelParamSet.shelfNorm() > 0.)
				shelfContrib = modelParamSet.shelfNorm() *
								shelfComponent(eOut, modelParamSet.mainE(),
											   modelParamSet.mainSigma(), 0.);

			response = modelParamSet.mainNorm() * (gContrib + shelfContrib);
			} break;
		case 2:
		case 3: {	//	compute contributions from main Gaussian photopeak
					//	+ Fluoresence and Escape
			double siGauss = gauss(eOut, modelParamSet.mainE(),
								   modelParamSet.mainSigma());
			if (modelParamSet.mainE() > _SI_K_EDGE) {
				//
				//	Flouresence Peak
				//
				double siFlo = modelParamSet.floNorm() *
							gauss(eOut, _SI_K_ALPHA, modelParamSet.floSigma());
				//
				//	Escape Peak
				//
				double escape = modelParamSet.mainE() - _SI_K_ALPHA,
					   siEsc = modelParamSet.escNorm() *
							gauss(eOut, escape,	modelParamSet.escSigma());

				siGauss += siFlo + siEsc;
			}
			response += modelParamSet.mainNorm() * siGauss;
			} break;
		case 4:
		case 5: {	//	King-Gauss model only
			double siKGauss = kingGauss(eOut, modelParamSet.mainE(),
										modelParamSet.mainSigma(),
										modelParamSet.KGaussGamma(),
										modelParamSet.KGaussDelta());
			if (modelParamSet.mainE() > _SI_K_EDGE) {
				//
				//	Flouresence Peak
				//
				double siFlo = modelParamSet.floNorm() *
										kingGauss(eOut, _SI_K_ALPHA,
												  modelParamSet.floSigma(),
												  modelParamSet.KGaussGamma(),
												  modelParamSet.KGaussDelta());
				//
				//	Escape Peak
				//
				double escape = modelParamSet.mainE() - _SI_K_ALPHA,
					   siEsc = modelParamSet.escNorm() *
										kingGauss(eOut, escape,
												  modelParamSet.escSigma(),
												  modelParamSet.KGaussGamma(),
												  modelParamSet.KGaussDelta());

				siKGauss += siFlo + siEsc;
			}

			response += modelParamSet.mainNorm() * siKGauss;
			} break;
		case 6:
		case 7: {	//	Red-Gauss Model only
			double siRedGauss = redGauss(eOut, modelParamSet.mainE(),
										 modelParamSet.mainSigma(),
										 modelParamSet.redSigma());
			if (modelParamSet.mainE() > _SI_K_EDGE) {
				//
				//	Flouresence Peak
				//
				double siFlo = modelParamSet.floNorm() *
										redGauss(eOut, _SI_K_ALPHA,
												 modelParamSet.floSigma(),
												 modelParamSet.redSigma());
				//
				//	Escape Peak
				//
				double escape = modelParamSet.mainE() - _SI_K_ALPHA,
					   siEsc = modelParamSet.escNorm() *
										redGauss(eOut, escape,
												 modelParamSet.escSigma(),
												 modelParamSet.redSigma());

				siRedGauss += siFlo + siEsc;
			}

			response += modelParamSet.mainNorm() * siRedGauss;
			} break;
		default:
			errstr<<"Unknown model ID "<<model<<" in construction of EMOS "
					"redistribution spectrum"
				  <<Sas::error("UnknownEmosRedistributionModel");
			break;
	}

	RETURN response;
} // EmosRedist::SiResponse()


//
//	ALGOID=1 spectrum construction - used Steve Sembay's EMOS redist
//	code in calmosalgo
//
Spectrum EmosRedist::spectrum1(double eIn, const BinnedAxis &eAxis,
							   unsigned eventGrade) const
{
	PROC

	static bool calMosAlgoInitialized = false;
	static float *echlo = 0, *echhi = 0;

	unsigned N = eAxis.bins();
	if (!calMosAlgoInitialized) {
		unsigned jmax, gmax;
		//
		//	JMAX/GMAX
		//
		if (eventGrade == ALL_GRADES) {
			jmax = CALMOSALGO::NALL;
			gmax = CALMOSALGO::GALL;
		} else if (eventGrade == GRADE_1) {
			jmax = CALMOSALGO::NZEROGRAD;
			gmax = CALMOSALGO::GZEROGRAD;
		} else {
			jmax = CALMOSALGO::NALL;
			gmax = CALMOSALGO::GALL;
			errstr<<"Can only select grade-1 or all grades in EMOS "
					"redistribution construction - will consider all"
				  <<Sas::warning("UnsupportedGradeSelection");
		}

		//
		//	EWIDTH/NEBIN
		//
		float ewidth = CALMOSALGO::EWIDTH;
		unsigned nebin = 1 + unsigned(eAxis.bin(N-1).hi() / ewidth);

		//
		//	THRESHOLD/BACKSIGMA/SI_KEDGE/SI_KALPHA
		//
		float threshold = _threshold,
			  backSigma = _backSigma,
			  SiKedge = _SI_K_EDGE,
			  SiKalpha = _SI_K_ALPHA;

		//
		//	ECHLO/ECHHI
		//
		delete [] echlo;
		delete [] echhi;
		echlo = new float[N];
		echhi = new float[N];
		for (unsigned n=0; n<N; ++n) {
			echlo[n] = eAxis.bin(n).lo();
			echhi[n] = eAxis.bin(n).hi();
		}

		CALMOSALGO_initRmfRow(jmax, gmax, _epar1, _param1, _models1, _grades1,
							  _nepts1, ewidth, nebin, threshold, backSigma,
							  echlo, echhi, SiKedge, SiKalpha);

		calMosAlgoInitialized = true;
	}

	if (!validateInput(eIn, eventGrade)) {
		ALTRETURNV(Spectrum(0));
	}
	checkAt();

	//
	//	construct spectrum
	//
	float	E = eIn,
			siabsd = absorptionLengthInSi(eIn) * 1.e6,	//	[um]
			gnorm[CALMOSALGO::MAXGRADE],
			rmfrow[N];

	double tot = 0.;
	for (unsigned n=0; n<CALMOSALGO::MAXGRADE; ++n) {
		gnorm[n] = eventGrade != GRADE_1 ? _gradeProbabilities[n](eIn) : 1.;
		tot += gnorm[n];
	}
	for (unsigned n=0; n<CALMOSALGO::MAXGRADE; gnorm[n++]/=tot);

	CALMOSALGO_rmfRow(E, siabsd, N, gnorm, rmfrow);

	Spectrum spec(N);
	for (unsigned n=0; n<N; ++n)
		spec[n] = rmfrow[n];

	RETURN spec.normalize();
} // EmosRedist::spectrum1()


Spectrum EmosRedist::spectrum(double eIn, const BinnedAxis &eAxis,
							  unsigned eventGrade) const
{
	PROC

	if (_algoid > 1)
		errstr<<"Unknown ALGOID "<<_algoid<<Sas::error("UnknownAlgoId");

	RETURN _algoid == 0
			? spectrum0(eIn, eAxis, eventGrade)
			: spectrum1(eIn, eAxis, eventGrade);
} // EmosRedist::spectrum()


EmosRedist::~EmosRedist()
{
	PROC

	delete [] _nepts1;
	delete [] _grades1;
	delete [] _models1;
	delete [] _epar1;
	delete [] _param1;

	ENDPROC
} // EmosRedist::~EmosRedist()


//
//	---( EpnRedist )--------------------------------------------------------
//
EpnRedist::EpnRedist(unsigned channels, double gain, Ccd,
					 const Proxy<BinnedAxis> &eBounds,
					 const Proxy<BinnedAxis> &eBins,
					 const Proxy<BinnedAxis> &binnedPIeBounds,
					 const DataSet *set,
					 CalAtomCollection<CanonicalRedist> *parent)
	: CanonicalRedist(channels, gain, eBounds, eBins, binnedPIeBounds),
	  CalChild<CalAtomCollection<CanonicalRedist> >(parent),
	  _atCalled(false)
{
	PROC

	if (!set->hasBlock("ABSORPTION")) {
		errstr<<"`"<<set->qualifiedName()<<"' has no Si/SiO2 absorption "
				"data table - computing absorption length from generic "
				"Henke data"<<Sas::warning("NoAbsorptionDataInCcf");

		//
		//	construct absorption data grid for Si/SiO2 from Henke data;
		//
		AbsorptionDataServer *acds = calServer.getAtom(acds);
	    const AbsCoefCurve &f2Si	= acds->henke("Si").f2(),
		                   &f2O		= acds->henke("O").f2();

	    double ASi     = acds->atomicData("Si").weight(),
		       AO      = acds->atomicData("O").weight(),			// [g/mol]
			   ASiO2   = ASi + 2. * AO;

	    double rhoSi   = acds->atomicData("Si").density() * 1.e-3,	// [g/cm^3]
			   rhoSiO2 = 2.64;

		vector<double> eGrid = f2Si.abscissa();
	    _nEnergiesSi = eGrid.size();
		_nEnergiesSiO2 = _nEnergiesSi;
		_eGridSi = new float [_nEnergiesSi];
		_eGridSiO2 = _eGridSi;
		_siabs = new float [_nEnergiesSi];
		_sio2abs = new float [_nEnergiesSi];

		for (int n=0; n<_nEnergiesSi; ++n) {
			double eV = eGrid[n],
				   f0 = PhysicalConstants::avogadro * 2. *      // [1/mol]
						PhysicalConstants::r0 * 1.e2 *          // [cm]
						PhysicalConstants::eVtimesAo/eV * 1.e-8;// [cm]

			_eGridSi[n] = eV;
			_siabs[n] = f0 * rhoSi/ASi * f2Si(eV) * 1.e-7;
			_sio2abs[n] = f0 * rhoSiO2/ASiO2 * (f2Si(eV) + 2.*f2O(eV)) * 1.e-7;
			//	1.e-7: 1/cm -> 1/nm
		}
	} else {
		CcfConstituentTable ccftab(set, "ABSORPTION");
		const Table *t = ccftab.block();
		ColumnData *cdCompound = t->column("COMPOUND")->data(),
				   *cdE = t->column("ENERGY")->data(),
				   *cdL = t->column("INV_LENGTH")->data();
		_nEnergiesSi = cdE->elements();
		_nEnergiesSiO2 = cdE->elements();
		if (UNT(_nEnergiesSi) != cdL->elements())
			errstr<<"Differnet vector lengths in `"<<t->qualifiedName()
				  <<"'"<<Sas::error("CcfIntegrityError");

		if (cdL->parent()->units() != "1/cm")
			errstr<<"Expected `"<<cdL->parent()->qualifiedName()
				  <<"' to be in 1/cm"<<Sas::error("CcfIntegrityError");

		if (CalOalUtility::stringColumnRow(cdCompound, 0) != "Si" ||
			CalOalUtility::stringColumnRow(cdCompound, 1) != "SiO2")
			errstr<<"Expected to find inverse absorption lengths for Si+SiO2"
					"in `"<<t->qualifiedName()
				  <<"'"<<Sas::error("CcfIntegrityError");
		delete cdCompound;

		real32 *e = cdE->real32Data(),
			   *l = cdL->real32Data();
		_eGridSi = new float [_nEnergiesSi];
		_siabs = new float [_nEnergiesSi];
		_eGridSiO2 = new float [_nEnergiesSiO2];
		_sio2abs = new float [_nEnergiesSiO2];
		for (int n=0; n<_nEnergiesSi; ++n) {
			_eGridSi[n] = e[cdE->position(0, n)];
			_siabs[n] = 1.e-7 * l[cdL->position(0, n)];	//	1/cm->1/nm
			_eGridSiO2[n] = e[cdE->position(1, n)];
			_sio2abs[n] = 1.e-7 * l[cdL->position(1, n)]; // 1/cm->1/nm
		}
		delete cdE;
		delete cdL;
	}

	ENDPROC
} // EpnRedist::EpnRedist()


void EpnRedist::checkAt() const
{
    PROC

    if (!_atCalled) {
        unsigned chipY = calServer.instrument()->maxY()/2;
        at(chipY);
        errstr<<"No chip line for redistribuion spectum specified - "
				"will retrieve data applicable at center ("
			  <<chipY<<")"
              <<Sas::message(Msg::Lib, Msg::Verbose);
        _atCalled = false;
    }

    ENDPROC
} // EpnRedist::checkAt()


const CanonicalRedist &EpnRedist::at(unsigned short chipY) const
{
	PROC

	_atCalled = true;
	CanonicalRedist::at(chipY);

	RETURN *this;
} // EpnRedist::at()


bool EpnRedist::validateInput(double E) const
{
	PROC

	if (E < _eBounds->lower() || E > _eBounds->upper()) {
		if (!_failSafe)
			errstr<<"Input energy for EPN redistribution ("<<E<<"eV) is out "
					"of valid bounds: "<<_eBounds->lower()<<"<=E<="
				  <<_eBounds->upper()
				  <<Sas::error("InvalidInput");
		ALTRETURNV(false);
	}

	RETURN true;
} // EpnRedist::validateInput()


Spectrum EpnRedist::spectrum(double eIn, const BinnedAxis &eAxis, unsigned)
const
{
	PROC

	if (!validateInput(eIn)) {
		ALTRETURNV(Spectrum(0));
	}
	checkAt();

	unsigned nChans = eAxis.bins();
	float response[nChans],
		  eMin[nChans],
		  eMax[nChans];
	for (unsigned n=0; n<nChans; ++n) {
		eMin[n] = eAxis.bin(n).lo();
		eMax[n] = eAxis.bin(n).hi();
	}

	int	line = _chipY;
	float eLo = eIn - .5 * eAxis.bin(0U).width(),
		  eHi = eIn + .5 * eAxis.bin(0U).width();

	char *cmode = "sw";
	Mode mode = calServer.state()->mode()->get();
	if (mode==Xmm::PRIME_FULL_WINDOW || mode==Xmm::PRIME_FULL_WINDOW_EXTENDED)
		cmode = "ff";

	CALPNALGO_resp(eLo, eHi, nChans, eMin, eMax, line, cmode,
				   _nEnergiesSi, _eGridSi, _siabs,
				   _nEnergiesSiO2, _eGridSiO2, _sio2abs, response);

	Spectrum spec(nChans);
	for (unsigned n=0; n<nChans; ++n)
		spec[n] = response[n];

	RETURN spec;
} // EpnRedist::spectrum()


EpnRedist::~EpnRedist()
{
	PROC

	delete [] _eGridSi; _eGridSi = 0;
	delete [] _eGridSiO2; _eGridSiO2 = 0;
	delete [] _siabs;
	delete [] _sio2abs;

	ENDPROC
} // EpnRedist::~EpnRedist()


//
//	---( RgsRedist )--------------------------------------------------------
//
RgsRedist::RgsRedist(unsigned channels, double gain, Ccd,
					 const Proxy<BinnedAxis> &eBounds,
					 const Proxy<BinnedAxis> &eBins,
					 const Proxy<BinnedAxis> &binnedPIeBounds,
					 const DataSet *set,
					 CalAtomCollection<CanonicalRedist> *parent)
	: CanonicalRedist(channels, gain, eBounds, eBins, binnedPIeBounds, parent),
	  CalChild<CalAtomCollection<CanonicalRedist> >(parent)
{
	PROC

	CcfConstituentTable ccftab(set, "CCD_REDISTRIBUTION");
	const Table *redistTable = ccftab.block();

	_fano = redistTable->attribute("FANO")->asReal();
	_SiThickness = redistTable->attribute("SI_THICK")->asReal();
	_b = redistTable->attribute("SCALEFAC")->asReal();
	_r = redistTable->attribute("E_RMS")->asReal();
	_w = redistTable->attribute("EH_E")->asReal();
	_eventSize = redistTable->attribute("EVT_SIZE")->asReal();
	_EThresh = redistTable->attribute("E_THR")->asReal();

	_partNorm = CalOalUtility::optionalAttribute(redistTable, "PARTNORM", 0.);
	_partDecy = CalOalUtility::optionalAttribute(redistTable, "PARTDECY", 0.);
	_partOffset = CalOalUtility::optionalAttribute(redistTable, "PARTOFFS",0.);
	_partRefE = CalOalUtility::optionalAttribute(redistTable, "PARTEREF",
												 1900.);

	ENDPROC
} // RgsRedist::RgsRedist()


double RgsRedist::collection(double e, double t) const
{
	PROC

	double a = pow(t/_b, 1./3.);
	RETURN _EThresh + (e - _EThresh) * (1. - exp(-a));
} // RgsRedist::collection()



Spectrum RgsRedist::noise(double energy, unsigned N) const
{
	PROC

//  	double sigma = sqrt(_fano * energy * _w + _eventSize * _r*_r * _w*_w),
//  		   sgm = sigma/sqrt(2.);
    // FIXME: these debug statements
	// sigma^2 in electrons
	double sigma_2_e = _fano * energy / _w + _eventSize * _r * _r;
	// sigma^2 in eV
	double sigma_2_eV =  sigma_2_e * _w * _w;
	// sigma^2 in PI
	double sigma_2_pi = sigma_2_eV * _gain * _gain;
	double sigma = sqrt(sigma_2_pi);
	const double sqrt2 = sqrt(2.);

	Spectrum spec(N);
	for (unsigned n=0; n<N; ++n) {
		double ss = DBL(n) - DBL(N/2);
		spec[n] = .5 * fabs(Slatec::erf((ss - .5)/sigma/sqrt2)
							- Slatec::erf((ss + .5)/sigma/sqrt2));

		SAS_DEBUG(cout<<n<<" "<<(ss - .5)/sigma/sqrt2<<" "
					  <<(ss + 0.5)/sigma/sqrt2<<" "
					  <<spec[n]<<endl;)
    }
	spec.normalize();
	spec.rotate(N/2);

	RETURN spec;
} // RgsRedist::noise()


Spectrum RgsRedist::noiseFreeSpectrum(double energy, const BinnedAxis &eAxis,
									  unsigned) const
{
	PROC

	AbsorptionDataServer *ads = calServer.getAtom(ads);
	const AbsCoefCurve &f2 = ads->henke("Si").f2();

	double eMax = collection(energy, _SiThickness),
		   tau = absorptionLength(f2(energy), energy);
	double bOverTau = _b / tau;

	//	Eq. (10), [2]
	double Ppart = _partNorm*(1.-exp(-_partDecy*(energy-_partOffset))) *
						absorptionLength(f2(_partRefE), _partRefE)/tau,
		   t = Ppart/energy;

	unsigned nBins = eAxis.bins();
	Spectrum spec(nBins);
	spec = 0.;

	for (unsigned i=0; i<nBins; i++) {
		double ec,
			   binWidth,
			   eLo = DBL(i) * _gain,
			   eHi = DBL(i + 1) * _gain;

		if (eHi <= _EThresh)
			continue; // below threshold, no charge collected
		if (eLo >= eMax)
			continue;
		if (eLo < _EThresh) {
			ec = .5 * (eHi + _EThresh);
			binWidth = eHi - _EThresh;
        } else {
			if (eHi >= eMax) {
				ec = .5 * (eLo + eMax);
				binWidth = eMax - eLo;
            } else {
				ec = .5 * (eLo + eHi);
				binWidth = eHi - eLo;
			}
		}

		double ln = log((energy - _EThresh)/(energy - ec));
		spec[i] = 3. * bOverTau / (energy - ec) * pow(ln, 2.) *
											exp(-bOverTau * pow(ln, 3.));
		if (ec < energy)
			spec[i] += t;

		spec[i] *= binWidth;
	}

	//	On a fine grid the spectrum would be normalized to norm
	//	below. Here we simply make sure that this is so by modifying
	//	the value in spectrum's peak channel.
	double norm = 1. - exp(-_SiThickness/tau);
//	double tot = 0.;
//	tot = accumulate(spec.begin(), spec.end(), tot);
//	spec[spec.peakChannel()] += norm - tot;

	spec.normalize(norm);

	RETURN spec;
} // RgsRedist::noiseFreeSpectrum()


Spectrum RgsRedist::spectrum(double energy,
							 const BinnedAxis &eAxis,
							 unsigned eventGrade) const
{
	PROC

	const unsigned NPADS = eAxis.bins();

	Spectrum totResponse(NPADS);
	totResponse = 0.;

	if (energy < eAxis.bin(UNT(0)).lo()) {
		if (!_failSafe)
			errstr<<"Input energy E="<<energy<<"eV is smaller than energy of "
					"lower bound of first bin of energy axis (E="
				  <<eAxis.bin(UNT(0)).lo()<<"eV) - returning null-spectrum"
				  <<Sas::warning("UnphysicalInput");
		ALTRETURNV(Spectrum(0));
	}
	if (energy > eAxis.bin(UNT(NPADS-1)).hi()) {
		if (!_failSafe)
			errstr<<"Input energy E="<<energy<<"eV is greater than energy of "
					"upper bound of last bin of energy axis (E="
				  <<eAxis.bin(UNT(NPADS-1)).hi()<<"eV) - returning "
					"null-spectrum"
				  <<Sas::warning("UnphysicalInput");
		ALTRETURNV(Spectrum(0));
	}

	Spectrum signal = noiseFreeSpectrum(energy, eAxis, eventGrade);

	unsigned nBins = eAxis.bins();
	//	zero-pad noise-free spectrum up to _channels+NPADS
	for (unsigned n=nBins; n<nBins+NPADS; ++n)
		signal.push_back(0.);

	//	construct wrapped-around noise spectrum of length _channels+NPADS
	Spectrum response = noise(energy, nBins+NPADS);

	//	convolve signal with response
	totResponse = Slatec::convolve(signal, response);

	//	get rid of padded channels
	totResponse.erase(totResponse.begin()+nBins, totResponse.end());
	totResponse.normalize();

	RETURN totResponse;
} // RgsRedist::spectrum()


RgsRedist::~RgsRedist()
{
	PROC
	ENDPROC
} // RgsRedist::~RgsRedist()
