//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	Redistribution.h --- classes to model canonical CCD responses
//
//	DESCRIPTION:
//		Several classes are declared to describe the canonical (ideal) response
//		of a CCD to monochromatic photons.
//
//	CHANGE HISTORY:
//		UL, 1999-09-13: 1.0.0: - first version from CalAtomsCommon
//		UL, 1999-09-15: 1.1.0: - added noiseFreeSpectrum() to Redistribution
//								 classes
//		UL, 1999-10-27: 2.0.0: - reintroduced eBins()
//							   - ebounds() -> eBounds()
//		UL, 2000-05-19: 2.1.0: - overhauled EmosRedist class to be in
//								 line with new redistribution model w.r.t
//								 grade selection
//		UL, 2000-05-22: 2.2.0: - fixed compilation problem under True64
//		UL, 2000-05-23: 2.3.0: - fixed a number of problems with MOS
//							     redistribution model
//		UL, 2000-05-30: 2.4.0: - added binnedPIeBounds()
//		UL, 2000-08-31: 2.5.0: - added EpnRedist class spectrum() method of
//								 which is interfaced to EPN redistribution
//								 code in calpnalgo package
//
//	AUTHOR:
//		U. Lammers, Mon Sep 13 16:37:27 MET DST 1999
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//		[2]: Calibration Access and Data Handbook, C. Erd et al. and
//			 references therein
//		[3]: MOS CCD RMF Description, S. Sembay et al., EPIC-LUX-RE-170
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/Redistribution.h,v 1.20 2001/03/15 17:38:11 ulammers Exp $
//

#ifndef REDISTRIBUTION_H
#define REDISTRIBUTION_H

#include <vector>
#include <string>

#include "Tracer.h"
#include "Dal.h"
#include "Ccf.h"
#include "CalChild.h"
#include "CalAtomsBase.h"
#include "CalCommon.h"
#include "Proxy.h"

template<class T>
T SSQR(T x) { return x*x; }

//
//	---( CanonicalRedist )----------------------------------------------------
//
class CanonicalRedist : public CalAtom,
	public virtual CalChild<CalAtomCollection<CanonicalRedist> > {
public:
	CanonicalRedist(unsigned channels, double gain,
					const Proxy<BinnedAxis> &eBounds,
					const Proxy<BinnedAxis> &eBins,
					const Proxy<BinnedAxis> &binnedPIeBounds,
					CalAtomCollection<CanonicalRedist> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate an CanonicalRedist object
		//


	static unsigned id() { PROC RETURN Ccf::Redist; }
		//
		//	PURPOSE:
		//		Return a uniue ID which for mapping Cal atoms onto
		//		CCF constituents
		//


	virtual void dump(ostream &os);
		//
		//	PURPOSE:
		//		Dump data members to an ostream
		//
		//	ARGUMENTS:
		//		In:
		//			os:	stream to print on
		//


	virtual const BinnedAxis &eBins() const;
		//
		//	PURPOSE:
		//		Return the energy axis on which the CCD redistribution is
		//		defined
		//


	virtual const BinnedAxis &eBounds() const;
		//
		//	PURPOSE:
		//		Return the energy axis on which the CCD redistribution is
		//		defined
		//


	virtual const BinnedAxis &binnedPIeBounds() const;
		//
		//	PURPOSE:
		//		Return the energy axis on which the CCD redistribution is
		//		defined
		//


	virtual const CanonicalRedist &failSafe(bool suppressError = false)
		const;
		//
		//	PURPOSE:
		//		Define behavior of redistribution spectrum generation
		//		in case of error conditions
		//
		//	ARGUMENTS:
		//		suppressError	: setting this to 'true' will suppress
		//						  all errors/warnings that might result during
		//						  the generation of a redistribution spectrum;
		//						  if such a condition arises a spectrum
		//						  of size zero shall be returned to the caller
		//


	virtual const CanonicalRedist &at(unsigned short chipY) const;
	virtual const CanonicalRedist &at(unsigned short chipX,
									  unsigned short chipY) const;
		//
		//	PURPOSE:
		//		Define the chip location <chipX>/<chipY> to which subsequent
		//		spectrum() invocations should refer to; if at() is not invoked
		//		prior to spectrum() the chip center will be used as	default
		//		location;
		//		The first form is to be used in conjunction with the
		//		redistribution model of EPN which is sensitive to the CCD
		//		line;
		//


	enum { GRADE_1 = 1, GRADE_2 = 2, GRADE_3 = 4, GRADE_4 = 8,
		   ALL_GRADES = GRADE_1|GRADE_2|GRADE_3|GRADE_4 };
	virtual Spectrum noiseFreeSpectrum(double E,
							  const BinnedAxis &eAxis,
							  unsigned eventGrade = ALL_GRADES) const;
	virtual Spectrum spectrum(double E, const BinnedAxis &eAxis,
							  unsigned eventGrade = ALL_GRADES) const;

	virtual Spectrum noiseFreeSpectrum(double E,
							  unsigned eventGrade = ALL_GRADES) const;
	virtual Spectrum spectrum(double E,
							  unsigned eventGrade = ALL_GRADES) const;

	virtual Spectrum noiseFreeSpectrum(unsigned index,
							  unsigned eventGrade = ALL_GRADES) const;
	virtual Spectrum spectrum(unsigned index,
							  unsigned eventGrades = ALL_GRADES) const;
		//
		//	PURPOSE:
		//		Construct normalized CCD redistribution spectrum for a photon
		//		of a given energy.
		//
		//	ARGUMENTS:
		//		index		: index of energy bin; 0<=index<=ebounds.bins()-1
		//		E			: energy in eV of incoming photon
		//		eAxis		: energy axis on which the redistribution spectrum
		//					  is to be evaluated; the forms which do not
		//					  take <eAxis> as argument use the eBins
		//					  axis from the CCF
		//		eventGrade	: number of pixels above threshold an event
		//					  consists of (1=single-pixel events,
		//					  2=bi-pixel-events etc.) - see enum above
		//					  for symbolic names; combinations can be
		//					  formed through logically or'ing, e.g.
		//					  construct spectrum for grade 1+3 events:
		//						spectrum(E, GRADE_1|GRADE_3);
		//					  default is ALL_GRADES; i.e. contributions
		//					  from all grades are taken into account;
		//


	virtual ~CanonicalRedist();
		//
		//	PURPOSE:
		//		Destrory CanonicalRedist object constructed with above
		//		constructor(s).
		//


	//	return absorption depth of photon with energy <E> [eV]
	double absorptionLength(double f2, double E) const;
	double absorptionLengthInSi(double E) const;

protected:
	Proxy<BinnedAxis>			_eBounds,
								_eBins,
								_binnedPIeBounds;
	double						_gain;
	unsigned					_channels;
	mutable unsigned short		_chipX, _chipY;

	//	evaluate value of Gaussian curve centered at <xCenter> with
	//	sigma <sigma> at point <x>
	inline double gauss(double x, double xCenter, double sigma) const
	{
		PROC

		double arg = sigma != 0. ? (x-xCenter)/sigma : 1.e20;
		arg *= arg;
		if (arg >= 32.)
			ALTRETURNV(0.);

		RETURN exp(-.5*arg);
	} // gauss()


	//	evaluate value of King-Gauss curve centered at <xCenter> with
	//	sigma/gamma/delta <sigma>/<gamma>/<delta> at point <x>
	inline double kingGauss(double x, double xCenter, double sigma,
							double gamma, double delta) const
	{
		PROC

		if (x < xCenter) {
			double d = SSQR((x-xCenter)/(sigma*gamma));
			ALTRETURNV(1./pow(1. + d, delta));
		}

		RETURN gauss(x, xCenter, sigma);
	} // kingGauss()


	//	evaluate value of red-Gauss curve centerd at <xCenter> with
	//	sigma/redSigma <sigma>/<redSigma> at point <x>
	inline double redGauss(double x, double xCenter, double sigma,
						   double redSigma) const
	{
		PROC
		RETURN gauss(x, xCenter, sigma * redSigma);
	} // redGauss()

	mutable bool _failSafe;		//	suppress errors/warnings in spectrum
								//	generation if set to true; output
								//	spectrum will have size zero in this
								//	case
private:
	double		 _SiA,			//	atomic weight Si
				 _SiDensity;	//	density [kg/m^3]

};


//
//	EMOS redistribution model
//
class EmosRedist : public virtual CanonicalRedist {
public:
	EmosRedist(unsigned channels, double gain, Ccd ccd,
			   const Proxy<BinnedAxis> &eBounds,
			   const Proxy<BinnedAxis> &eBins,
			   const Proxy<BinnedAxis> &binnedPIeBounds,
			   const DataSet *set,
			   CalAtomCollection<CanonicalRedist> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate EmosRedist object from CCF constituent
		//


	const CanonicalRedist &at(unsigned short chipX,
							  unsigned short chipY) const;
		//
		//	PURPOSE:
		//		see above
		//


	Spectrum spectrum(double E, const BinnedAxis &eAxis,
					  unsigned eventGrade = ALL_GRADES) const;
		//
		//	PURPOSE:
		//		Construct redistribution spectrum corresponding to
		//		input energy <E> and pixel grade <eventGrade>
		//


	~EmosRedist();
		//
		//	PURPOSE:
		//		Destroy EMOS redistribution object.
		//

private:
	typedef enum { GRADE_1_MAIN = 0,
				   GRADE_1_LEAK,
				   GRADE_2_MAIN,
				   GRADE_2_LEAK,
				   GRADE_3_MAIN,
				   GRADE_4_MAIN } EventTypes;
	unsigned toGrade(unsigned eventType) const {
		PROC
		const unsigned eg[] = { 1, 1, 2, 2, 3, 4 };
		RETURN eg[eventType];
	}

	static const unsigned EVENT_TYPES = GRADE_4_MAIN+1;
	class Parameter {
	public:
		static const unsigned N_PARAM = 6;
		Parameter() { PROC ENDPROC }
		Parameter(float pvec[N_PARAM], double SI_K_EDGE,
				  double minVal, double maxVal)
			: _SI_K_EDGE(SI_K_EDGE), _minVal(minVal), _maxVal(maxVal)
		{
			PROC
			for (unsigned n=0; n<N_PARAM; ++n)
				_p[n] = pvec[n];
			ENDPROC
		}
		double value(double E, bool alwaysHighSet = false) const
		{
			PROC
			double val = E<_SI_K_EDGE && !alwaysHighSet
							? _p[0] + _p[1]*E + _p[2]*E*E
							: _p[3] + _p[4]*E + _p[5]*E*E;

			if (val < _minVal)
				ALTRETURNV(_minVal);
			if (val > _maxVal)
				ALTRETURNV(_maxVal);

			RETURN val;
		}	
		~Parameter() {}
	private:
		double _p[N_PARAM],
			   _SI_K_EDGE,
			   _minVal, _maxVal;
	};
	vector<Parameter>		_param[EVENT_TYPES];
	vector<unsigned short>	_grades, _models;

	double		_SI_K_EDGE,
				_SI_K_ALPHA;

	unsigned									_maxGrades;
	mutable vector<CcdQuantumEfficiencyCurve>	_gradeProbabilities;
	mutable bool								_atCalled;

	LogarithmicDoubleGrid						_depth;

	class FractionalChargeLossGrid : public vector<double> {
	public:
		FractionalChargeLossGrid(double alpha, double beta,
								 const LogarithmicDoubleGrid &depthGrid);

		double loss(double depth) const
		{
			PROC
			double v = _alpha + _beta * depth;
			if (v > 1.) {
				ALTRETURNV(1.);
			} else if (v < 0.) {
				ALTRETURNV(0.);
			}
			RETURN v;
		} // loss()

		~FractionalChargeLossGrid() { PROC ENDPROC }

	private:
		double _alpha, _beta;
	};

	class AbsorptionDensityGrid : public vector<double> {
	public:
		AbsorptionDensityGrid(const EmosRedist *emosRedist, double E,
							  const LogarithmicDoubleGrid &depthGrid);
		~AbsorptionDensityGrid() { PROC ENDPROC }
	};

	class ModelParameterSet {
	public:
		ModelParameterSet(double E,
						  double SiKalpha,
						  const vector<Parameter> &p,
						  bool evenModel);

		double mainE() const { PROC RETURN _mainE; }
		double mainNorm() const { PROC RETURN _mainNorm; }
		double mainSigma() const { PROC RETURN _mainSigma; }
		double sLossAlpha() const { PROC RETURN _sLossAlpha; }
		double sLossBeta() const { PROC RETURN _sLossBeta; }
		double redSigma() const { PROC RETURN _redSigma; }
		double KGaussGamma() const { PROC RETURN _KGaussGamma; }
		double KGaussDelta() const { PROC RETURN _KGaussDelta; }
		double shelfNorm() const { PROC RETURN _shelfNorm; }
		double floNorm() const { PROC RETURN _floNorm; }
		double floSigma() const { PROC RETURN _floSigma; }
		double escE() const { PROC RETURN _escE; }
		double escNorm() const { PROC RETURN _escNorm ; }
		double escSigma() const { PROC RETURN _escSigma; }

		~ModelParameterSet() { PROC ENDPROC }
	private:
		double _mainE,		//	energy of main photopeak
			   _mainNorm,	//	normalization of main photopeak
			   _mainSigma,	//	width of spread main function
			   _sLossAlpha,	//	surface loss function parameter alpha
			   _sLossBeta,	//	surface loss function parameter beta
			   _redSigma,	//	relative sigma "red" tail in redgauss model
			   _KGaussGamma,//	relative sigma King component in KGauss model
			   _KGaussDelta,//	exponent in King component in KGauss model
			   _shelfNorm,	//	relative normalisation of shelf function
			   _floNorm,	//	relative normalisation of fluoresence peak
			   _floSigma,	//	width of fluoresence peak
			   _escE,		//	energy of fluoresence peak
			   _escNorm,	//	relative normalisation of escape peak
			   _escSigma;	//	width of escape peak
	};

	bool validateInput(double E, unsigned eventGrade) const;
	void checkAt() const;

	inline double shelfComponent(double e, double mainE, double mainSigma,
							     double eShelfMin) const
	{
		PROC
		double loLim = -5. * mainSigma,
			   hiLim = 5. * mainSigma,
			   eLo = eShelfMin - e,
			   eHi = mainE - e;

		if (eLo < loLim && eHi < loLim ||
			eLo > hiLim && eHi > hiLim)
			ALTRETURNV(0.);

		if (eLo < loLim && eHi > hiLim)
			ALTRETURNV(1.);

		RETURN .5 * (Slatec::erf(eHi/mainSigma) - Slatec::erf(eLo/mainSigma));
	} // shelfComponent


	double SiResponse(unsigned short model, double eOut,
					  const FractionalChargeLossGrid &fracLossGrid,
					  const AbsorptionDensityGrid &absDensityGrid,
					  const ModelParameterSet &modelParamSet) const;


	//	ALGOID 1: EMOS redistribution code in calmosalgo
	double		_threshold,	//	no response below this threshold [eV]
				_backSigma;	//	sigma of noise component to convolve signal
							//	with [eV]
	mutable int		*_nepts1,
					*_grades1,
					*_models1;
	mutable float	*_epar1,
					*_param1;

	void constructForAlgo0(const Table *);	//	algo-0
	void constructForAlgo1(const Table *);	//	algo-1

	Spectrum spectrum0(double, const BinnedAxis &, unsigned) const;
	Spectrum spectrum1(double, const BinnedAxis &, unsigned) const;
};


//
//	EPN redistribution model
//
class EpnRedist : public virtual CanonicalRedist {
public:
	EpnRedist(unsigned channels, double gain, Ccd ccd,
			  const Proxy<BinnedAxis> &eBounds,
			  const Proxy<BinnedAxis> &eBins,
			  const Proxy<BinnedAxis> &binnedPIeBounds,
			  const DataSet *set,
			  CalAtomCollection<CanonicalRedist> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate EpnRedist object from CCF constituent
		//


	const CanonicalRedist &at(unsigned short chipY) const;
		//
		//	PURPOSE:
		//		see above
		//


	Spectrum spectrum(double E, const BinnedAxis &eAxis,
					  unsigned eventGrade = ALL_GRADES) const;
		//
		//	PURPOSE:
		//		Construct redistribution spectrum corresponding to
		//		input energy <E> and pixel grade <eventGrade>
		//


	~EpnRedist();
		//
		//	PURPOSE:
		//		Destroy EPN redistribution object.
		//

private:
	mutable bool	_atCalled;
	int				_nEnergiesSi, _nEnergiesSiO2;
	float			*_eGridSi, *_eGridSiO2,
					*_siabs,
					*_sio2abs;

	bool validateInput(double E) const;
	void checkAt() const;
};


//
//	RGS redistribution model
//
class RgsRedist : public virtual CanonicalRedist {
public:
	RgsRedist(unsigned channel, double gain, Ccd ccd,
			  const Proxy<BinnedAxis> &eBounds,
			  const Proxy<BinnedAxis> &eBins,
			  const Proxy<BinnedAxis> &binnedPIeBounds,
			  const DataSet *set,
			  CalAtomCollection<CanonicalRedist> *parent = 0);
		//
		//	PURPOSE:
		//		Instantiate RgsRedist object from CCF constituent
		//


	Spectrum spectrum(double E, const BinnedAxis &eAxis,
					  unsigned eventGrade = ALL_GRADES) const;
	Spectrum noiseFreeSpectrum(double E, const BinnedAxis &eAxis,
							   unsigned eventGrade = ALL_GRADES) const;
		//
		//	PURPOSE:
		//		Construct (noise free) redistribution spectrum corresponding to
		//		input energy <E> on an binned energy axis eAxis; the argument
		//		eventGrade is is currently ignored; if the input energy <E>
		//		is greater than the energy of the upper bound of the highest
		//		bin leads a warning message will be issued and a null-spectrum
		//		returned
		//


	~RgsRedist();
		//
		//	PURPOSE:
		//		Destroy RGS redistribution object.
		//

private:
	//	data from CCF::Redist
	double	_fano,			//	[unitless] FANO factor
			_SiThickness,	//	[cm] CCD Si thickness
			_b,				//	[unitless] collection scale depth
			_r,				//	[e] readout noise
			_w,				//	[eV/e] energy to produce e-h pair
			_eventSize,		//	[pixel] for OCB 1x1
			_EThresh,		//	[eV] acceptance threshold energy

			_partNorm,		//	g in Eq. 10, [2]
			_partDecy,		//	[1/eV], c in Eq. 10, [2]
			_partOffset,	//	[eV], d in Eq. 10, [2]
			_partRefE;		//	[eV], constant argument of tau in Eq. 10, [2]

	//	data from CCF::AbsCoef
	double	_SiA,			//	atomic weight of Si
			_SiDensity;		//	[kg/m^3] density of Si

	//	data from CCF::MiscData
	double	_r0,			//	[m] classical electron radius
			_avogadro,		//	[1/mol] number of particles in 1 mol material
			_c,				//	[m/s] light velocity
			_h,				//	[Js] Planck's constant
			_eVToJ;			//	[J/eV] J to Ev

	//	methods
	double collection(double e, double t) const;
	Spectrum noise(double energy, unsigned N) const;
};

#endif

