//-----------------------------------------------------------------------------
//
//						XMM Science Analysis System (SAS)
//					   (c) 1997-2010 European Space Agency
//
//-----------------------------------------------------------------------------
//
//	TimeCorrector.cc --- implementation of TimeCorrector class
//
//	DESCRIPTION:
//
//	CHANGE HISTORY:
//		UL, 1999-11-23: 1.0.0: - first version
//		UL, 1999-11-29: 1.1.0: - bug fix for treatment of timing mode
//
//	AUTHOR:
//		U. Lammers, Mon Oct  6 15:32:57 WET DST 1997
//
//	REFERENCES:
//		[1]: Software Specification Document for the XMM Science Analysis
//			 Subsystem, M. Dahlem et. al., 
//
//		[2]: Time resolution capabilities of the XMM EPIC pn-CCD in different
//			 readout modes, M. Kuster et al.
//
//	$Header: /data/xmm12/cvsadmin/soc_repos/sas/packages/cal/src/TimeCorrector.cc,v 1.8 2000/12/20 16:04:53 ulammers Exp $
//

#include <iomanip>

#include "errstr.h"
#include "common.h"
#include "TimeCorrector.h"
#include "CalServer.h"
#include "Xmm.h"

//
//	---( TimeCorrector )---------------------------------------------------
//
CalAtomCollection<TimeCorrector>::CalAtomCollection(
								const XmmInstrument *instrument,
								const Ccf *ccf,
								Xmm::Filter filter,
								AccuracyLevel accuracy)
{
	PROC

	_filter		= filter;
	_accuracy	= accuracy;
	_nCCDs		= instrument->nCcds();
	_nNodes		= instrument->nNodes();
	_nModes		= instrument->nModes();
	_nEntries	= 1;
	TimeCorrector *null = 0;
	_collection = new vector<TimeCorrector *>(_nEntries, null);
	TimeCorrector *&corrector = (*_collection)[0];

	//
	//	read the principal data from the CCF
	//
	CifEntry *e = ccf->entry(instrument, Ccf::TimeCorr);
	DataSet *dataSet = e->open();

	//
	//	the rest is instrument specific
	//
	switch (instrument->id()) {
		case Xmm::EPN: {
			CcfConstituentTable ccftab(dataSet, "TIMECORR");
			const Table *table = ccftab.block();
			ColumnData	*cdmodeId = table->column("MODE_ID")->data(),
						*cdfrmt = table->column("FRAME_TIME")->data();
			int8		*modeId = cdmodeId->int8Data();
			real32		*frmt	= cdfrmt->real32Data();
			
			unsigned N = cdfrmt->elements();
			if (N != 3)
				errstr<<"Expected to see three elements in column "
					  <<cdfrmt->parent()->qualifiedName()<< " but there are "
					  <<N<<Sas::error("DataMismatch");
			vector<vector<float> >	coefs(_nModes);
			for (unsigned i=0; i<table->rows(); ++i) {
				unsigned mode = modeId[i];
				if (mode > _nModes)
					errstr<<"Illegal modeId ("<<mode<<") in table `"
						  <<table->qualifiedName()<<"'"
						  <<Sas::error("IllegalModeId");

				vector<float>	v(3);
				v[0] = frmt[cdfrmt->position(i, 0)];
				v[1] = frmt[cdfrmt->position(i, 1)];
				v[2] = frmt[cdfrmt->position(i, 2)];

				coefs[mode] = v;
			}
			corrector = new EpnTimeCorrector(coefs, this);
		}
		break;
		default:
			corrector = new TimeCorrector(this);
			break;
	}
	dataSetServer->close(dataSet);

	ENDPROC
} // CalAtomCollection<TimeCorrector>::CalAtomCollection()


void TimeCorrector::dump(ostream &os)
{
	PROC

	os<<"generic TimeCorrector{}";

	ENDPROC
} // TimeCorrector::dump()


TimeCorrector *CalAtomCollection<TimeCorrector>::select(
					Ccd ccd,
					Node node,
					unsigned) const
{
	PROC

	_ccd = ccd;
	_node = node;

	RETURN (*_collection)[0];
} // CalAtomCollection<TimeCorrector>::select



void TimeCorrector::noOp() const
{
	PROC

//	errstr<<"Generic time corrector cannot perform fine time adjustments - "
//			"leaving input unchanged"
//		  <<Sas::warning("NoOperation");

	ENDPROC
} // TimeCorrector::noOp()


CalReal64Vector &TimeCorrector::correct(CalReal64Vector &time) const
{
	PROC

	noOp();

	RETURN time;
} // TimeCorrector::correct()


CalReal64Vector &TimeCorrector::correct(CalReal64Vector &time,
										CalInt16Vector &) const
{
	PROC

	noOp();

	RETURN time;
} // TimeCorrector::correct()


CalReal64Vector &TimeCorrector::correct(CalReal64Vector &time,
										CalInt16Vector &, int16) const
{
	PROC

	noOp();

	RETURN time;
} // TimeCorrector::correct()


//
//	-------------( EpnTimeCorrector )----------------------------------------
//
CalReal64Vector &EpnTimeCorrector::correct(CalReal64Vector &time) const
{
	//	the following implements Equations (1)-(6) + (10) in Ref. [2]

	PROC

	//	retrieve mode / current CCD number and break latter in quadrant and
	//	ccd id [0, 2]
	Mode mode = calServer.state()->mode()->get();
	unsigned modeId = xmmPayLoad.epn.modeIndex(mode);

	Ccd ccd = calServer.state()->ccd()->get();
	unsigned ccdId = EPNcam::toCcdId(ccd);

	double deltaT = 0.;

	switch (mode) {
		case Xmm::PRIME_FULL_WINDOW:
		case Xmm::PRIME_FULL_WINDOW_EXTENDED:
		case Xmm::PRIME_LARGE_WINDOW:
			deltaT = _coefs[modeId][ccdId];
			break;

		case Xmm::PRIME_SMALL_WINDOW:
			if (ccd != 4)
				errstr<<"In `"<<mode<<"' mode only CCD #4 can be active but "
						"the current CCD is "<<ccd
					  <<Sas::error("IllegaCcd");
			deltaT = _coefs[modeId][0];
			break;

		case Xmm::FAST_BURST:
			deltaT = _coefs[modeId][0];
			break;

		default:
			errstr<<"Scalar fine time adjustment in mode `"<<mode<<"' it not "
					"supported"<<Sas::error("UnsupportedOperation");
			break;
	}

	//	convert deltaT from ms to s
	deltaT *= 1.e-3;

	unsigned N = time.size();
	for (unsigned n=0; n<N; ++n)
		time[n] -= deltaT;	//	Eq. (1)-(6) + (10) in [2]

	RETURN time;
} // EpnTimeCorrector::correct()


CalReal64Vector &EpnTimeCorrector::correct(CalReal64Vector &time,
										   CalInt16Vector &rawy) const
{
	PROC

	//	the following implements Eq. (7) in Ref. [2]

	//	get mode from Cal state
	Mode mode = calServer.state()->mode()->get();
	unsigned modeId = xmmPayLoad.epn.modeIndex(mode);

	if (mode != Xmm::FAST_TIMING)
		errstr<<"Fine-time adjustment with RAWY and no source position in "
				"mode `"<<mode<<"' is not supported - this is only possible "
				"in mode"<<Xmm::FAST_TIMING
			  <<Sas::error("UnsupportedOperation");

	double p1 = _coefs[modeId][0] * 1.e-3,
		   p2 = _coefs[modeId][1] * 1.e-3;

	unsigned N = time.size();
	for (unsigned n=0; n<N; ++n)
		time[n] -= (9-INT(DBL(rawy[n]-1)/20.)) * p1 - p2;	//	Eq. (7) in [2]

	RETURN time;
} // EpnTimeCorrector::correct()


CalReal64Vector &EpnTimeCorrector::correct(CalReal64Vector &time,
										   CalInt16Vector &rawy,
										   int16 sourceY) const
{
	PROC

	//	the following implements Eq. (9) + (11) in Ref. [2]

	//	get mode from Cal state
	Mode mode = calServer.state()->mode()->get();
	unsigned modeId = xmmPayLoad.epn.modeIndex(mode);

	unsigned N = time.size();
	if (mode == Xmm::FAST_TIMING) {
		double p = _coefs[modeId][2] * 1.e-3;

		for (unsigned n=0; n<N; ++n)
			time[n] += (rawy[n]-1 - 200 - INT(.1*DBL(sourceY-1))) * p;
														//	Eq. (9) in [2]
	} else if (mode == Xmm::FAST_BURST) {
		double p = _coefs[modeId][1] * 1.e-3,
			   a = _coefs[modeId][2] * 1.e-3;

		for (unsigned n=0; n<N; ++n)
			time[n] -= a + DBL(sourceY-1 - rawy[n]-1) * p;
														//	Eq. (11) in [2]
	} else
		errstr<<"Fine-time adjustment with source position in mode "<<mode
			  <<" is not supported - must be "<<Xmm::FAST_TIMING
			  <<" or "<<Xmm::FAST_BURST
			  <<Sas::error("UnsupportedOperation");

	RETURN time;
} // EpnTimeCorrect::correct()

